/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 
         International, see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.
--------------------------------------------------------------------------------

You can use this macro to set up a basic GNG environment. Global variables in
this macro will be accessible at the ROOT command line after the macro was 
loaded.

Common usage in the ROOT command line would be, e.g.:

root [0] .x tgng_load.C
root [1] .L macros/basic_setup.C
root [2] init_gng(new TGNGVectorData(3), new TGNGVectorData(3), new TGNGMinkowskiDistance(2));

Now, use the globally defined TGNG object "gng" for your purposes. Have a look
at randomtest.C as an example or just run:

    testgng_rnd(Int_t count, Int_t dimension);

as a test. In the latter case, the gng should be initialized with TGNGVectorData
and the parameter "dimension" should match the dimension used during 
initialization.

If you use randomtest.C you should also have a look at analysetest.C for an
example on how to read back the data generated by randomtest.C

*/

// global variables:

// main GNG instance
TGNG *gng = 0;

// some visualizations
TGNGMatrixVis    *vis1 = 0;
TGNGFruReVis     *vis2 = 0;
TGNGHiClusterVis *vis3 = 0;

// a canvas
TCanvas *gng_can = 0;

void init_gng(TGNGDataElement     *prototypeA, 
	          TGNGDataElement     *prototypeB, 
	          TGNGDistanceMeasure *distMeasure)
{
	// create the gng and its extensions
	gng  = new TGNG("mygng","first gng test");

	vis1 = new TGNGMatrixVis();
	vis2 = new TGNGFruReVis();
	vis3 = new TGNGHiClusterVis();

	// Add these and other extensions. Note that extensions are owned by the
	// gng and their memory has not to be released. Thus, if we have some 
	// extensions with which we do not plan to interact with, we can safely add
	// them directly to the GNG without storing them in some global variable.

	gng->AddExtension(new TGNGLogExt());  // adding simple stdout logging
	gng->AddExtension(new TGNGIdExt());   // adding unique unit IDs
	gng->AddExtension(new TGNGLishExt()); // adding local input space histograms

	// adding hierarchical clustering
	TGNGHierarchicalClusterExt *hCluster = new TGNGHierarchicalClusterExt();
	hCluster->SetDistanceType(TGNGHierarchicalClusterExt::dtLISH);
	hCluster->SetLinkageType(TGNGHierarchicalClusterExt::ltSingle);
	gng->AddExtension(hCluster);

	// adding visualizations
	gng->AddExtension(vis1);
	gng->AddExtension(vis2);
	gng->AddExtension(vis3);

	// Create a canvas and setup visualizations
	gng_can = new TCanvas();

	// we want 3 drawing pads 
	gng_can->Divide(3,1);

    // draw visualizations in the pads
    gng_can->cd(1);
    gPad->SetMargin(0,0,0,0);
    vis1->Draw("colit!e"); // see documentation of TGNGMatrixVis::Draw() for an
                           // explanation of these options

    vis1->SetUpdateInterval(0.1); // update the visualization every 0.1 seconds
                                  // if there is new input

    gng_can->cd(2);
    vis2->Draw("t");
    vis2->SetUpdateInterval(0.1);

    gng_can->cd(3);
    gPad->SetMargin(0,0,0,0);
    vis3->Draw("t");
    vis3->SetUpdateInterval(1); // update the clustering visualization somewhat
                                // slower

    // set gng parameters a bit more conservative than the values proposed by
    // Fritzke
	gng->SetMaxUnitCnt(50);	
	gng->SetLambda(2000);
	gng->SetT_max(500);
	gng->SetEpsilon_b(0.01);
	gng->SetEpsilon_n(0.0001);

	// Initialize the gng
	gng->Init(prototypeA,prototypeB,distMeasure);
}

void testgng_rnd(Int_t count, Int_t dimension)
{
	TGNGVectorData *data = new TGNGVectorData(dimension);
	for (Int_t i = 0; i < count; ++i) {
		printf("input %d of %d           \r",i+1,count);
		fflush(stdout);
		for (Int_t j = 0; j < dimension; ++j) {
			data->fVector[j] = gRandom->Uniform();
		}
		gng->Feed(data);
	}
	printf("\ndone.\n");
	delete data;
}
