/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 
         International, see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.
--------------------------------------------------------------------------------

This macro contains a number of functions that set up RGNG experiments for 
different types of input spaces. Typically the functions in this macro are
used by "run"-Macros that define individual experiments.

Common usage in the ROOT command line would be, e.g.:

root [0] .x tgng_load.C
root [1] .L macros/grid_inputs.C+
root [2] .L macros/grid_example.C
root [3] run("example.root",2000000)

See documentation of structs and functions in this macro for further details.
*/


#include <cstring>
#include <vector>

#include <TMath.h>
#include <TRandom.h>
#include <TTree.h>
#include <TFile.h>
#include <TVector2.h>
#include <TObjArray.h>
#include <TTimeStamp.h>
#include <TCanvas.h>

#include "src/core/TGNG.h"
#include "src/core/TGNGUnit.h"
#include "src/distmeasures/TGNGMinkowskiDistance.h"
#include "src/datatypes/TGNGVectorData.h"
#include "src/datatypes/TGNGRecursiveData.h"
#include "src/extensions/log/TGNGFullLoggerExt.h"
#include "src/extensions/log/TGNGRecLogExt.h"
#include "src/extensions/functional/TGNGLishExt.h"
#include "src/extensions/functional/TGNGActivityExt.h"

/*
	This structure contains all parameters for a RGNG grid cell experiment
*/
struct LongTestParameters {
	Int_t topLevelNodes;			// maximum number of top layer units
	Int_t bottomLevelNodes;			// maximum number of bottom layer units

	Double_t tlPrimaryRate;			// primary TL learning rate (epsilon_b)
	Double_t tlPrimaryRateEnd;		// primary TL learning rate after ramping (if enabled)
	Double_t tlSecondaryRate; 		// secondary TL learning rate (epsilon_n)
	Double_t tlSecondaryRateEnd; 	// secondary TL learning rate after ramping (if enabled)
	Double_t tlRamp; 				// TL ramp duration of learning rate change 	
    Double_t tlGamma;				// TL recovery rate from learning inhibition(if enabled)
	Bool_t   tlUseRamp; 			// enables learning rate ramping in the TL
	Bool_t   tlUseLearnInhib;		// enables learning inhibition in the TL
	UInt_t   tlLambda;				// TL interval at which new units are added
	UInt_t   tlTmax;				// maximum age of TL edges

	Double_t blPrimaryRate; 		// primary BL learning rate (epsilon_b)
	Double_t blPrimaryRateEnd; 		// primary BL learning rate after ramping (if enabled)
	Double_t blSecondaryRate; 		// secondary BL learning rate (epsilon_n)
	Double_t blSecondaryRateEnd; 	// secondary BL learning rate after ramping (if enabled)
	Double_t blRamp; 				// BL ramp duration of learning rate change 	
    Double_t blGamma; 				// BL recovery rate from learning inhibition(if enabled)
	Bool_t   blUseRamp; 			// enables learning rate ramping in the BL
	Bool_t   blUseLearnInhib;		// enables learning inhibition in the BL
	UInt_t   blLambda;				// BL interval at which new units are added
	UInt_t   blTmax;				// maximum age of BL edges

    Double_t nbAttenuation; 		// TL and BL tertiary learning rate or epsilon_r

    Double_t minkowskiParam;		// exponent of the minkowski distance
    TGNGMinkowskiDistance::RecursiveMode recMode; // recursive mode of the minkowski distance

    Int_t inputDim;					// dimension of the input space in case of two orthogonal, one-dimensional attractor inputs
    Int_t inpFade;					// fading of position indication in case of two orthogonal, one-dimensional attractor inputs
    Int_t inpResample;				// resampling of input space dimensions to avoid aliasing artefacts

    Int_t logInterval;				// interval at which the whole RGNG should be logged
};



/*
	Helper function that generates a list of "nrOfPlaces" two-dimensional
	vectors with random positions between 0 and 1 in each dimension.
*/
TObjArray* getRandomPlaces(Int_t nrOfPlaces)
{
	TObjArray *result = new TObjArray();
	result->SetOwner(kTRUE);
	for (Int_t i = 0; i < nrOfPlaces; ++i)
		result->Add(new TVector2(gRandom->Uniform(),gRandom->Uniform()));
	return result;
}

/*
	Helper function that selects a random subset of size "subsetSize" from
	"nrOfPlaces" entries.
*/
std::vector<Bool_t>* getRandomSubset(Int_t nrOfPlaces, Int_t subsetSize)
{
	std::vector<Bool_t> *result = new std::vector<Bool_t>(nrOfPlaces,kFALSE);
	for (Int_t i = 0; i < subsetSize; ++i) {
		Int_t rIdx;
		while (result->at(rIdx = gRandom->Integer(nrOfPlaces)));
		result->at(rIdx) = kTRUE;
	}
	return result;
}

/*
	Helper function that determines the activities of a number of place cells
	whose centers are provided in "places". The activity of the place cells is
	modelled as a gaussian function with standard deviation "sigma" and sampled
	at the reference position "refPos". The resulting activities of the active
	subset of place cells ("subset") are stored in the output vector "data" 
	starting at the given offset "offset".
*/
void setPlaceActivity(TVector2 *refPos, TObjArray *places, Double_t sigma, 
	std::vector<Bool_t> *subset, TGNGVectorData *data, Int_t offset)
{
	Double_t tss = 2*sigma*sigma;
	Int_t nrOfPlaces = places->GetEntriesFast();
	for (Int_t i = 0; i < nrOfPlaces; ++i) {
		if (subset->at(i) == kFALSE) {
			data->fVector[offset++] = 0.0;
			continue;
		}
		TVector2 *curPos = (TVector2*)places->At(i);
		Double_t ddx = curPos->X() - refPos->X();
		Double_t ddy = curPos->Y() - refPos->Y();
		data->fVector[offset++] = TMath::Exp(-(ddx*ddx+ddy*ddy)/tss);
	}
}

/*
	Debug function that plots artificial place cell activity as gaussian 
	distributions with standard deviation "sigma" at the positions provided
	in "places". The resulting map has a resolution of "res" x "res" and is
	plotted with the option "option" on canvas "can". If no canvas is provided
	a new one is created.
*/
void plotPlaces(TObjArray *places, Double_t sigma, Int_t res, Option_t* option = "", TCanvas *can = 0)
{
	if (can == 0) {
		can = new TCanvas();
	}

	TMatrixD mat(res,res);

	Int_t nrOfPlaces = places->GetEntriesFast();

	Double_t tss = 2*sigma*sigma;

	for (Int_t x = 0; x < res; ++x)
		for (Int_t y = 0; y < res; ++y) {
			Double_t dx = (Double_t)x / (Double_t)res;
			Double_t dy = (Double_t)y / (Double_t)res;
			Double_t minDist = 2*res;
			Int_t minIdx = 0;
			for (Int_t k = 0; k < nrOfPlaces; ++k) {
				TVector2 *curPos = (TVector2*)places->At(k);
				Double_t ddx = curPos->X() - dx;
				Double_t ddy = curPos->Y() - dy;
				Double_t curDist = ddx*ddx+ddy*ddy;
				if (curDist < minDist) {
					minDist = curDist;
					minIdx  = k;
				}
			}
			TVector2 *minPos = (TVector2*)places->At(minIdx);
			Double_t ddx = minPos->X() - dx;
			Double_t ddy = minPos->Y() - dy;
			mat(y,x) = TMath::Exp(-(ddx*ddx+ddy*ddy)/tss);
		}

	mat.Draw(option);

}


/*
	This structure summarizes the parameters of a group of place cells.
*/
struct PlaceParameters {
	Int_t                numberOfCells; // number of cells in the place cell group
	Double_t             sigma;         // standard deviation of the cells firing fields
	TObjArray           *places;        // centers of the cells firing fields
	std::vector<Bool_t> *subset;        // active subset of cells
};

/*
	Helper function to determine remaining computation time
*/
int getRemaining(Int_t slice, Int_t remaining, char *output)
{
	static Bool_t initialized = kFALSE;
	static TTimeStamp old;
	if (initialized == kFALSE) {
		TTimeStamp now;
		initialized = kTRUE;
		old = now;
		return 0;
	}
	TTimeStamp now;
	Double_t diff = now.AsDouble() - old.AsDouble();
	old = now;
	diff *= (Double_t)remaining;
	diff /= (Double_t)slice;
	now.SetSec(now.GetSec() + diff);
	return sprintf(output,"EST: %s", now.AsString("l"));
}

/*
	This function feeds "cnt" random position vectors into the RGNG "gng".
	The position vectors have the dimension "dim" and represent the position
	as output of two, one-dimensional attractor networks. The width of the
	"activity bump" in these attractor networks is determined by the parameter
	"fade". The higher "fade" is, the narrower the bump is.
*/
void feedRandom(Int_t cnt, TGNG *gng, Int_t dim, Int_t fade)
{
	
	Int_t dim2 = dim / 2;

	TVector2 rawInput;

	Double_t dx;
	Double_t dy;

	char buffer[500] = "";

	TGNGVectorData *data = new TGNGVectorData(dim);
	for (Int_t i = 0; i < cnt; ++i) {

		if ((i % 10000) == 0) 
		{
			getRemaining(10000,cnt-i-1,buffer);
		}

		if ((i % 100) == 0) 
		{
			printf("input %d of %d   %s           \r",i+1,cnt,buffer);
			fflush(stdout);
		}


		data->fVector.Zero();

		dx = gRandom->Uniform();
		dy = gRandom->Uniform();

		rawInput.Set(dx,dy);

		Int_t x = (Int_t)(TMath::Floor(dx * (Double_t)(dim2) + 0.5));
		Int_t y = (Int_t)(TMath::Floor(dy * (Double_t)(dim2) + 0.5));

		if (x >= dim2) x = (dim2)-1;
		if (y >= dim2) y = (dim2)-1;

		Int_t fadeDist = dim/fade;

		for (Int_t j = x; j < x + fadeDist; ++j)
			data->fVector[j % dim2] = 1.0 - ((Double_t)(j-x) / (Double_t)fadeDist);

		for (Int_t j = x+dim2; j > x+dim2-fadeDist; --j)
			data->fVector[j % dim2] = 1.0 - ((Double_t)(x+dim2-j) / (Double_t)fadeDist);

		for (Int_t j = y; j < y + fadeDist; ++j)
			data->fVector[dim2 + (j % dim2)] = 1.0 - ((Double_t)(j-y) / (Double_t)fadeDist);

		for (Int_t j = y+dim2; j > y+dim2-fadeDist; --j)
			data->fVector[dim2 + (j % dim2)] = 1.0 - ((Double_t)(y+dim2-j) / (Double_t)fadeDist);

		gng->Feed(data,0,&rawInput);
	}
	printf("\ndone.\n");
	delete data;
	
}

/*
	This function feeds "cnt" random position vectors into the RGNG "gng".
	The position vectors have the dimension "dim" and represent the position
	as output of two, one-dimensional attractor networks. The width of the
	"activity bump" in these attractor networks is determined by the parameter
	"fade". The higher "fade" is, the narrower the bump is. The additional
	parameter "resample" determines an internal increase of the dimension of
	the input vector. For example, with "resample = 8" the internal dimension
	of the vector would be (dim*resample) while the final vector that is fed
	to the gng has still only "dim" dimensions. The internal vector is 
	downsampled by averaging to the final dim dimensions. This resampling avoids
	potential aliasing artefacts.
*/
void feedRandom(Int_t cnt, TGNG *gng, Int_t dim, Int_t fade, Int_t resample)
{
	
	Int_t dim2 = dim / 2;

	TVector2 rawInput;

	Double_t dx;
	Double_t dy;

	Int_t dimR  = dim * resample;
	Int_t dim2R = dimR / 2;

	Double_t *rData = new Double_t[dimR];

	char buffer[500] = "";

	TGNGVectorData *data = new TGNGVectorData(dim);
	for (Int_t i = 0; i < cnt; ++i) {

		if ((i % 10000) == 0) 
		{
			getRemaining(10000,cnt-i-1,buffer);
		}

		if ((i % 1000) == 0) 
		{
			printf("input %d of %d   %s           \r",i+1,cnt,buffer);
			fflush(stdout);
		}

		// build resample data
		//data->fVector.Zero();
		std::memset(rData,0,sizeof(Double_t)*dimR);

		dx = gRandom->Uniform();
		dy = gRandom->Uniform();

		rawInput.Set(dx,dy);

		Int_t x = (Int_t)(TMath::Floor(dx * (Double_t)(dim2R) + 0.5));
		Int_t y = (Int_t)(TMath::Floor(dy * (Double_t)(dim2R) + 0.5));

		if (x >= dim2R) x = (dim2R)-1;
		if (y >= dim2R) y = (dim2R)-1;

		Int_t fadeDist = dimR/fade;

		for (Int_t j = x; j < x + fadeDist; ++j)
			rData[j % dim2R] = 1.0 - ((Double_t)(j-x) / (Double_t)fadeDist);

		for (Int_t j = x+dim2R; j > x+dim2R-fadeDist; --j)
			rData[j % dim2R] = 1.0 - ((Double_t)(x+dim2R-j) / (Double_t)fadeDist);

		for (Int_t j = y; j < y + fadeDist; ++j)
			rData[dim2R + (j % dim2R)] = 1.0 - ((Double_t)(j-y) / (Double_t)fadeDist);

		for (Int_t j = y+dim2R; j > y+dim2R-fadeDist; --j)
			rData[dim2R + (j % dim2R)] = 1.0 - ((Double_t)(y+dim2R-j) / (Double_t)fadeDist);

		// resample rData into data->fVector
		for (Int_t j = 0; j < dim; ++j) {
			data->fVector[j] = 0.0;
			for (Int_t k = j * resample; k < (j+1) * resample; ++k) {
				data->fVector[j] += rData[k];
			}
			data->fVector[j] /= (Double_t)resample;
		}

		gng->Feed(data,0,&rawInput);
	}
	printf("\ndone.\n");
	delete data;

	delete[] rData;	
}

/*
	This function feeds "cnt" random position vectors into the RGNG "gng" like
	the feedRandom() functions. The difference is, that the coordinate system
	of the assumed two, one-dimensional attractor networks is rotated by 
	"angle".
*/
void feedRandomRot(Int_t cnt, TGNG *gng, Int_t dim, Int_t fade, Int_t resample, Double_t angle)
{
	
	Int_t dim2 = dim / 2;

	TVector2 rawInput;

	TVector2 rotation;

	Double_t dx;
	Double_t dy;

	Int_t dimR  = dim * resample;
	Int_t dim2R = dimR / 2;

	Double_t *rData = new Double_t[dimR];

	char buffer[500] = "";

	TGNGVectorData *data = new TGNGVectorData(dim);
	for (Int_t i = 0; i < cnt; ++i) {

		if ((i % 10000) == 0) 
		{
			getRemaining(10000,cnt-i-1,buffer);
		}

		if ((i % 1000) == 0) 
		{
			printf("input %d of %d   %s           \r",i+1,cnt,buffer);
			fflush(stdout);
		}

		// build resample data
		//data->fVector.Zero();
		std::memset(rData,0,sizeof(Double_t)*dimR);

		dx = gRandom->Uniform();
		dy = gRandom->Uniform();

		rawInput.Set(dx,dy);

		Double_t rdx = ((dx-0.5)*TMath::Cos(angle)-(dy-0.5)*TMath::Sin(angle)) + 0.5;
		Double_t rdy = ((dx-0.5)*TMath::Sin(angle)+(dy-0.5)*TMath::Cos(angle)) + 0.5;

		if (rdx < 0.0) rdx += 1.0;
		if (rdx > 1.0) rdx -= 1.0;
		if (rdy < 0.0) rdy += 1.0;
		if (rdy > 1.0) rdy -= 1.0;

		Int_t x = (Int_t)(TMath::Floor(rdx * (Double_t)(dim2R) + 0.5));
		Int_t y = (Int_t)(TMath::Floor(rdy * (Double_t)(dim2R) + 0.5));

		if (x >= dim2R) x = (dim2R)-1;
		if (y >= dim2R) y = (dim2R)-1;

		Int_t fadeDist = dimR/fade;

		for (Int_t j = x; j < x + fadeDist; ++j)
			rData[j % dim2R] = 1.0 - ((Double_t)(j-x) / (Double_t)fadeDist);

		for (Int_t j = x+dim2R; j > x+dim2R-fadeDist; --j)
			rData[j % dim2R] = 1.0 - ((Double_t)(x+dim2R-j) / (Double_t)fadeDist);

		for (Int_t j = y; j < y + fadeDist; ++j)
			rData[dim2R + (j % dim2R)] = 1.0 - ((Double_t)(j-y) / (Double_t)fadeDist);

		for (Int_t j = y+dim2R; j > y+dim2R-fadeDist; --j)
			rData[dim2R + (j % dim2R)] = 1.0 - ((Double_t)(y+dim2R-j) / (Double_t)fadeDist);

		// resample rData into data->fVector
		for (Int_t j = 0; j < dim; ++j) {
			data->fVector[j] = 0.0;
			for (Int_t k = j * resample; k < (j+1) * resample; ++k) {
				data->fVector[j] += rData[k];
			}
			data->fVector[j] /= (Double_t)resample;
		}

		gng->Feed(data,0,&rawInput);
	}
	printf("\ndone.\n");
	delete data;

	delete[] rData;	
}


/*
	This function feeds "cnt" artifical place cell group inputs into the RGNG
	"gng". The place cell groups are defined as an array of PlaceParameter
	structs ("places"). The dimension of the resulting input vector corresponds
	to the overall number of place cells defined in the "places" array.
*/
void feedRandomPlace(Int_t cnt, TGNG *gng, std::vector<PlaceParameters*> *places)
{
	Int_t feedbackSize = 0;

	for (PlaceParameters *pp : *places)
		feedbackSize += pp->numberOfCells;

	TVector2 rawInput;

	Double_t dx;
	Double_t dy;

	char buffer[500] = "";

	TGNGVectorData *data = new TGNGVectorData(feedbackSize);
	for (Int_t i = 0; i < cnt; ++i) {
		if ((i % 10000) == 0) 
		{
			getRemaining(10000,cnt-i-1,buffer);
		}

		if ((i % 100) == 0) 
		{
			printf("input %d of %d   %s           \r",i+1,cnt,buffer);
			fflush(stdout);
		}

		// build resample data
		data->fVector.Zero();

		dx = gRandom->Uniform();
		dy = gRandom->Uniform();

		rawInput.Set(dx,dy);

		// fill in place cell data
		Int_t offset = 0;
		for (PlaceParameters *pp : *places) {
			setPlaceActivity(&rawInput, pp->places, pp->sigma, pp->subset, data, offset);
			offset += pp->numberOfCells;
		}

		gng->Feed(data,0,&rawInput);
	}
	printf("\ndone.\n");
	delete data;

}

/*
	This function feeds "cnt" constant (center) position vectors into the RGNG 
	"gng". The position vectors have the dimension "dim" and represent the 
	position as output of two, one-dimensional attractor networks. The width of 
	the	"activity bump" in these attractor networks is determined by the 
	parameter "fade". The higher "fade" is, the narrower the bump is. 
*/
void feedConst(Int_t cnt, TGNG *gng, Int_t dim, Int_t fade)
{
	
	Int_t dim2 = dim / 2;

	char buffer[500] = "";

	TGNGVectorData *data = new TGNGVectorData(dim);
	for (Int_t i = 0; i < cnt; ++i) {
		if ((i % 10000) == 0) 
		{
			getRemaining(10000,cnt-i-1,buffer);
		}

		if ((i % 100) == 0) 
		{
			printf("input %d of %d   %s           \r",i+1,cnt,buffer);
			fflush(stdout);
		}

		data->fVector.Zero();

		Int_t x = (Int_t)(TMath::Floor(0.5 * (Double_t)(dim2) + 0.5));
		Int_t y = (Int_t)(TMath::Floor(0.5 * (Double_t)(dim2) + 0.5));

		if (x >= dim2) x = (dim2)-1;
		if (y >= dim2) y = (dim2)-1;

		Int_t fadeDist = dim/fade;

		for (Int_t j = x; j < x + fadeDist; ++j)
			data->fVector[j % dim2] = 1.0 - ((Double_t)(j-x) / (Double_t)fadeDist);

		for (Int_t j = x+dim2; j > x+dim2-fadeDist; --j)
			data->fVector[j % dim2] = 1.0 - ((Double_t)(x+dim2-j) / (Double_t)fadeDist);

		for (Int_t j = y; j < y + fadeDist; ++j)
			data->fVector[dim2 + (j % dim2)] = 1.0 - ((Double_t)(j-y) / (Double_t)fadeDist);

		for (Int_t j = y+dim2; j > y+dim2-fadeDist; --j)
			data->fVector[dim2 + (j % dim2)] = 1.0 - ((Double_t)(y+dim2-j) / (Double_t)fadeDist);

		gng->Feed(data);
	}
	printf("\ndone.\n");
	delete data;
	
}

/*
	This function feeds posTree->GetEntries() position vectors into the RGNG 
	"gng". The positions stem from the TTree "posTree", which is expected to
	provide two branches "posx" and "posy" of double-valued locations. The 
	position vectors have the dimension "dim" and represent the position as 
	output of two, one-dimensional attractor networks. The width of the	
	"activity bump" in these attractor networks is determined by the parameter 
	"fade". The higher "fade" is, the narrower the bump is. The parameters
	"scale" and "shift" are used to adapt the positions read from the tree by:

		dx = (dx + shift) * scale;
		if (dx > 1) dx = 1;
		if (dx < 0) dx = 0;

		dy = (dy + shift) * scale;
		if (dy > 1) dy = 1;
		if (dy < 0) dy = 0;

*/
void feedPosTree(TTree *posTree, TGNG *gng, Int_t dim, Int_t fade, Double_t scale, Double_t shift)
{

	Double_t dx;
	Double_t dy;

	Int_t nrOfEntries = posTree->GetEntries();

	TVector2 rawInput;

    posTree->SetBranchStatus("*",0);
    posTree->SetBranchStatus("posx",1);
    posTree->SetBranchStatus("posy",1);

    posTree->SetBranchAddress("posx",&dx);
    posTree->SetBranchAddress("posy",&dy);
	
	Int_t dim2 = dim / 2;

	char buffer[500] = "";

	TGNGVectorData *data = new TGNGVectorData(dim);
	for (Int_t i = 0; i < nrOfEntries; ++i) {
		if ((i % 10000) == 0) 
		{
			getRemaining(10000,nrOfEntries-i-1,buffer);
		}

		if ((i % 100) == 0) 
		{
			printf("input %d of %d   %s           \r",i+1,nrOfEntries,buffer);
			fflush(stdout);
		}

		data->fVector.Zero();

		posTree->GetEntry(i);

		if (TMath::IsNaN(dx) || TMath::IsNaN(dy)) {
			printf("NaaaaaaaaaaaaN\n");
			continue;
		}

		dx = (dx + shift) * scale;
		if (dx > 1) dx = 1;
		if (dx < 0) dx = 0;

		dy = (dy + shift) * scale;
		if (dy > 1) dy = 1;
		if (dy < 0) dy = 0;

		rawInput.Set(dx,dy);

		Int_t x = (Int_t)(TMath::Floor(dx * (Double_t)(dim2) + 0.5));
		Int_t y = (Int_t)(TMath::Floor(dy * (Double_t)(dim2) + 0.5));

		if (x >= dim2) x = (dim2)-1;
		if (y >= dim2) y = (dim2)-1;

		if (x < 0) x = 0;
		if (y < 0) y = 0;

		//printf("feeding: %d / %d\n", x,y);

		Int_t fadeDist = dim/fade;

		for (Int_t j = x; j < x + fadeDist; ++j)
			data->fVector[j % dim2] = 1.0 - ((Double_t)(j-x) / (Double_t)fadeDist);

		for (Int_t j = x+dim2; j > x+dim2-fadeDist; --j)
			data->fVector[j % dim2] = 1.0 - ((Double_t)(x+dim2-j) / (Double_t)fadeDist);

		for (Int_t j = y; j < y + fadeDist; ++j)
			data->fVector[dim2 + (j % dim2)] = 1.0 - ((Double_t)(j-y) / (Double_t)fadeDist);

		for (Int_t j = y+dim2; j > y+dim2-fadeDist; --j)
			data->fVector[dim2 + (j % dim2)] = 1.0 - ((Double_t)(y+dim2-j) / (Double_t)fadeDist);

		gng->Feed(data,0,&rawInput);
		//gng->Peek(data);
	}
	printf("\ndone.\n");
	delete data;
	
}


/*
	This function feeds posTree->GetEntries() position vectors into the RGNG 
	"gng". The positions stem from the TTree "posTree", which is expected to
	provide two branches "posx" and "posy" of double-valued locations. The 
	position vectors have the dimension "dim" and represent the position as 
	output of two, one-dimensional attractor networks. The width of the	
	"activity bump" in these attractor networks is determined by the parameter 
	"fade". The higher "fade" is, the narrower the bump is. The parameters
	"scale" and "shift" are used to adapt the positions read from the tree by:

		dx = (dx + shift) * scale;
		if (dx > 1) dx = 1;
		if (dx < 0) dx = 0;

		dy = (dy + shift) * scale;
		if (dy > 1) dy = 1;
		if (dy < 0) dy = 0;
	
	The additional parameter "resample" determines an internal increase of the 
	dimension of the input vector. For example, with "resample = 8" the internal
	dimension of the vector would be (dim*resample) while the final vector that 
	is fed to the gng has still only "dim" dimensions. The internal vector is 
	downsampled by averaging to the final dim dimensions. This resampling avoids
	potential aliasing artefacts.

	The parameters "angle" and "yScale" allow to rotate and skew the coordinate
	system of the presumed attractor network inputs.
*/
void feedPosTree(TTree *posTree, TGNG *gng, Int_t dim, Int_t fade, 
	Double_t scale, Double_t shift, Int_t resample, Double_t angle = 0.0, 
	Double_t yScale = 1.0)
{

	Double_t dx;
	Double_t dy;

	Int_t nrOfEntries = posTree->GetEntries();

	TVector2 rawInput;

    posTree->SetBranchStatus("*",0);
    posTree->SetBranchStatus("posx",1);
    posTree->SetBranchStatus("posy",1);

    posTree->SetBranchAddress("posx",&dx);
    posTree->SetBranchAddress("posy",&dy);
	
	Int_t dim2 = dim / 2;

	Int_t dimR  = dim * resample;
	Int_t dim2R = dimR / 2;

	Double_t *rData = new Double_t[dimR];

	char buffer[500] = "";

	TGNGVectorData *data = new TGNGVectorData(dim);
	for (Int_t i = 0; i < nrOfEntries; ++i) {
		if ((i % 10000) == 0) 
		{
			getRemaining(10000,nrOfEntries-i-1,buffer);
		}

		if ((i % 100) == 0) 
		{
			printf("input %d of %d   %s           \r",i+1,nrOfEntries,buffer);
			fflush(stdout);
		}

		//data->fVector.Zero();
		std::memset(rData,0,sizeof(Double_t)*dimR);

		posTree->GetEntry(i);

		if (TMath::IsNaN(dx) || TMath::IsNaN(dy)) {
			printf("NaaaaaaaaaaaaN\n");
			continue;
		}

		dx = (dx + shift) * scale;
		if (dx > 1) dx = 1;
		if (dx < 0) dx = 0;

		dy = (dy + shift) * scale;
		if (dy > 1) dy = 1;
		if (dy < 0) dy = 0;

		rawInput.Set(dx,dy*yScale);

		Double_t rdx = dx;
		Double_t rdy = dy;

		if (angle != 0.0) {

			rdx = ((dx-0.5)*TMath::Cos(angle)-(dy-0.5)*TMath::Sin(angle)) + 0.5;
			rdy = ((dx-0.5)*TMath::Sin(angle)+(dy-0.5)*TMath::Cos(angle)) + 0.5;

			if (rdx < 0.0) rdx += 1.0;
			if (rdx > 1.0) rdx -= 1.0;
			if (rdy < 0.0) rdy += 1.0;
			if (rdy > 1.0) rdy -= 1.0;
		}

		Int_t x = (Int_t)(TMath::Floor(rdx * (Double_t)(dim2R) + 0.5));
		Int_t y = (Int_t)(TMath::Floor(rdy * (Double_t)(dim2R) + 0.5));

		if (x >= dim2R) x = (dim2R)-1;
		if (y >= dim2R) y = (dim2R)-1;

		if (x < 0) x = 0;
		if (y < 0) y = 0;

		//printf("feeding: %d / %d\n", x,y);

		Int_t fadeDist = dimR/fade;

		for (Int_t j = x; j < x + fadeDist; ++j)
			rData[j % dim2R] = 1.0 - ((Double_t)(j-x) / (Double_t)fadeDist);

		for (Int_t j = x+dim2R; j > x+dim2R-fadeDist; --j)
			rData[j % dim2R] = 1.0 - ((Double_t)(x+dim2R-j) / (Double_t)fadeDist);

		for (Int_t j = y; j < y + fadeDist; ++j)
			rData[dim2R + (j % dim2R)] = 1.0 - ((Double_t)(j-y) / (Double_t)fadeDist);

		for (Int_t j = y+dim2R; j > y+dim2R-fadeDist; --j)
			rData[dim2R + (j % dim2R)] = 1.0 - ((Double_t)(y+dim2R-j) / (Double_t)fadeDist);

		// resample rData into data->fVector
		for (Int_t j = 0; j < dim; ++j) {
			data->fVector[j] = 0.0;
			for (Int_t k = j * resample; k < (j+1) * resample; ++k) {
				data->fVector[j] += rData[k];
			}
			data->fVector[j] /= (Double_t)resample;
		}

		gng->Feed(data,0,&rawInput);
		//gng->Peek(data);
	}
	printf("\ndone.\n");
	delete data;

	delete[] rData;	
	
}

/*
	This function feeds posTree->GetEntries() position vectors into the RGNG 
	"gng". The positions stem from the TTree "posTree", which is expected to
	provide two branches "posx" and "posy" of double-valued locations. The 
	position vectors have the dimension "dim" and represent the position as 
	output of two, one-dimensional attractor networks. The width of the	
	"activity bump" in these attractor networks is determined by the parameter 
	"fade". The higher "fade" is, the narrower the bump is. The parameters
	"scale" and "shift" are used to adapt the positions read from the tree by:

		dx = (dx + shift) * scale;
		if (dx > 1) dx = 1;
		if (dx < 0) dx = 0;

		dy = (dy + shift) * scale;
		if (dy > 1) dy = 1;
		if (dy < 0) dy = 0;
	
	The additional parameter "resample" determines an internal increase of the 
	dimension of the input vector. For example, with "resample = 8" the internal
	dimension of the vector would be (dim*resample) while the final vector that 
	is fed to the gng has still only "dim" dimensions. The internal vector is 
	downsampled by averaging to the final dim dimensions. This resampling avoids
	potential aliasing artefacts.

	The parameter "noiseLevel" allows to add uniform random noise to the input 
	vector. The noise is added per vector element as follows:

		data->fVector[j] += (gRandom->Uniform() - 0.5) * 2.0 * noiseLevel;
		if (data->fVector[j] > 1.0) data->fVector[j] = 1.0;
		if (data->fVector[j] < 0.0) data->fVector[j] = 0.0;
*/
void feedPosTreeNoise(TTree *posTree, TGNG *gng, Int_t dim, Int_t fade, Double_t scale, Double_t shift, Int_t resample, Double_t noiseLevel)
{

	Double_t dx;
	Double_t dy;

	Int_t nrOfEntries = posTree->GetEntries();

	TVector2 rawInput;

    posTree->SetBranchStatus("*",0);
    posTree->SetBranchStatus("posx",1);
    posTree->SetBranchStatus("posy",1);

    posTree->SetBranchAddress("posx",&dx);
    posTree->SetBranchAddress("posy",&dy);
	
	Int_t dim2 = dim / 2;

	Int_t dimR  = dim * resample;
	Int_t dim2R = dimR / 2;

	Double_t *rData = new Double_t[dimR];

	char buffer[500] = "";

	TGNGVectorData *data = new TGNGVectorData(dim);
	for (Int_t i = 0; i < nrOfEntries; ++i) {
		if ((i % 10000) == 0) 
		{
			getRemaining(10000,nrOfEntries-i-1,buffer);
		}

		if ((i % 100) == 0) 
		{
			printf("input %d of %d   %s           \r",i+1,nrOfEntries,buffer);
			fflush(stdout);
		}

		//data->fVector.Zero();
		std::memset(rData,0,sizeof(Double_t)*dimR);

		posTree->GetEntry(i);

		if (TMath::IsNaN(dx) || TMath::IsNaN(dy)) {
			printf("NaaaaaaaaaaaaN\n");
			continue;
		}

		dx = (dx + shift) * scale;
		if (dx > 1) dx = 1;
		if (dx < 0) dx = 0;

		dy = (dy + shift) * scale;
		if (dy > 1) dy = 1;
		if (dy < 0) dy = 0;

		rawInput.Set(dx,dy);

		Int_t x = (Int_t)(TMath::Floor(dx * (Double_t)(dim2R) + 0.5));
		Int_t y = (Int_t)(TMath::Floor(dy * (Double_t)(dim2R) + 0.5));

		if (x >= dim2R) x = (dim2R)-1;
		if (y >= dim2R) y = (dim2R)-1;

		if (x < 0) x = 0;
		if (y < 0) y = 0;

		//printf("feeding: %d / %d\n", x,y);

		Int_t fadeDist = dimR/fade;

		for (Int_t j = x; j < x + fadeDist; ++j)
			rData[j % dim2R] = 1.0 - ((Double_t)(j-x) / (Double_t)fadeDist);

		for (Int_t j = x+dim2R; j > x+dim2R-fadeDist; --j)
			rData[j % dim2R] = 1.0 - ((Double_t)(x+dim2R-j) / (Double_t)fadeDist);

		for (Int_t j = y; j < y + fadeDist; ++j)
			rData[dim2R + (j % dim2R)] = 1.0 - ((Double_t)(j-y) / (Double_t)fadeDist);

		for (Int_t j = y+dim2R; j > y+dim2R-fadeDist; --j)
			rData[dim2R + (j % dim2R)] = 1.0 - ((Double_t)(y+dim2R-j) / (Double_t)fadeDist);

		// resample rData into data->fVector
		for (Int_t j = 0; j < dim; ++j) {
			data->fVector[j] = 0.0;
			for (Int_t k = j * resample; k < (j+1) * resample; ++k) {
				data->fVector[j] += rData[k];
			}
			data->fVector[j] /= (Double_t)resample;

			// add noise
			data->fVector[j] += (gRandom->Uniform() - 0.5) * 2.0 * noiseLevel;
			if (data->fVector[j] > 1.0) data->fVector[j] = 1.0;
			if (data->fVector[j] < 0.0) data->fVector[j] = 0.0;
		}

		gng->Feed(data,0,&rawInput);
		//gng->Peek(data);
	}
	printf("\ndone.\n");
	delete data;

	delete[] rData;	
	
}


/*
	This function feeds "cnt" random input vectors that simulate an efference
	copy of eye muscle output signals into the RGNG "gng". The input vector
	is a concatenation of four "dim"-dimensional vectors v_x0, v_x1, v_y0, v_y1
	which are calculated as follows:

	v_x0_i := max[ min[ 1 - delta * ((i+1)/dim - x),       1 ], 0 ],
	v_x1_i := max[ min[ 1 - delta * ((i+1)/dim - (1 - x)), 1 ], 0 ],
	v_y0_i := max[ min[ 1 - delta * ((i+1)/dim - y),       1 ], 0 ],
	v_y1_i := max[ min[ 1 - delta * ((i+1)/dim - (1 - y)), 1 ], 0 ],

	with i from 0 to dim-1, and delta = 4.

	The parameter "noiseLevel" allows to add uniform random noise to the input 
	vector. The noise is added per vector element as follows:

		data->fVector[j] += (gRandom->Uniform() - 0.5) * 2.0 * noiseLevel;
		if (data->fVector[j] > 1.0) data->fVector[j] = 1.0;
		if (data->fVector[j] < 0.0) data->fVector[j] = 0.0;
*/
void feedRandomEye(Int_t cnt, TGNG *gng, Int_t dim, Double_t noiseLevel = 0.0)
{
	
	TVector2 rawInput;

	Double_t dx;
	Double_t dy;

	char buffer[500] = "";

	Int_t x0offset = 0*dim;
	Int_t x1offset = 1*dim;
	Int_t y0offset = 2*dim;
	Int_t y1offset = 3*dim;

	TGNGVectorData *data = new TGNGVectorData(dim*4);
	for (Int_t i = 0; i < cnt; ++i) {

		if ((i % 10000) == 0) 
		{
			getRemaining(10000,cnt-i-1,buffer);
		}

		if ((i % 100) == 0) 
		{
			printf("input %d of %d   %s           \r",i+1,cnt,buffer);
			fflush(stdout);
		}


		data->fVector.Zero();

		dx = gRandom->Uniform();
		dy = gRandom->Uniform();

		rawInput.Set(dx,dy);

		Double_t sx0 = dx;
		Double_t sx1 = 1.0 - dx;
		Double_t sy0 = dy;
		Double_t sy1 = 1.0 - dy;

		Double_t stepWidth = 1.0 / (Double_t)dim;
		Int_t x0Thres = (Int_t)((Double_t)dim * sx0);
		Int_t x1Thres = (Int_t)((Double_t)dim * sx1);
		Int_t y0Thres = (Int_t)((Double_t)dim * sy0);
		Int_t y1Thres = (Int_t)((Double_t)dim * sy1);
		for (Int_t j = 0; j < dim; ++j) {
			Double_t cp = (Double_t)(j+1) * stepWidth;
			data->fVector[x0offset+j] = (j < x0Thres) ? 1.0 : 1.0 - (cp - sx0) * 4.;
			data->fVector[x1offset+j] = (j < x1Thres) ? 1.0 : 1.0 - (cp - sx1) * 4.;
			data->fVector[y0offset+j] = (j < y0Thres) ? 1.0 : 1.0 - (cp - sy0) * 4.;
			data->fVector[y1offset+j] = (j < y1Thres) ? 1.0 : 1.0 - (cp - sy1) * 4.;
		}			

		// add noise
		for (Int_t j = 0; j < dim*4; ++j) {
			data->fVector[j] += (gRandom->Uniform() - 0.5) * 2.0 * noiseLevel;
			if (data->fVector[j] > 1.0) data->fVector[j] = 1.0;
			if (data->fVector[j] < 0.0) data->fVector[j] = 0.0;
		}


		gng->Feed(data,0,&rawInput);
	}
	printf("\ndone.\n");
	delete data;
	
}


/*
	This function sets up an RGNG according to the "params" configuration, and
	uses the "feedRandom()" function to feed "sampleCnt" inputs to the RGNG. The
	whole process is logged using a TGNGFullLoggerExt as well as a TGNGRecLogExt
	into the file "logfile".
*/
void LongTest(LongTestParameters &params, const char* logfile, Int_t sampleCnt)
{
	TGNG *topGNG = new TGNG("topgng","top layer GNG");

	TGNGFullLoggerExt *logger = new TGNGFullLoggerExt(logfile);
	logger->SetGNGLogInterval(params.logInterval);
	topGNG->AddExtension(logger);
	topGNG->AddExtension(new TGNGRecLogExt(logger->GetCurrentLogFile()));

	TGNGMinkowskiDistance *distMeasure = new TGNGMinkowskiDistance(params.minkowskiParam);
	distMeasure->SetRecursiveMode(params.recMode);

	TGNGRecursiveData *rd1 = new TGNGRecursiveData();
	TGNGRecursiveData *rd2 = new TGNGRecursiveData();

	rd1->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);
	rd1->fGNG.SetLambda(params.blLambda);
	rd1->fGNG.SetT_max(params.blTmax);
	rd1->fGNG.SetLambdaDelay(0);
	rd1->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd1->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd1->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd1->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd1->fGNG.SetRampTime(params.blRamp);
	rd1->SetNbAttenuation(params.nbAttenuation);
	rd1->fGNG.SetGamma(params.blGamma);

	rd1->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd1->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd1->fGNG.AddExtension(new TGNGActivityExt());

	rd1->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	rd2->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);	
	rd2->fGNG.SetLambda(params.blLambda);
	rd2->fGNG.SetT_max(params.blTmax);
	rd2->fGNG.SetLambdaDelay(0);
	rd2->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd2->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd2->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd2->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd2->fGNG.SetRampTime(params.blRamp);
	rd2->SetNbAttenuation(params.nbAttenuation);
	rd2->fGNG.SetGamma(params.blGamma);

	rd2->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd2->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd2->fGNG.AddExtension(new TGNGActivityExt());

	rd2->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	topGNG->Init(rd1,rd2,distMeasure);

	topGNG->SetMaxUnitCnt(params.topLevelNodes);	
	topGNG->SetLambda(params.tlLambda);
	topGNG->SetT_max(params.tlTmax);
	topGNG->SetEpsilon_b(params.tlPrimaryRate);
	topGNG->SetEpsilon_be(params.tlPrimaryRateEnd);
	topGNG->SetEpsilon_n(params.tlSecondaryRate);
	topGNG->SetEpsilon_ne(params.tlSecondaryRateEnd);
	topGNG->SetRampTime(params.tlRamp);
	topGNG->SetGamma(params.tlGamma);

	topGNG->SetLearningInhibition(params.tlUseLearnInhib);
	topGNG->SetEpsilonRamp(params.tlUseRamp);

	feedRandom(sampleCnt, topGNG, params.inputDim,params.inpFade,params.inpResample);

	delete topGNG;
	delete distMeasure;
}

/*
	This function sets up an RGNG according to the "params" configuration, and
	uses the "feedRandomEye()" function to feed "sampleCnt" inputs to the RGNG. 
	The	whole process is logged using a TGNGFullLoggerExt as well as a 
	TGNGRecLogExt into the file "logfile". The additional parameter "noiseLevel"
	can be used to vary the noise in the input signal (see feedRandomEye() ).
*/
void LongTestEye(LongTestParameters &params, const char* logfile, Int_t sampleCnt, 
	Double_t noiseLevel = 0.0)
{
	TGNG *topGNG = new TGNG("topgng","top layer GNG");

	TGNGFullLoggerExt *logger = new TGNGFullLoggerExt(logfile);
	logger->SetGNGLogInterval(params.logInterval);
	topGNG->AddExtension(logger);
	topGNG->AddExtension(new TGNGRecLogExt(logger->GetCurrentLogFile()));

	TGNGMinkowskiDistance *distMeasure = new TGNGMinkowskiDistance(params.minkowskiParam);
	distMeasure->SetRecursiveMode(params.recMode);

	TGNGRecursiveData *rd1 = new TGNGRecursiveData();
	TGNGRecursiveData *rd2 = new TGNGRecursiveData();

	rd1->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);
	rd1->fGNG.SetLambda(params.blLambda);
	rd1->fGNG.SetT_max(params.blTmax);
	rd1->fGNG.SetLambdaDelay(0);
	rd1->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd1->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd1->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd1->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd1->fGNG.SetRampTime(params.blRamp);
	rd1->SetNbAttenuation(params.nbAttenuation);
	rd1->fGNG.SetGamma(params.blGamma);

	rd1->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd1->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd1->fGNG.AddExtension(new TGNGActivityExt());

	rd1->fGNG.Init(new TGNGVectorData(params.inputDim*4), 
		           new TGNGVectorData(params.inputDim*4), distMeasure);

	rd2->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);	
	rd2->fGNG.SetLambda(params.blLambda);
	rd2->fGNG.SetT_max(params.blTmax);
	rd2->fGNG.SetLambdaDelay(0);
	rd2->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd2->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd2->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd2->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd2->fGNG.SetRampTime(params.blRamp);
	rd2->SetNbAttenuation(params.nbAttenuation);
	rd2->fGNG.SetGamma(params.blGamma);

	rd2->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd2->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd2->fGNG.AddExtension(new TGNGActivityExt());

	rd2->fGNG.Init(new TGNGVectorData(params.inputDim*4), 
		           new TGNGVectorData(params.inputDim*4), distMeasure);

	topGNG->Init(rd1,rd2,distMeasure);

	topGNG->SetMaxUnitCnt(params.topLevelNodes);	
	topGNG->SetLambda(params.tlLambda);
	topGNG->SetT_max(params.tlTmax);
	topGNG->SetEpsilon_b(params.tlPrimaryRate);
	topGNG->SetEpsilon_be(params.tlPrimaryRateEnd);
	topGNG->SetEpsilon_n(params.tlSecondaryRate);
	topGNG->SetEpsilon_ne(params.tlSecondaryRateEnd);
	topGNG->SetRampTime(params.tlRamp);
	topGNG->SetGamma(params.tlGamma);

	topGNG->SetLearningInhibition(params.tlUseLearnInhib);
	topGNG->SetEpsilonRamp(params.tlUseRamp);

	feedRandomEye(sampleCnt, topGNG, params.inputDim,noiseLevel);

	delete topGNG;
	delete distMeasure;
}

/*
	This function sets up an RGNG according to the "params" configuration, and
	uses the "feedRandomRot()" function to feed "sampleCnt" inputs to the RGNG. 
	The	whole process is logged using a TGNGFullLoggerExt as well as a 
	TGNGRecLogExt into the file "logfile". The additional parameter "angle"
	can be used to rotate the input space coordinate system in the input signal 
	(see feedRandomRot() ).
*/
void LongTestRot(LongTestParameters &params, const char* logfile, Int_t sampleCnt, Double_t angle)
{
	TGNG *topGNG = new TGNG("topgng","top layer GNG");

	TGNGFullLoggerExt *logger = new TGNGFullLoggerExt(logfile);
	logger->SetGNGLogInterval(params.logInterval);
	topGNG->AddExtension(logger);
	topGNG->AddExtension(new TGNGRecLogExt(logger->GetCurrentLogFile()));

	TGNGMinkowskiDistance *distMeasure = new TGNGMinkowskiDistance(params.minkowskiParam);
	distMeasure->SetRecursiveMode(params.recMode);

	TGNGRecursiveData *rd1 = new TGNGRecursiveData();
	TGNGRecursiveData *rd2 = new TGNGRecursiveData();

	rd1->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);
	rd1->fGNG.SetLambda(params.blLambda);
	rd1->fGNG.SetT_max(params.blTmax);
	rd1->fGNG.SetLambdaDelay(0);
	rd1->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd1->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd1->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd1->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd1->fGNG.SetRampTime(params.blRamp);
	rd1->SetNbAttenuation(params.nbAttenuation);
	rd1->fGNG.SetGamma(params.blGamma);

	rd1->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd1->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd1->fGNG.AddExtension(new TGNGActivityExt());

	rd1->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	rd2->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);	
	rd2->fGNG.SetLambda(params.blLambda);
	rd2->fGNG.SetT_max(params.blTmax);
	rd2->fGNG.SetLambdaDelay(0);
	rd2->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd2->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd2->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd2->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd2->fGNG.SetRampTime(params.blRamp);
	rd2->SetNbAttenuation(params.nbAttenuation);
	rd2->fGNG.SetGamma(params.blGamma);

	rd2->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd2->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd2->fGNG.AddExtension(new TGNGActivityExt());

	rd2->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	//topGNG->AddExtension(new TGNGLishExt());

	topGNG->Init(rd1,rd2,distMeasure);

	topGNG->SetMaxUnitCnt(params.topLevelNodes);	
	topGNG->SetLambda(params.tlLambda);
	topGNG->SetT_max(params.tlTmax);
	topGNG->SetEpsilon_b(params.tlPrimaryRate);
	topGNG->SetEpsilon_be(params.tlPrimaryRateEnd);
	topGNG->SetEpsilon_n(params.tlSecondaryRate);
	topGNG->SetEpsilon_ne(params.tlSecondaryRateEnd);
	topGNG->SetRampTime(params.tlRamp);
	topGNG->SetGamma(params.tlGamma);

	topGNG->SetLearningInhibition(params.tlUseLearnInhib);
	topGNG->SetEpsilonRamp(params.tlUseRamp);

	feedRandomRot(sampleCnt, topGNG, params.inputDim,params.inpFade,params.inpResample,angle);

	delete topGNG;
	delete distMeasure;
}

/*
	This function sets up an RGNG according to the "params" configuration, and
	uses the "feedRandomPlace()" function to feed "sampleCnt" inputs to the 
	RGNG. The whole process is logged using a TGNGFullLoggerExt as well as a 
	TGNGRecLogExt into the file "logfile". The additional parameter "places"
	allows to specify the place cells that are used as input space 
	(see feedRandomPlace() ).
*/
void LongTestPlace(LongTestParameters &params, const char* logfile, Int_t sampleCnt, std::vector<PlaceParameters*> *places)
{
	TGNG *topGNG = new TGNG("topgng","top layer GNG");

	TGNGFullLoggerExt *logger = new TGNGFullLoggerExt(logfile);
	logger->SetGNGLogInterval(params.logInterval);
	topGNG->AddExtension(logger);
	topGNG->AddExtension(new TGNGRecLogExt(logger->GetCurrentLogFile()));

	TGNGMinkowskiDistance *distMeasure = new TGNGMinkowskiDistance(params.minkowskiParam);
	distMeasure->SetRecursiveMode(params.recMode);

	TGNGRecursiveData *rd1 = new TGNGRecursiveData();
	TGNGRecursiveData *rd2 = new TGNGRecursiveData();

	rd1->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);
	rd1->fGNG.SetLambda(params.blLambda);
	rd1->fGNG.SetT_max(params.blTmax);
	rd1->fGNG.SetLambdaDelay(0);
	rd1->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd1->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd1->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd1->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd1->fGNG.SetRampTime(params.blRamp);
	rd1->SetNbAttenuation(params.nbAttenuation);
	rd1->fGNG.SetGamma(params.blGamma);

	rd1->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd1->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd1->fGNG.AddExtension(new TGNGActivityExt());

	Int_t inpdim = 0;

	for (PlaceParameters *pp : *places)
		inpdim += pp->numberOfCells;
	
	rd1->fGNG.Init(new TGNGVectorData(inpdim), 
		           new TGNGVectorData(inpdim), distMeasure);

	rd2->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);	
	rd2->fGNG.SetLambda(params.blLambda);
	rd2->fGNG.SetT_max(params.blTmax);
	rd2->fGNG.SetLambdaDelay(0);
	rd2->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd2->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd2->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd2->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd2->fGNG.SetRampTime(params.blRamp);
	rd2->SetNbAttenuation(params.nbAttenuation);
	rd2->fGNG.SetGamma(params.blGamma);

	rd2->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd2->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd2->fGNG.AddExtension(new TGNGActivityExt());

	rd2->fGNG.Init(new TGNGVectorData(inpdim), 
		           new TGNGVectorData(inpdim), distMeasure);

	topGNG->Init(rd1,rd2,distMeasure);

	topGNG->SetMaxUnitCnt(params.topLevelNodes);	
	topGNG->SetLambda(params.tlLambda);
	topGNG->SetT_max(params.tlTmax);
	topGNG->SetEpsilon_b(params.tlPrimaryRate);
	topGNG->SetEpsilon_be(params.tlPrimaryRateEnd);
	topGNG->SetEpsilon_n(params.tlSecondaryRate);
	topGNG->SetEpsilon_ne(params.tlSecondaryRateEnd);
	topGNG->SetRampTime(params.tlRamp);
	topGNG->SetGamma(params.tlGamma);

	topGNG->SetLearningInhibition(params.tlUseLearnInhib);
	topGNG->SetEpsilonRamp(params.tlUseRamp);

	feedRandomPlace(sampleCnt, topGNG, places);

	delete topGNG;
	delete distMeasure;
}


/*
	This function sets up an RGNG according to the "params" configuration, and
	uses the "feedPosTree()" function to feed posTree->GetEntries() inputs 
	to the RGNG. The whole process is logged using a TGNGFullLoggerExt as well 
	as a TGNGRecLogExt into the file "logfile". The additional parameter 
	"posTree" allows to specify which locations are fed into the GNG (see 
	feedPosTree() ). The parameters "scale" and "shift" allow to map the 
	locations in "posTree" on the range between 0 and 1 (again, see 
	feedPosTree() ). The parameter "repeat" specifies how often the locations 
	provided in "posTree" should be fed into the gng.
*/
void LongTestTree(LongTestParameters &params, const char* logfile, TTree *posTree,Double_t scale, Double_t shift, Int_t repeat = 1)
{
	TGNG *topGNG = new TGNG("topgng","top layer GNG");

	TGNGFullLoggerExt *logger = new TGNGFullLoggerExt(logfile);
	logger->SetGNGLogInterval(params.logInterval);
	topGNG->AddExtension(logger);
	topGNG->AddExtension(new TGNGRecLogExt(logger->GetCurrentLogFile()));

	TGNGMinkowskiDistance *distMeasure = new TGNGMinkowskiDistance(params.minkowskiParam);
	distMeasure->SetRecursiveMode(params.recMode);

	TGNGRecursiveData *rd1 = new TGNGRecursiveData();
	TGNGRecursiveData *rd2 = new TGNGRecursiveData();

	rd1->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);
	rd1->fGNG.SetLambda(params.blLambda);
	rd1->fGNG.SetT_max(params.blTmax);
	rd1->fGNG.SetLambdaDelay(0);
	rd1->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd1->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd1->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd1->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd1->fGNG.SetRampTime(params.blRamp);
	rd1->SetNbAttenuation(params.nbAttenuation);
	rd1->fGNG.SetGamma(params.blGamma);

	rd1->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd1->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd1->fGNG.AddExtension(new TGNGActivityExt());

	rd1->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	rd2->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);	
	rd2->fGNG.SetLambda(params.blLambda);
	rd2->fGNG.SetT_max(params.blTmax);
	rd2->fGNG.SetLambdaDelay(0);
	rd2->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd2->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd2->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd2->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd2->fGNG.SetRampTime(params.blRamp);
	rd2->SetNbAttenuation(params.nbAttenuation);
	rd2->fGNG.SetGamma(params.blGamma);

	rd2->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd2->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd2->fGNG.AddExtension(new TGNGActivityExt());

	rd2->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	topGNG->Init(rd1,rd2,distMeasure);

	topGNG->SetMaxUnitCnt(params.topLevelNodes);	
	topGNG->SetLambda(params.tlLambda);
	topGNG->SetT_max(params.tlTmax);
	topGNG->SetEpsilon_b(params.tlPrimaryRate);
	topGNG->SetEpsilon_be(params.tlPrimaryRateEnd);
	topGNG->SetEpsilon_n(params.tlSecondaryRate);
	topGNG->SetEpsilon_ne(params.tlSecondaryRateEnd);
	topGNG->SetRampTime(params.tlRamp);
	topGNG->SetGamma(params.tlGamma);

	topGNG->SetLearningInhibition(params.tlUseLearnInhib);
	topGNG->SetEpsilonRamp(params.tlUseRamp);

	for (Int_t i = 0; i < repeat; ++i) {
		std::printf("Run %d of %d\n", i+1,repeat);
		feedPosTree(posTree, topGNG, params.inputDim,params.inpFade, scale, shift, params.inpResample);
	}

	delete topGNG;
	delete distMeasure;
}

/*
	This function sets up an RGNG according to the "params" configuration, and
	uses the "feedPosTree()" function to feed posTree->GetEntries() inputs 
	to the RGNG. The whole process is logged using a TGNGFullLoggerExt as well 
	as a TGNGRecLogExt into the file "logfile". The additional parameter 
	"posTree" allows to specify which locations are fed into the GNG (see 
	feedPosTree() ). The parameters "scale" and "shift" allow to map the 
	locations in "posTree" on the range between 0 and 1 (again, see 
	feedPosTree() ). The parameter "repeatbf" specifies how often the 
	locations provided in "posTree" should be fed into the gng with angle = 0.
	After this, the gng is fed with one iteration of the posTree-Values using
	the angle "angle". Finally, the parameter "repeataf" determines, how often
	the posTree-Values are fed into the gng again with angle = 0.
*/
void LongTestTreeRot(LongTestParameters &params, const char* logfile, TTree *posTree,Double_t scale, 
	Double_t shift, Int_t repeatbf = 1, Int_t repeataf = 1, Double_t angle = 0.0)
{
	TGNG *topGNG = new TGNG("topgng","top layer GNG");

	TGNGFullLoggerExt *logger = new TGNGFullLoggerExt(logfile);
	logger->SetGNGLogInterval(params.logInterval);
	topGNG->AddExtension(logger);
	topGNG->AddExtension(new TGNGRecLogExt(logger->GetCurrentLogFile()));

	TGNGMinkowskiDistance *distMeasure = new TGNGMinkowskiDistance(params.minkowskiParam);
	distMeasure->SetRecursiveMode(params.recMode);

	TGNGRecursiveData *rd1 = new TGNGRecursiveData();
	TGNGRecursiveData *rd2 = new TGNGRecursiveData();

	rd1->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);
	rd1->fGNG.SetLambda(params.blLambda);
	rd1->fGNG.SetT_max(params.blTmax);
	rd1->fGNG.SetLambdaDelay(0);
	rd1->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd1->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd1->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd1->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd1->fGNG.SetRampTime(params.blRamp);
	rd1->SetNbAttenuation(params.nbAttenuation);
	rd1->fGNG.SetGamma(params.blGamma);

	rd1->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd1->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd1->fGNG.AddExtension(new TGNGActivityExt());

	rd1->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	rd2->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);	
	rd2->fGNG.SetLambda(params.blLambda);
	rd2->fGNG.SetT_max(params.blTmax);
	rd2->fGNG.SetLambdaDelay(0);
	rd2->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd2->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd2->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd2->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd2->fGNG.SetRampTime(params.blRamp);
	rd2->SetNbAttenuation(params.nbAttenuation);
	rd2->fGNG.SetGamma(params.blGamma);

	rd2->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd2->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd2->fGNG.AddExtension(new TGNGActivityExt());

	rd2->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	topGNG->Init(rd1,rd2,distMeasure);

	topGNG->SetMaxUnitCnt(params.topLevelNodes);	
	topGNG->SetLambda(params.tlLambda);
	topGNG->SetT_max(params.tlTmax);
	topGNG->SetEpsilon_b(params.tlPrimaryRate);
	topGNG->SetEpsilon_be(params.tlPrimaryRateEnd);
	topGNG->SetEpsilon_n(params.tlSecondaryRate);
	topGNG->SetEpsilon_ne(params.tlSecondaryRateEnd);
	topGNG->SetRampTime(params.tlRamp);
	topGNG->SetGamma(params.tlGamma);

	topGNG->SetLearningInhibition(params.tlUseLearnInhib);
	topGNG->SetEpsilonRamp(params.tlUseRamp);

	for (Int_t i = 0; i < repeatbf; ++i) {
		std::printf("Before run %d of %d\n", i+1,repeatbf);
		feedPosTree(posTree, topGNG, params.inputDim,params.inpFade, scale, shift, params.inpResample);
	}
	std::printf("Rotate run with angle %f\n",angle);
	feedPosTree(posTree, topGNG, params.inputDim,params.inpFade, scale, shift, params.inpResample, angle);
	for (Int_t i = 0; i < repeataf; ++i) {
		std::printf("After run %d of %d\n", i+1,repeataf);
		feedPosTree(posTree, topGNG, params.inputDim,params.inpFade, scale, shift, params.inpResample);
	}

	delete topGNG;
	delete distMeasure;
}

/*
	This function sets up an RGNG according to the "params" configuration, and
	uses the "feedPosTree()" function to feed posTree->GetEntries() inputs 
	to the RGNG. The whole process is logged using a TGNGFullLoggerExt as well 
	as a TGNGRecLogExt into the file "logfile". The additional parameter 
	"posTree" allows to specify which locations are fed into the GNG (see 
	feedPosTree() ). The parameters "scale" and "shift" allow to map the 
	locations in "posTree" on the range between 0 and 1 (again, see 
	feedPosTree() ). The parameter "repeatbf" specifies how often the 
	locations provided in "posTree" should be fed into the gng with yScale = 1.
	After this, the gng is fed with one iteration of the posTree-Values using
	the yScale "rescale". Finally, the parameter "repeataf" determines, how 
	often the posTree-Values are fed into the gng again with yScale = 1.
*/
void LongTestTreeScale(LongTestParameters &params, const char* logfile, TTree *posTree,Double_t scale, 
	Double_t shift, Int_t repeatbf = 1, Int_t repeataf = 1, Double_t rescale = 1.0)
{
	TGNG *topGNG = new TGNG("topgng","top layer GNG");

	TGNGFullLoggerExt *logger = new TGNGFullLoggerExt(logfile);
	logger->SetGNGLogInterval(params.logInterval);
	topGNG->AddExtension(logger);
	topGNG->AddExtension(new TGNGRecLogExt(logger->GetCurrentLogFile()));

	TGNGMinkowskiDistance *distMeasure = new TGNGMinkowskiDistance(params.minkowskiParam);
	distMeasure->SetRecursiveMode(params.recMode);

	TGNGRecursiveData *rd1 = new TGNGRecursiveData();
	TGNGRecursiveData *rd2 = new TGNGRecursiveData();

	rd1->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);
	rd1->fGNG.SetLambda(params.blLambda);
	rd1->fGNG.SetT_max(params.blTmax);
	rd1->fGNG.SetLambdaDelay(0);
	rd1->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd1->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd1->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd1->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd1->fGNG.SetRampTime(params.blRamp);
	rd1->SetNbAttenuation(params.nbAttenuation);
	rd1->fGNG.SetGamma(params.blGamma);

	rd1->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd1->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd1->fGNG.AddExtension(new TGNGActivityExt());

	rd1->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	rd2->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);	
	rd2->fGNG.SetLambda(params.blLambda);
	rd2->fGNG.SetT_max(params.blTmax);
	rd2->fGNG.SetLambdaDelay(0);
	rd2->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd2->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd2->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd2->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd2->fGNG.SetRampTime(params.blRamp);
	rd2->SetNbAttenuation(params.nbAttenuation);
	rd2->fGNG.SetGamma(params.blGamma);

	rd2->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd2->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd2->fGNG.AddExtension(new TGNGActivityExt());

	rd2->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	topGNG->Init(rd1,rd2,distMeasure);

	topGNG->SetMaxUnitCnt(params.topLevelNodes);	
	topGNG->SetLambda(params.tlLambda);
	topGNG->SetT_max(params.tlTmax);
	topGNG->SetEpsilon_b(params.tlPrimaryRate);
	topGNG->SetEpsilon_be(params.tlPrimaryRateEnd);
	topGNG->SetEpsilon_n(params.tlSecondaryRate);
	topGNG->SetEpsilon_ne(params.tlSecondaryRateEnd);
	topGNG->SetRampTime(params.tlRamp);
	topGNG->SetGamma(params.tlGamma);

	topGNG->SetLearningInhibition(params.tlUseLearnInhib);
	topGNG->SetEpsilonRamp(params.tlUseRamp);

	for (Int_t i = 0; i < repeatbf; ++i) {
		std::printf("Before run %d of %d\n", i+1,repeatbf);
		feedPosTree(posTree, topGNG, params.inputDim,params.inpFade, scale, shift, params.inpResample);
	}
	std::printf("Rescale run with scale %f\n",rescale);
	feedPosTree(posTree, topGNG, params.inputDim,params.inpFade, scale, shift, params.inpResample, 0.0,rescale);
	for (Int_t i = 0; i < repeataf; ++i) {
		std::printf("After run %d of %d\n", i+1,repeataf);
		feedPosTree(posTree, topGNG, params.inputDim,params.inpFade, scale, shift, params.inpResample);
	}

	delete topGNG;
	delete distMeasure;
}

/*
	This function sets up an RGNG according to the "params" configuration, and
	uses the "feedRandom()" function and the "feedPosTree()" function to feed 
	"delay" + posTree->GetEntries() inputs to the RGNG. The whole process is 
	logged using a TGNGFullLoggerExt as well as a TGNGRecLogExt into the file 
	"logfile". The additional parameter "posTree" allows to specify which 
	locations are fed into the GNG (see feedPosTree() ). The parameters "scale" 
	and "shift" allow to map the locations in "posTree" on the range between 0 
	and 1 (again, see feedPosTree() ). The parameter "delay" specifies how many
	random positions should be fed to the gng via feedRandom() before the 
	posTree-Values are used.
*/
void LongTestDelayedTree(LongTestParameters &params, const char* logfile, TTree *posTree,Double_t scale, Double_t shift, Int_t delay)
{
	TGNG *topGNG = new TGNG("topgng","top layer GNG");

	TGNGFullLoggerExt *logger = new TGNGFullLoggerExt(logfile);
	logger->SetGNGLogInterval(params.logInterval);
	topGNG->AddExtension(logger);
	topGNG->AddExtension(new TGNGRecLogExt(logger->GetCurrentLogFile()));

	TGNGMinkowskiDistance *distMeasure = new TGNGMinkowskiDistance(params.minkowskiParam);
	distMeasure->SetRecursiveMode(params.recMode);

	TGNGRecursiveData *rd1 = new TGNGRecursiveData();
	TGNGRecursiveData *rd2 = new TGNGRecursiveData();

	rd1->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);
	rd1->fGNG.SetLambda(params.blLambda);
	rd1->fGNG.SetT_max(params.blTmax);
	rd1->fGNG.SetLambdaDelay(0);
	rd1->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd1->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd1->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd1->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd1->fGNG.SetRampTime(params.blRamp);
	rd1->SetNbAttenuation(params.nbAttenuation);
	rd1->fGNG.SetGamma(params.blGamma);

	rd1->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd1->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd1->fGNG.AddExtension(new TGNGActivityExt());

	rd1->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	rd2->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);	
	rd2->fGNG.SetLambda(params.blLambda);
	rd2->fGNG.SetT_max(params.blTmax);
	rd2->fGNG.SetLambdaDelay(0);
	rd2->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd2->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd2->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd2->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd2->fGNG.SetRampTime(params.blRamp);
	rd2->SetNbAttenuation(params.nbAttenuation);
	rd2->fGNG.SetGamma(params.blGamma);

	rd2->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd2->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd2->fGNG.AddExtension(new TGNGActivityExt());

	rd2->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	topGNG->Init(rd1,rd2,distMeasure);

	topGNG->SetMaxUnitCnt(params.topLevelNodes);	
	topGNG->SetLambda(params.tlLambda);
	topGNG->SetT_max(params.tlTmax);
	topGNG->SetEpsilon_b(params.tlPrimaryRate);
	topGNG->SetEpsilon_be(params.tlPrimaryRateEnd);
	topGNG->SetEpsilon_n(params.tlSecondaryRate);
	topGNG->SetEpsilon_ne(params.tlSecondaryRateEnd);
	topGNG->SetRampTime(params.tlRamp);
	topGNG->SetGamma(params.tlGamma);

	topGNG->SetLearningInhibition(params.tlUseLearnInhib);
	topGNG->SetEpsilonRamp(params.tlUseRamp);

	feedRandom(delay, topGNG, params.inputDim,params.inpFade,params.inpResample);
	feedPosTree(posTree, topGNG, params.inputDim,params.inpFade, scale, shift, params.inpResample);

	delete topGNG;
	delete distMeasure;
}

/*
	This function sets up an RGNG according to the "params" configuration, and
	uses the "feedPosTreeNoise()" function to feed posTree->GetEntries() inputs 
	to the RGNG. The whole process is logged using a TGNGFullLoggerExt as well 
	as a TGNGRecLogExt into the file "logfile". The additional parameter 
	"posTree" allows to specify which locations are fed into the GNG (see 
	feedPosTreeNoise() ). The parameters "scale" and "shift" allow to map the 
	locations in "posTree" on the range between 0 and 1 (again, see 
	feedPosTreeNoise() ). The parameter "repeat" specifies how often the 
	locations provided in "posTree" should be fed into the gng. The parameter
	"noiseLevel" allows to specify a level of noise that should be added to
	the input vectors (again, see feedPosTreeNoise() ).
*/
void LongTestTreeNoise(LongTestParameters &params, const char* logfile, TTree *posTree,Double_t scale, Double_t shift, Double_t noiseLevel, Int_t repeat = 1)
{
	TGNG *topGNG = new TGNG("topgng","top layer GNG");

	TGNGFullLoggerExt *logger = new TGNGFullLoggerExt(logfile);
	logger->SetGNGLogInterval(params.logInterval);
	topGNG->AddExtension(logger);
	topGNG->AddExtension(new TGNGRecLogExt(logger->GetCurrentLogFile()));

	TGNGMinkowskiDistance *distMeasure = new TGNGMinkowskiDistance(params.minkowskiParam);
	distMeasure->SetRecursiveMode(params.recMode);

	TGNGRecursiveData *rd1 = new TGNGRecursiveData();
	TGNGRecursiveData *rd2 = new TGNGRecursiveData();

	rd1->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);
	rd1->fGNG.SetLambda(params.blLambda);
	rd1->fGNG.SetT_max(params.blTmax);
	rd1->fGNG.SetLambdaDelay(0);
	rd1->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd1->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd1->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd1->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd1->fGNG.SetRampTime(params.blRamp);
	rd1->SetNbAttenuation(params.nbAttenuation);
	rd1->fGNG.SetGamma(params.blGamma);

	rd1->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd1->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd1->fGNG.AddExtension(new TGNGActivityExt());

	rd1->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	rd2->fGNG.SetMaxUnitCnt(params.bottomLevelNodes);	
	rd2->fGNG.SetLambda(params.blLambda);
	rd2->fGNG.SetT_max(params.blTmax);
	rd2->fGNG.SetLambdaDelay(0);
	rd2->fGNG.SetEpsilon_b(params.blPrimaryRate);
	rd2->fGNG.SetEpsilon_be(params.blPrimaryRateEnd);
	rd2->fGNG.SetEpsilon_n(params.blSecondaryRate);
	rd2->fGNG.SetEpsilon_ne(params.blSecondaryRateEnd);
	rd2->fGNG.SetRampTime(params.blRamp);
	rd2->SetNbAttenuation(params.nbAttenuation);
	rd2->fGNG.SetGamma(params.blGamma);

	rd2->fGNG.SetLearningInhibition(params.blUseLearnInhib);
	rd2->fGNG.SetEpsilonRamp(params.blUseRamp);

	rd2->fGNG.AddExtension(new TGNGActivityExt());

	rd2->fGNG.Init(new TGNGVectorData(params.inputDim), 
		           new TGNGVectorData(params.inputDim), distMeasure);

	topGNG->Init(rd1,rd2,distMeasure);

	topGNG->SetMaxUnitCnt(params.topLevelNodes);	
	topGNG->SetLambda(params.tlLambda);
	topGNG->SetT_max(params.tlTmax);
	topGNG->SetEpsilon_b(params.tlPrimaryRate);
	topGNG->SetEpsilon_be(params.tlPrimaryRateEnd);
	topGNG->SetEpsilon_n(params.tlSecondaryRate);
	topGNG->SetEpsilon_ne(params.tlSecondaryRateEnd);
	topGNG->SetRampTime(params.tlRamp);
	topGNG->SetGamma(params.tlGamma);

	topGNG->SetLearningInhibition(params.tlUseLearnInhib);
	topGNG->SetEpsilonRamp(params.tlUseRamp);

	for (Int_t i = 0; i < repeat; ++i) {
		std::printf("Run %d of %d\n", i+1,repeat);
		feedPosTreeNoise(posTree, topGNG, params.inputDim,params.inpFade, scale, shift, params.inpResample,noiseLevel);
	}

	delete topGNG;
	delete distMeasure;
}
