/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 
         International, see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.
--------------------------------------------------------------------------------

This macro contains functions to analyse RGNG grid cell simulations.

Common usage in the ROOT command line would be, e.g.:

root [0] .x tgng_load.C
root [1] .L macros/rgng_analysis.C+
root [2] AnalyseFile("inputfile.root", "outputfile.root", nrOfSteps, 
				      shift, scale, mapRes, samplDur, mxthres, mxnb);

To analyse a whole directiory of simulation data use:

AnalyseDir("directory","inputfile prefix","output file prefix", nrOfSteps, 
				      shift, scale, mapRes, samplDur, mxthres, mxnb);

The nrOfSteps defines the intervals at which the analysis of the input data
should take place. For example, if your input has 3 million time steps and
your nrOfSteps is 30, then you get an analysis of the input data every 100000 
time steps. The parameters shift and scale allow to analyse only part of each
TL unit's firing fields. For an analysis that includes all firing fields the
values are shift = 0, and scale = 1.0. If you want to look at the center quarter
of all firing fields, the value would be shift = -0.25, scale = 2.0. The 
parameter mapRes describes the firing rate map resolution. If you assume a
standard 150cm x 150cm environment that is measured with a 3cm resolution the
mapRes should be 50. The samplDur describes the number of inputs over which
each rate map should be integrated. The default value is 30000 samples which
corresponds to a 10 minute trial sampled at 50Hz. If you look at only a fraction
of the firing fields, it could be necessary to increase the sampling duration
to, e.g., 60000 samples. Please NOTE: The sampling duration should fit into the
analysis interval you specify by the nrOfSteps. The parameter mxthres defines
which value should be used as the lower bound threshold for the maxima detection
in the rate map autocorrelograms. The default value of 0.2 is the one that is
used throughout the grid cell literature. The parameter mxnb describes the
neighborhood size in which the search for local maxima should take place. The
default value is 2 which amounts to a neighborhood size of 5x5 (-2 to +2).

*/

#include <vector>
#include <utility>
#include <queue>

#include <TTree.h>
#include <TObjArray.h>
#include <TH2D.h>
#include <TH3D.h>
#include <TMatrixT.h>
#include <TMath.h>
#include <TVector2.h>
#include <TFile.h>
#include <TSystemFile.h>
#include <TSystem.h>
#include <TSystemDirectory.h>
#include <TList.h>
#include <TRandom.h>

#include "src/core/TGNG.h"
#include "src/core/TGNGUnit.h"
#include "src/core/TNG.h"
#include "src/core/TNGUnit.h"
#include "src/core/TGNGP.h"
#include "src/distmeasures/TGNGMinkowskiDistance.h"
#include "src/datatypes/TGNGVectorData.h"
#include "src/datatypes/TGNGRecursiveData.h"


// utility functions to analyse recurse gngs

TTree* LoadGNGTree(TFile *file) 
{
	return (TTree*)file->Get("gngTree_0");
}

TH2D* GetGradDev(TTree *gngTree, Int_t steps, ULong64_t maxTime = 0, UInt_t maxVal = 0)
{
	ULong64_t  time;
	TGNG      *gng = 0;

	gngTree->SetBranchAddress("inputCnt",&time);
	gngTree->SetBranchAddress("gng",&gng);

	gngTree->GetEntry(0);

	UInt_t mxUnitCnt = gng->GetMaxUnitCnt();

	if ((maxVal != 0) && (maxVal < mxUnitCnt)) {
		mxUnitCnt = maxVal;
	}

	Int_t nrOfEntries = gngTree->GetEntries();

	if (maxTime != 0) {
		gngTree->SetBranchStatus("*",0); //disable all branches
	    gngTree->SetBranchStatus("inputCnt",1);

	    for (Int_t i = 0; i < nrOfEntries; ++i) {
	    	gngTree->GetEntry(i);
	    	if (time > maxTime) {
	    		nrOfEntries = i;
	    		break;
	    	}
	    }

	    gngTree->SetBranchStatus("*",1); //enable all branches
	}

	Int_t stepWidth = nrOfEntries / steps;

	if (stepWidth == 0) {
		stepWidth = 1;
		steps = nrOfEntries;
	}

	delete gng;
	gng = 0;

	TH2D *result = new TH2D("candleHist","distribution of grads",steps,0,(Double_t)steps,(Int_t)mxUnitCnt,0,(Double_t)mxUnitCnt);
	result->SetDirectory(0);

	for (Int_t i = 0; i < steps; ++i) {
		gngTree->GetEntry(i * stepWidth);

		gng->FillGradDistribution(result,(Double_t)i + 0.5);
		result->GetXaxis()->SetBinLabel(i+1,Form("%llu",time));

		delete gng;
		gng = 0;
	}

	return result;
}

TH2D* GetDistDev(TTree *gngTree, Int_t steps, Double_t minVal, Double_t maxVal, ULong64_t maxTime = 0, TGNGMinkowskiDistance::LinkageType lt = TGNGMinkowskiDistance::ltSingle)
{
	ULong64_t  time;
	TGNG      *gng = 0;

	gngTree->SetBranchAddress("inputCnt",&time);
	gngTree->SetBranchAddress("gng",&gng);

	Int_t nrOfEntries = gngTree->GetEntries();

	if (maxTime != 0) {
		gngTree->SetBranchStatus("*",0); //disable all branches
	    gngTree->SetBranchStatus("inputCnt",1);

	    for (Int_t i = 0; i < nrOfEntries; ++i) {
	    	gngTree->GetEntry(i);
	    	if (time > maxTime) {
	    		nrOfEntries = i;
	    		break;
	    	}
	    }

	    gngTree->SetBranchStatus("*",1); //enable all branches
	}

	Int_t stepWidth = nrOfEntries / steps;

	if (stepWidth == 0) {
		stepWidth = 1;
		steps = nrOfEntries;
	}

	TH2D *result = new TH2D("candleHist","distribution of distances",steps,0,(Double_t)steps,100,minVal,maxVal);
	result->SetDirectory(0);

	for (Int_t i = 0; i < steps; ++i) {
		gngTree->GetEntry(i * stepWidth);

		TGNGMinkowskiDistance *dm = static_cast<TGNGMinkowskiDistance*>(gng->GetDistanceMeasure());
		dm->SetLinkageType(lt);

		gng->FillDistanceDistribution(result,(Double_t)i + 0.5);
		result->GetXaxis()->SetBinLabel(i+1,Form("%llu",time));

		delete gng;
		gng = 0;
	}

	return result;
}


TH2D* GetNbDistDev(TTree *gngTree, Int_t steps, Double_t minVal, Double_t maxVal, ULong64_t maxTime = 0, TGNGMinkowskiDistance::LinkageType lt = TGNGMinkowskiDistance::ltSingle)
{
	ULong64_t  time;
	TGNG      *gng = 0;

	gngTree->SetBranchAddress("inputCnt",&time);
	gngTree->SetBranchAddress("gng",&gng);

	Int_t nrOfEntries = gngTree->GetEntries();

	if (maxTime != 0) {
		gngTree->SetBranchStatus("*",0); //disable all branches
	    gngTree->SetBranchStatus("inputCnt",1);

	    for (Int_t i = 0; i < nrOfEntries; ++i) {
	    	gngTree->GetEntry(i);
	    	if (time > maxTime) {
	    		nrOfEntries = i;
	    		break;
	    	}
	    }

	    gngTree->SetBranchStatus("*",1); //enable all branches
	}

	Int_t stepWidth = nrOfEntries / steps;

	if (stepWidth == 0) {
		stepWidth = 1;
		steps = nrOfEntries;
	}

	TH2D *result = new TH2D("candleHist","distribution of distances",steps,0,(Double_t)steps,100,minVal,maxVal);
	result->SetDirectory(0);

	for (Int_t i = 0; i < steps; ++i) {
		gngTree->GetEntry(i * stepWidth);

		TGNGMinkowskiDistance *dm = static_cast<TGNGMinkowskiDistance*>(gng->GetDistanceMeasure());
		dm->SetLinkageType(lt);

		gng->FillNbDistanceDistribution(result,(Double_t)i + 0.5);
		result->GetXaxis()->SetBinLabel(i+1,Form("%llu",time));

		delete gng;
		gng = 0;
	}

	return result;
}

void GetUnitCenters(TGNG *gng, TObjArray *output)
{
	Int_t nrOfUnits = gng->fUnits.GetEntriesFast();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(gng->fUnits.At(i));
		TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);

		Double_t curX = curData->fVector[0];
		Double_t curY = curData->fVector[1];

		Int_t rowCnt = curData->fVector.GetNrows();
		if (rowCnt > 2) {
			// determine x and y as maximum values in lower and upper half
			Double_t mxv = 0;
			Int_t    mxi = 0;
			Double_t myv = 0;
			Int_t    myi = rowCnt / 2;
			for (Int_t r = 0; r < rowCnt / 2; ++r) {
				if (curData->fVector[r] > mxv) {
					mxv = curData->fVector[r];
					mxi = r;
				}
				Int_t ru = r + (rowCnt / 2);
				if (curData->fVector[ru] > myv) {
					myv = curData->fVector[ru];
					myi = ru;
				}					
			}

			Double_t mxv0 = 0;
			if (mxi > 0) {
				mxv0 = curData->fVector[mxi-1];
			} else {
				mxv0 = curData->fVector[(rowCnt / 2)-1];
			}

			Double_t mxv1 = 0;
			if (mxi < (rowCnt / 2)-1) {
				mxv1 = curData->fVector[mxi+1];
			} else {
				mxv1 = curData->fVector[0];
			}

			Double_t dxi = (Double_t)mxi - (mxv1-mxv0) / (2.0 * (mxv0+mxv1)-4.0*mxv);


			Double_t myv0 = 0;
			if (myi > (rowCnt / 2)) {
				myv0 = curData->fVector[myi-1];
			} else {
				myv0 = curData->fVector[rowCnt-1];
			}

			Double_t myv1 = 0;
			if (myi < rowCnt-1) {
				myv1 = curData->fVector[myi+1];
			} else {
				myv1 = curData->fVector[rowCnt / 2];
			}

			Double_t dyi = (Double_t)(myi-(rowCnt / 2)) - (myv1-myv0) / (2.0 * (myv0+myv1)-4.0*myv);

			curX = dxi / (Double_t)(rowCnt / 2);
			curY = dyi / (Double_t)(rowCnt / 2);
		}

		output->Add(new TVector2(curX,curY));

	}
}


void GetUnitCenters(TNG *ng, TObjArray *output)
{
	Int_t nrOfUnits = ng->fUnits.GetEntriesFast();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TNGUnit *curUnit = static_cast<TNGUnit*>(ng->fUnits.At(i));
		TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);

		Double_t curX = curData->fVector[0];
		Double_t curY = curData->fVector[1];

		Int_t rowCnt = curData->fVector.GetNrows();
		if (rowCnt > 2) {
			// determine x and y as maximum values in lower and upper half
			Double_t mxv = 0;
			Int_t    mxi = 0;
			Double_t myv = 0;
			Int_t    myi = rowCnt / 2;
			for (Int_t r = 0; r < rowCnt / 2; ++r) {
				if (curData->fVector[r] > mxv) {
					mxv = curData->fVector[r];
					mxi = r;
				}
				Int_t ru = r + (rowCnt / 2);
				if (curData->fVector[ru] > myv) {
					myv = curData->fVector[ru];
					myi = ru;
				}					
			}

			Double_t mxv0 = 0;
			if (mxi > 0) {
				mxv0 = curData->fVector[mxi-1];
			} else {
				mxv0 = curData->fVector[(rowCnt / 2)-1];
			}

			Double_t mxv1 = 0;
			if (mxi < (rowCnt / 2)-1) {
				mxv1 = curData->fVector[mxi+1];
			} else {
				mxv1 = curData->fVector[0];
			}

			Double_t dxi = (Double_t)mxi - (mxv1-mxv0) / (2.0 * (mxv0+mxv1)-4.0*mxv);


			Double_t myv0 = 0;
			if (myi > (rowCnt / 2)) {
				myv0 = curData->fVector[myi-1];
			} else {
				myv0 = curData->fVector[rowCnt-1];
			}

			Double_t myv1 = 0;
			if (myi < rowCnt-1) {
				myv1 = curData->fVector[myi+1];
			} else {
				myv1 = curData->fVector[rowCnt / 2];
			}

			Double_t dyi = (Double_t)(myi-(rowCnt / 2)) - (myv1-myv0) / (2.0 * (myv0+myv1)-4.0*myv);

			curX = dxi / (Double_t)(rowCnt / 2);
			curY = dyi / (Double_t)(rowCnt / 2);
		}

		output->Add(new TVector2(curX,curY));

	}
}

void GetUnitCenters(TGNGP *gngp, Int_t tl, TObjArray *output)
{
	TObjArray *topLayerUnit = static_cast<TObjArray*>(gngp->fUnits.At(tl));

	Int_t nrOfUnits = topLayerUnit->GetEntriesFast();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(i));
		TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);

		Double_t curX = curData->fVector[0];
		Double_t curY = curData->fVector[1];

		Int_t rowCnt = curData->fVector.GetNrows();
		if (rowCnt > 2) {
			// determine x and y as maximum values in lower and upper half
			Double_t mxv = 0;
			Int_t    mxi = 0;
			Double_t myv = 0;
			Int_t    myi = rowCnt / 2;
			for (Int_t r = 0; r < rowCnt / 2; ++r) {
				if (curData->fVector[r] > mxv) {
					mxv = curData->fVector[r];
					mxi = r;
				}
				Int_t ru = r + (rowCnt / 2);
				if (curData->fVector[ru] > myv) {
					myv = curData->fVector[ru];
					myi = ru;
				}					
			}

			Double_t mxv0 = 0;
			if (mxi > 0) {
				mxv0 = curData->fVector[mxi-1];
			} else {
				mxv0 = curData->fVector[(rowCnt / 2)-1];
			}

			Double_t mxv1 = 0;
			if (mxi < (rowCnt / 2)-1) {
				mxv1 = curData->fVector[mxi+1];
			} else {
				mxv1 = curData->fVector[0];
			}

			Double_t dxi = (Double_t)mxi - (mxv1-mxv0) / (2.0 * (mxv0+mxv1)-4.0*mxv);


			Double_t myv0 = 0;
			if (myi > (rowCnt / 2)) {
				myv0 = curData->fVector[myi-1];
			} else {
				myv0 = curData->fVector[rowCnt-1];
			}

			Double_t myv1 = 0;
			if (myi < rowCnt-1) {
				myv1 = curData->fVector[myi+1];
			} else {
				myv1 = curData->fVector[rowCnt / 2];
			}

			Double_t dyi = (Double_t)(myi-(rowCnt / 2)) - (myv1-myv0) / (2.0 * (myv0+myv1)-4.0*myv);

			curX = dxi / (Double_t)(rowCnt / 2);
			curY = dyi / (Double_t)(rowCnt / 2);
		}

		output->Add(new TVector2(curX,curY));

	}
}


TMatrixD* DrawPoints(TObjArray *points, Int_t rows, Int_t cols, Double_t sigma, Bool_t wrap = kTRUE, Double_t scale = 1.0, Int_t rowShift = 0, Int_t colShift = 0)
{
	TMatrixD *result = new TMatrixD(rows,cols);
	result->Zero();

	Int_t nrOfUnits = points->GetEntriesFast();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TVector2 *curPoint = static_cast<TVector2*>(points->At(i));

		Int_t ix = curPoint->X() * scale * (Double_t)cols + (Double_t)colShift;
		Int_t iy = curPoint->Y() * scale * (Double_t)rows + (Double_t)rowShift;

		// draw gaussian around iy,ix
		Int_t lxw = sigma * (Double_t)cols * 6.0;
		Int_t lyw = sigma * (Double_t)rows * 6.0;

		for (Int_t lx = -lxw; lx <= lxw; ++lx) {
			for (Int_t ly = -lyw; ly <= lyw; ++ly) {
				Int_t cx = ix + lx;
				Int_t cy = iy + ly;

				if (wrap == kTRUE) {
					while (cx < 0) cx += cols;
					while (cx >= cols) cx -= cols;
					while (cy < 0) cy += rows;
					while (cy >= rows) cy -= rows;
				}
				if ((cx < 0) || (cx >= cols) || (cy < 0) || (cy >= rows))
					continue;

				Double_t dx = (Double_t)lx / (Double_t)cols;
				Double_t dy = (Double_t)ly / (Double_t)rows;

				Double_t val = TMath::Exp(-(dx*dx+dy*dy)/(2.0*sigma*sigma));
				if ((*result)(cy,cx) < val)
					(*result)(cy,cx) = val;
			}
		}
	}

	return result;
}


TMatrixD* DrawUnitCenters(TGNG *gng, Int_t rows, Int_t cols, Double_t sigma, Bool_t wrap = kTRUE, Double_t scale = 1.0, Int_t rowShift = 0, Int_t colShift = 0)
{
	TObjArray tmp;
	tmp.SetOwner(kTRUE);
	GetUnitCenters(gng,&tmp);
	return DrawPoints(&tmp,rows,cols,sigma,wrap,scale,rowShift,colShift);
}

TMatrixD* DrawUnitCenters(TNG *ng, Int_t rows, Int_t cols, Double_t sigma, Bool_t wrap = kTRUE, Double_t scale = 1.0, Int_t rowShift = 0, Int_t colShift = 0)
{
	TObjArray tmp;
	tmp.SetOwner(kTRUE);
	GetUnitCenters(ng,&tmp);
	return DrawPoints(&tmp,rows,cols,sigma,wrap,scale,rowShift,colShift);
}

TMatrixD* DrawUnitCenters(TGNGP *gngp, Int_t tl, Int_t rows, Int_t cols, Double_t sigma, Bool_t wrap = kTRUE, Double_t scale = 1.0, Int_t rowShift = 0, Int_t colShift = 0)
{
	TObjArray tmp;
	tmp.SetOwner(kTRUE);
	GetUnitCenters(gngp,tl,&tmp);
	return DrawPoints(&tmp,rows,cols,sigma,wrap,scale,rowShift,colShift);
}

// autocorrelation calculated as Pearson’s product moment correlation coefficient 
// as described in the supporting material to Sargolini et al. 2006

void GetCrossCorrelation(TMatrixD &inputA, TMatrixD &inputB, TMatrixD &output)
{
	Int_t nRows = inputA.GetNrows();
	Int_t nCols = inputA.GetNcols();

	output.ResizeTo(2*nRows+1,2*nCols+1);
	output.Zero();

	Int_t x,y,dx,dy;

	for (dy = -nRows; dy <= nRows; ++dy) {
		//printf("processing row %d\n",dy);
		for (dx = -nCols; dx <= nCols; ++dx) {
			output(dy+nRows,dx+nCols) = 0;
			Double_t inputSum = 0;
			Double_t inputSum2 = 0;
			Double_t shiftSum = 0;
			Double_t shiftSum2 = 0;
			Double_t n = 0;
			Double_t productSum = 0;
			for (y = 0; y < nRows; ++y)
				for (x = 0; x < nCols; ++x) {
					Int_t rx = x+dx;
					Int_t ry = y+dy;
					Double_t value = 0;
					Double_t sv = 0;
					Double_t nv = 0;
					if ((rx >= 0) && (rx < nCols) && (ry >= 0) && (ry < nRows)) {
						sv = inputB(ry,rx);
						nv = inputA(y,x);
						value = nv * sv;
						n += 1.0;
					}
					productSum += value;
					shiftSum  += sv;
					shiftSum2 += sv*sv;
					inputSum  += nv;
					inputSum2 += nv*nv;
				}
			if (n >= 20) {
				Double_t div = TMath::Sqrt(n * inputSum2 - TMath::Power(inputSum,2.0)) * 
					           TMath::Sqrt(n * shiftSum2 - TMath::Power(shiftSum,2.0));
				if (div != 0)
					output(dy+nRows,dx+nCols) = (n * productSum - (inputSum * shiftSum)) / div;
			}
		}		
	}
}

void GetCrossCorrelation2(TMatrixD &inputA, TMatrixD &inputB, TMatrixD &output)
{
	Int_t nRows = inputA.GetNrows();
	Int_t nCols = inputA.GetNcols();

	output.ResizeTo(2*nRows+1,2*nCols+1);
	output.Zero();

	Int_t nCols2 = 2*nCols+1; 

	double *arrayA = inputA.GetMatrixArray();
	double *arrayB = inputB.GetMatrixArray();
	double *arrayO = output.GetMatrixArray();

//fElements[arown*this->fNcols+acoln])

	Int_t x,y,dx,dy;

	for (dy = -nRows; dy <= nRows; ++dy) {
		//printf("processing row %d\n",dy);
		for (dx = -nCols; dx <= nCols; ++dx) {
			//output(dy+nRows,dx+nCols) = 0;
			Double_t inputSum = 0;
			Double_t inputSum2 = 0;
			Double_t shiftSum = 0;
			Double_t shiftSum2 = 0;
			Double_t n = 0;
			Double_t productSum = 0;
			for (y = 0; y < nRows; ++y)
				for (x = 0; x < nCols; ++x) {
					Int_t rx = x+dx;
					Int_t ry = y+dy;
					Double_t value = 0;
					Double_t sv = 0;
					Double_t nv = 0;
					if ((rx >= 0) && (rx < nCols) && (ry >= 0) && (ry < nRows)) {
						//sv = inputB(ry,rx);
						sv = arrayB[ry * nCols + rx];
						//nv = inputA(y,x);
						nv = arrayA[y * nCols + x];
						value = nv * sv;
						n += 1.0;
					}
					productSum += value;
					shiftSum  += sv;
					shiftSum2 += sv*sv;
					inputSum  += nv;
					inputSum2 += nv*nv;
				}
			if (n >= 20) {
				Double_t div = TMath::Sqrt(n * inputSum2 - TMath::Power(inputSum,2.0)) * 
					           TMath::Sqrt(n * shiftSum2 - TMath::Power(shiftSum,2.0));
				if (div != 0) {
					//output(dy+nRows,dx+nCols) = (n * productSum - (inputSum * shiftSum)) / div;
					arrayO[(dy+nRows) * nCols2 + (dx+nCols)] = (n * productSum - (inputSum * shiftSum)) / div;
				}
			}
		}		
	}
}

Double_t GetCorrelation(TMatrixD &inputA, TMatrixD &inputB)
{
	Int_t nRows = inputA.GetNrows();
	Int_t nCols = inputA.GetNcols();

	Int_t x,y;

	Double_t result = 0;

	Double_t inputSum = 0;
	Double_t inputSum2 = 0;
	Double_t shiftSum = 0;
	Double_t shiftSum2 = 0;
	Double_t n = 0;
	Double_t productSum = 0;
	for (y = 0; y < nRows; ++y)
		for (x = 0; x < nCols; ++x) {
			Double_t value = 0;
			Double_t sv = 0;
			Double_t nv = 0;

			sv = inputB(y,x);
			nv = inputA(y,x);
			value = nv * sv;
			n += 1.0;

			productSum += value;
			shiftSum  += sv;
			shiftSum2 += sv*sv;
			inputSum  += nv;
			inputSum2 += nv*nv;
		}

	Double_t div = TMath::Sqrt(n * inputSum2 - TMath::Power(inputSum,2.0)) * 
		           TMath::Sqrt(n * shiftSum2 - TMath::Power(shiftSum,2.0));
	if (div != 0)
		result = (n * productSum - (inputSum * shiftSum)) / div;

	return result;
}

TMatrixD* GetAutoCorrelation(TMatrixD* input)
{
	TMatrixD *result = new TMatrixD();
	//GetCrossCorrelation(*input,*input,*result);
	GetCrossCorrelation2(*input,*input,*result);
	return result;
}

void GetMax(TMatrixD &ac, Int_t nb, TObjArray *maxPoints)
{
	Int_t nCols = ac.GetNcols();
	Int_t nRows = ac.GetNrows();

	Int_t xmid = nCols / 2;
	Int_t ymid = nRows / 2;

	maxPoints->Clear();

	Int_t x,y,dx,dy;
	for (y = nb; y < nRows-nb; ++y)
		for (x = nb; x < nCols-nb; ++x) {
			Double_t mv = ac(y,x);
			for (dy = -nb; dy <= nb; ++dy)
				for (dx = -nb; dx <= nb; ++dx)
					if ((ac(y+dy,x+dx) >= mv) && ((dx != 0) || (dy != 0))) {
						mv = 0;
						break;
					}
			if (mv > 0)
				maxPoints->Add(new TVector2((Double_t)(x-xmid),(Double_t)(y-ymid)));
		}
}

void ReduceTo6(TObjArray *maxPoints)
{
	// search 0,0
	Int_t i;
	Int_t  cnt = maxPoints->GetEntriesFast();
	Int_t vcnt = maxPoints->GetEntriesFast();
	for (i = 0; i < cnt; ++i) {
		TVector2 *vec = (TVector2*)maxPoints->At(i);
		if ((vec) && (vec->X() == 0) && (vec->Y() == 0)) {
			maxPoints->RemoveAt(i);
			delete vec;
			--vcnt;
			break;
		}
	}

	while (vcnt > 6) {
		Double_t  maxLength = 0;
		TVector2 *maxVec = 0;
		Int_t     maxIdx = -1;
		for (i = 0; i < cnt; ++i) {
			TVector2 *vec = (TVector2*)maxPoints->At(i);
			if ((vec) && (vec->Mod() > maxLength)) {
				maxLength = vec->Mod();
				maxVec    = vec;
				maxIdx    = i;
			}
		}
		if (maxVec) {
			maxPoints->RemoveAt(maxIdx);
			delete maxVec;
			--vcnt;			
		} else {
			break;
		}
	}

	maxPoints->Compress();

}

Double_t GetGridSpacing(TObjArray *maxPoints)
{
	UInt_t nrOfVals = maxPoints->GetEntriesFast();
	/*
	if (nrOfVals != 6) {
		printf("nr of values not 6: %d\n", nrOfVals);
		return -1;
	}
	*/
	if (nrOfVals == 0) {
		//printf("no maxima found\n");
		return -1;
	}

	Double_t values[6];
	UInt_t i;
	for (i = 0; i < nrOfVals; ++i) {
		TVector2 *vec = (TVector2*)maxPoints->At(i);
		if (vec) {
			values[i] = vec->Mod();
		} else {
			printf("a vector was null!\n");
			return -1;
		}
	}

	return TMath::Median((Long64_t)nrOfVals,(const double*)(&values));
}

Double_t GetMaxDist(TObjArray *maxPoints)
{
	UInt_t nrOfVals = maxPoints->GetEntriesFast();

	if (nrOfVals == 0) {
		//printf("no maxima found\n");
		return -1;
	}

	Double_t values[6];
	UInt_t i;
	for (i = 0; i < nrOfVals; ++i) {
		TVector2 *vec = (TVector2*)maxPoints->At(i);
		if (vec) {
			values[i] = vec->Mod();
		} else {
			printf("a vector was null!\n");
			return -1;
		}
	}

	i = TMath::LocMax((Long64_t)nrOfVals,(const double*)(&values));

	return values[i];
}



Double_t GetGridOrientation(TObjArray *maxPoints)
{
	Double_t  mx = 0;
	Double_t  my = 0;
	TVector2 *mxVec = 0;
	UInt_t nrOfVals = maxPoints->GetEntriesFast();

	if (nrOfVals == 0) {
		//printf("no maxima found\n");
		return -5;
	}

	UInt_t i;
	for (i = 0; i < nrOfVals; ++i) {
		TVector2 *vec = (TVector2*)maxPoints->At(i);
		if ((vec) && (vec->X() > mx) && (vec->Y() > 0)) {
			mx = vec->X();
			my = vec->Y();
			mxVec = vec;
		}
	}
	if ((mxVec) && (mx > 0))
		return TMath::ATan(my/mx);
	
	return 2;
}

Double_t GetFieldSize(TMatrixD &input, Double_t thres = 0.2, Int_t cx = -1, Int_t cy = -1)
{
	Double_t result = 1.0;

	Int_t nRows = input.GetNrows();
	Int_t nCols = input.GetNcols();

	if (cx == -1) cx = nCols / 2;
	if (cy == -1) cy = nRows / 2;

	std::vector<Bool_t> tmp(nRows*nCols,kFALSE);
	tmp[nCols * cy + cx] = kTRUE;

	std::queue< std::pair<Int_t,Int_t> > points;
	points.push(std::make_pair(cx,cy));

	while (points.size() > 0) {
		auto cp = points.front();
		points.pop();

		cx = cp.first;
		cy = cp.second;

		Double_t compVal = input(cy,cx);

		for (Int_t dy = -1; dy <= 1; ++dy) {
			for (Int_t dx = -1; dx <= 1; ++dx) {
				Int_t x = cx + dx;
				Int_t y = cy + dy;
				if ((x < 0) || (x >= nCols) || (y < 0) || (y >= nRows))
					continue;
				Double_t curVal = input(y,x);
				if ((tmp[nCols * y + x] == kTRUE) || (curVal > compVal) || (curVal < thres))
					continue;

				points.push(std::make_pair(x,y));
				tmp[nCols * y + x] = kTRUE;
				result += 1.0;

			}
		}
	}

	return result;
}

void SetCircle(TMatrixD &matrix, Int_t cx, Int_t cy, Double_t radius, Double_t val, Bool_t inverse = kFALSE)
{
	Int_t nRows = matrix.GetNrows();
	Int_t nCols = matrix.GetNcols();

	if (inverse == kFALSE) {
		Int_t scan = (Int_t)radius + 1;

		for (Int_t dy = -scan; dy <= scan; ++dy) {
			for (Int_t dx = -scan; dx <= scan; ++dx) {
				Int_t x = cx + dx;
				Int_t y = cy + dy;
				Double_t dist = TMath::Sqrt(dy*dy+dx*dx);
				if ((x < 0) || (x >= nCols) || (y < 0) || (y >= nRows))
					continue;
				if (dist < radius)
					matrix(y,x) = val;
				
			}
		}
	} else {
		for (Int_t y = 0; y < nRows; ++y) {
			for (Int_t x = 0; x < nCols; ++x) {
				Int_t dx = x - cx;
				Int_t dy = y - cy;
				Double_t dist = TMath::Sqrt(dy*dy+dx*dx);
				if (dist > radius)
					matrix(y,x) = val;
				
			}
		}		
	}
}

TMatrixD* GetMaskedAC(TMatrixD &ac, Double_t thres = 0.2, Int_t mxnb = 10)
{
	Int_t nRows = ac.GetNrows();
	Int_t nCols = ac.GetNcols();

	TMatrixD *result = new TMatrixD(ac);

	TObjArray mp;

	GetMax(*result, mxnb, &mp);
	ReduceTo6(&mp);

	Double_t fr = TMath::Sqrt(GetFieldSize(*result, thres) / TMath::Pi());

	Double_t mxd = GetMaxDist(&mp) + fr;

	SetCircle(*result,nCols/2,nRows/2,fr,0);
	SetCircle(*result,nCols/2,nRows/2,mxd,0,kTRUE);

	return result;
}


void RotateMatrix(TMatrixD &input, TMatrixD &output, Double_t rad, Double_t cx = 0, Double_t cy = 0)
{
	Int_t nRows = input.GetNrows();
	Int_t nCols = input.GetNcols();

	output.ResizeTo(nRows,nCols);
	output.Zero();

	Double_t cosRad = TMath::Cos(-rad);
	Double_t sinRad = TMath::Sin(-rad);

	for (Int_t y = 0; y < nRows; ++y) {
		for (Int_t x = 0; x < nCols; ++x) {
			// get src coords
			Double_t sx = (((Double_t)x - cx) *  cosRad + ((Double_t)y - cy) * sinRad) + cx;
			Double_t sy = (((Double_t)x - cx) * -sinRad + ((Double_t)y - cy) * cosRad) + cy;

			if ((sx < 0) || (sx >= nCols-1) || (sy < 0) || (sy >= nRows-1))
				continue;

			Int_t tlx = (Int_t)sx;
			Int_t tly = (Int_t)sy;

			Double_t fx = sx - TMath::Floor(sx);
			Double_t fy = sy - TMath::Floor(sy);

			output(y,x) = (input(tly,tlx)   * (1.0 - fx) + input(tly,tlx+1)   * fx) * (1.0 - fy) + 
						  (input(tly+1,tlx) * (1.0 - fx) + input(tly+1,tlx+1) * fx) * fy;

		}
	}

}

Double_t CalculateGridness(TMatrixD &ac, Double_t thres = 0.2, Int_t mxnb = 10)
{
	Int_t nRows = ac.GetNrows();
	Int_t nCols = ac.GetNcols();

	Double_t result = -1;

	TMatrixD *mac = GetMaskedAC(ac,thres,mxnb);

	TMatrixD rot;

	RotateMatrix(*mac,rot,60.0*TMath::Pi()/180.0,nCols/2,nRows/2);
	Double_t cr60 = GetCorrelation(*mac,rot);

	RotateMatrix(*mac,rot,120.0*TMath::Pi()/180.0,nCols/2,nRows/2);
	Double_t cr120 = GetCorrelation(*mac,rot);

	RotateMatrix(*mac,rot,30.0*TMath::Pi()/180.0,nCols/2,nRows/2);
	Double_t cr30 = GetCorrelation(*mac,rot);

	RotateMatrix(*mac,rot,90.0*TMath::Pi()/180.0,nCols/2,nRows/2);
	Double_t cr90 = GetCorrelation(*mac,rot);

	RotateMatrix(*mac,rot,150.0*TMath::Pi()/180.0,nCols/2,nRows/2);
	Double_t cr150 = GetCorrelation(*mac,rot);

	result = TMath::Min(cr60,cr120) - TMath::Max(cr30,TMath::Max(cr90,cr150));

	delete mac;

	return result;
}



TH2D* GetGridnessDev(TTree *gngTree, Int_t steps, ULong64_t maxTime = 0, Double_t sigma = 0.03, Double_t thres = 0.2)
{
	ULong64_t  time;
	TGNG      *gng = 0;

	gngTree->SetBranchAddress("inputCnt",&time);
	gngTree->SetBranchAddress("gng",&gng);

	Int_t nrOfEntries = gngTree->GetEntries();

	if (maxTime != 0) {
		gngTree->SetBranchStatus("*",0); //disable all branches
	    gngTree->SetBranchStatus("inputCnt",1);

	    for (Int_t i = 0; i < nrOfEntries; ++i) {
	    	gngTree->GetEntry(i);
	    	if (time > maxTime) {
	    		nrOfEntries = i;
	    		break;
	    	}
	    }

	    gngTree->SetBranchStatus("*",1); //enable all branches
	}

	Int_t stepWidth = nrOfEntries / steps;

	if (stepWidth == 0) {
		stepWidth = 1;
		steps = nrOfEntries;
	}

	TH2D *result = new TH2D("candleHist","distribution of gridness",steps,0,(Double_t)steps,20,-1.0,1.0);
	result->SetDirectory(0);

	for (Int_t i = 0; i < steps; ++i) {
		gngTree->GetEntry(i * stepWidth);

		Int_t nrOfUnits = gng->fUnits.GetEntriesFast();
		for (Int_t j = 0; j < nrOfUnits; ++j) {
			printf("step %d: gng %d of %d           \r",i+1,j+1,nrOfUnits);
			fflush(stdout);

			TGNGUnit *curUnit = static_cast<TGNGUnit*>(gng->fUnits.At(j));
			TGNGRecursiveData *rd = static_cast<TGNGRecursiveData*>(curUnit->fPrototype);

			TMatrixD *act = DrawUnitCenters(&(rd->fGNG), 128, 128, sigma);
			TMatrixD *aco = GetAutoCorrelation(act);

			Double_t gridness = CalculateGridness(*aco,thres);

			result->Fill((Double_t)i + 0.5,gridness);

			delete aco;
			delete act;
		}
	
		result->GetXaxis()->SetBinLabel(i+1,Form("%llu",time));

		delete gng;
		gng = 0;
	}
	printf("\ndone.\n");

	return result;
}




void SmoothRateMap(TMatrixD &fire, TMatrixD &pos, TMatrixD &output)
{
	Int_t nCols = output.GetNcols();
	Int_t nRows = output.GetNrows();

	output.Zero();

	Int_t iSize = fire.GetNrows();

	//Double_t xScale = (Double_t)nCols / (Double_t)iSize;
	//Double_t yScale = (Double_t)nRows / (Double_t)iSize;

	Double_t xScale = (Double_t)iSize / (Double_t)nCols;
	Double_t yScale = (Double_t)iSize / (Double_t)nRows;

	Double_t sigma = 3;

	Int_t x,y,dx,dy;
	// generate output
	for (y = 0; y < nRows; ++y)
		for (x = 0; x < nCols; ++x) {
			Double_t actv = 0;
			Double_t posv = 0;
			for (dy = 0; dy < iSize; ++dy)
				for (dx = 0; dx < iSize; ++dx) {
					Double_t iX = (Double_t)dx;// * xScale;
					Double_t iY = (Double_t)dy;// * yScale;
					Double_t eX = (Double_t)x * xScale;
					Double_t eY = (Double_t)y * yScale;
					actv += fire(dy,dx) * TMath::Exp(-0.5*(TMath::Power( (iX-eX)/sigma,2.0)) -
					                                  0.5*(TMath::Power( (iY-eY)/sigma,2.0))   );
					posv += pos(dy,dx) * TMath::Exp(-0.5*(TMath::Power( (iX-eX)/sigma,2.0)) -
					                                 0.5*(TMath::Power( (iY-eY)/sigma,2.0))   );
				}
			if (posv > 0)
				output(y,x) = actv / posv;

		}

}

void SmoothRateMap2(TMatrixD &fire, TMatrixD &pos, TMatrixD &output, Double_t sigma = 3.0)
{
	Int_t nCols = output.GetNcols();
	Int_t nRows = output.GetNrows();

	output.Zero();

	Int_t iSize = fire.GetNrows();

	//Double_t xScale = (Double_t)nCols / (Double_t)iSize;
	//Double_t yScale = (Double_t)nRows / (Double_t)iSize;

	Double_t xScale = (Double_t)iSize / ((Double_t)nCols + (6 * sigma));
	Double_t yScale = (Double_t)iSize / ((Double_t)nRows + (6 * sigma));

	Int_t x,y,dx,dy;
	// generate output
	for (y = 0; y < nRows; ++y)
		for (x = 0; x < nCols; ++x) {
			Double_t actv = 0;
			Double_t posv = 0;
			for (dy = 0; dy < iSize; ++dy)
				for (dx = 0; dx < iSize; ++dx) {
					Double_t iX = (Double_t)dx + 0.5;// * xScale;
					Double_t iY = (Double_t)dy + 0.5;// * yScale;
					Double_t eX = ((Double_t)x + 0.5) * xScale + (3 * sigma);
					Double_t eY = ((Double_t)y + 0.5) * yScale + (3 * sigma);
					actv += fire(dy,dx) * TMath::Exp(-0.5*(TMath::Power( (iX-eX)/sigma,2.0)) -
					                                  0.5*(TMath::Power( (iY-eY)/sigma,2.0))   );

					posv += pos(dy,dx) * TMath::Exp(-0.5*(TMath::Power( (iX-eX)/sigma,2.0)) -
					                                 0.5*(TMath::Power( (iY-eY)/sigma,2.0))   );
				}
			if (posv > 1)
				output(y,x) = actv / posv;

		}

}


void SmoothRateMapBC(TMatrixD &fire, TMatrixD &pos, TMatrixD &output)
{
	// BoxCar-Filter used by Stensola2012

	Double_t bcFilter[5][5] = {
		{0.0025, 0.0125, 0.0200, 0.0125, 0.0025},
		{0.0125, 0.0625, 0.1000, 0.0625, 0.0125},
		{0.0200, 0.1000, 0.1600, 0.1000, 0.0200},
		{0.0125, 0.0625, 0.1000, 0.0625, 0.0125},
		{0.0025, 0.0125, 0.0200, 0.0125, 0.0025}
	};


	Int_t nCols = output.GetNcols();
	Int_t nRows = output.GetNrows();

	output.Zero();

	Int_t x,y,dx,dy;
	// generate output
	for (y = 2; y < nRows-2; ++y)
		for (x = 2; x < nCols-2; ++x) {
			Double_t actv = 0;
			Double_t posv = 0;
			for (dy = -2; dy <= 2; ++dy)
				for (dx = -2; dx <= 2; ++dx) {
					actv += fire(y+dy,x+dx) * bcFilter[dy+2][dx+2];
					posv += pos(y+dy,x+dx) * bcFilter[dy+2][dx+2];
				}
			if (posv > 1)
				output(y,x) = actv / posv;
		}

}


void SmoothRateMapBC2(TMatrixD &fire, TMatrixD &pos, TMatrixD &output)
{
	// BoxCar-Filter used by Stensola2012

	Double_t bcFilter[5][5] = {
		{0.0025, 0.0125, 0.0200, 0.0125, 0.0025},
		{0.0125, 0.0625, 0.1000, 0.0625, 0.0125},
		{0.0200, 0.1000, 0.1600, 0.1000, 0.0200},
		{0.0125, 0.0625, 0.1000, 0.0625, 0.0125},
		{0.0025, 0.0125, 0.0200, 0.0125, 0.0025}
	};


	Int_t nCols = output.GetNcols();
	Int_t nRows = output.GetNrows();

	output.Zero();

	Int_t x,y,dx,dy;
	// generate output
	for (y = 0; y < nRows; ++y)
		for (x = 0; x < nCols; ++x) {
			Double_t actv = 0;
			Double_t posv = 0;
			for (dy = -2; dy <= 2; ++dy)
				for (dx = -2; dx <= 2; ++dx) {
					Int_t ax = x+dx;
					Int_t ay = y+dy;
					if ((ax < 0) || (ax >= nCols) || (ay < 0) || (ay >= nRows))
						continue;
					actv += fire(y+dy,x+dx) * bcFilter[dy+2][dx+2];
					posv += pos(y+dy,x+dx) * bcFilter[dy+2][dx+2];
				}
			if (posv > 1)
				output(y,x) = actv / posv;
		}

}

void GetCellActivity(TTree *act, TTree *feed, Int_t cellID, ULong64_t start, ULong64_t stop, TMatrixD &fire, TMatrixD &posCnt, Double_t shift = 0, Double_t scale = 1.0)
{
	Int_t iSize = fire.GetNcols();

	//TMatrixD fireCnt(iSize,iSize);

	fire.Zero();
	//fireCnt.Zero();
	posCnt.Zero();


	ULong64_t aTime = 0;
	Int_t     aUnit;
	Double_t  aRatio;

	act->SetBranchStatus("*",0);

	act->SetBranchStatus("inputCnt",1);
	act->SetBranchAddress("inputCnt",&aTime);

	act->SetBranchStatus("unitIdx",1);
	act->SetBranchAddress("unitIdx",&aUnit);

	act->SetBranchStatus("ratio",1);
	act->SetBranchAddress("ratio",&aRatio);

	Int_t actSize = act->GetEntries();

	ULong64_t  fTime = 0;
	TObject   *fRawInput = 0;
	TVector2  *fPos = 0;

	feed->SetBranchStatus("*",0);

	feed->SetBranchStatus("inputCnt",1);
	feed->SetBranchAddress("inputCnt",&fTime);

	feed->SetBranchStatus("rawinput",1);
	feed->SetBranchAddress("rawinput",&fRawInput);

	Int_t feedSize = feed->GetEntries();

	Int_t actIdx  = 0;
	Int_t feedIdx = 0;

	// find start idx
	act->SetBranchStatus("*",0);
	act->SetBranchStatus("inputCnt",1);
	for (actIdx = 0; actIdx < actSize; ++actIdx) {
		act->GetEntry(actIdx);
		if (aTime >= start)
			break;
	}

	feed->SetBranchStatus("*",0);
	feed->SetBranchStatus("inputCnt",1);
	for (feedIdx = 0; feedIdx < feedSize; ++feedIdx) {
		feed->GetEntry(feedIdx);
		if (fTime >= start)
			break;
	}

	// activate other branches
	act->SetBranchStatus("unitIdx",1);
	act->SetBranchStatus("ratio",1);
	feed->SetBranchStatus("rawinput",1);

	// go from start to stop
	for (ULong64_t t = start; t < stop; ++t) {
		// find position in act
		while (actIdx < actSize) {
			act->GetEntry(actIdx++);
			if (aTime < t)
				continue;
			if ((aTime == t) && (aUnit != cellID))
				continue;
			break;
		}
		if ((aTime != t) || (aUnit != cellID))
			continue;
		// find position in feed
		while (feedIdx < feedSize) {
			feed->GetEntry(feedIdx++);
			if (fTime < t)
				continue;
			break;
		}		
		if (fTime != t)
			continue;
		// convert rawInput to vector - if possible
		fPos = dynamic_cast<TVector2*>(fRawInput);
		if (fPos == 0)
			continue;
		// fill matrices
		Int_t ixpos = (Int_t)(TMath::Floor((fPos->X() + shift) * scale * (Double_t)iSize));
		Int_t iypos = (Int_t)(TMath::Floor((fPos->Y() + shift) * scale * (Double_t)iSize));

		if ((ixpos < 0) || (ixpos >= iSize) || (iypos < 0) || (iypos >= iSize))
			continue;

		posCnt(iypos,ixpos) += 1;

		//fire(iypos,ixpos) += aRatio;
		fire(iypos,ixpos) += TMath::Exp(-0.5*TMath::Power((1.0 - aRatio)/0.2,2.0));

		/*
		//if ( aRatio*aRatio > gRandom->Uniform()) {
		if ((1.0 - aRatio) < TMath::Abs(gRandom->Gaus(0.0,thres))) {
		//if (( aRatio > gRandom->Uniform()) && (aRatio > thres)) {

			if (constAct)
				fire(iypos,ixpos) += 1;
			else
				fire(iypos,ixpos) += aRatio;
			//fireCnt(iypos,ixpos) += 1;

		}
		*/

	}

}



void GetCellActivities(TTree *act, TTree *feed, Int_t steps, Int_t sampleDur, Int_t mapSize, TObjArray &fire, TObjArray &posCnt, Double_t shift = 0, Double_t scale = 1.0)
{
	// the array fire will contain a TObjArray* for each step. Those
	// TObjArrays then contain TMatrixD* holding the raw activity maps of the
	// individual Top-Layer units.

	// the array posCnt will contain a TObjArray* for each step. Those
	// TObjArrays then contain TMatrixD* holding the position count maps of the
	// individual Top-Layer units.

	fire.Clear();
	posCnt.Clear();

	ULong64_t aTime = 0;
	Int_t     aUnit;
	Double_t  aRatio;

	act->SetBranchStatus("*",0);

	act->SetBranchStatus("inputCnt",1);
	act->SetBranchAddress("inputCnt",&aTime);

	act->SetBranchStatus("unitIdx",1);
	act->SetBranchAddress("unitIdx",&aUnit);

	act->SetBranchStatus("ratio",1);
	act->SetBranchAddress("ratio",&aRatio);

	Long64_t actSize = act->GetEntries();

	ULong64_t  fTime = 0;
	TObject   *fRawInput = 0;
	TVector2  *fPos = 0;

	feed->SetBranchStatus("*",0);

	feed->SetBranchStatus("inputCnt",1);
	feed->SetBranchAddress("inputCnt",&fTime);

	feed->SetBranchStatus("rawinput",1);
	feed->SetBranchAddress("rawinput",&fRawInput);

	Int_t feedSize = feed->GetEntries();

	// get stepsize and maxUnit ID
	Int_t maxUnit = -1;
	ULong64_t  maxTime = 0;

	for (Long64_t i = 0; i < actSize; ++i) {
		if ((i % 1000) == 0) {
			printf("scanning entry %lld of %lld                  \r",i+1,actSize);
			fflush(stdout);
		}
		act->GetEntry(i);
		if (aTime > maxTime)
			maxTime = aTime;
		if (aUnit > maxUnit)
			maxUnit = aUnit;
	}

	Int_t stepWidth = maxTime / steps;

	if (stepWidth < sampleDur) {
		stepWidth = sampleDur;
		steps = maxTime / stepWidth;
	}

	Long64_t actIdx  = 0;
	Long64_t feedIdx = 0;

	for (Int_t s = 0; s < steps; ++s) {
		TObjArray *curFireArray = new TObjArray();
		curFireArray->SetOwner(kTRUE);
		TObjArray *curPosArray = new TObjArray();
		curPosArray->SetOwner(kTRUE);
		for (Int_t i = 0; i <= maxUnit; ++i) {
			TMatrixD *fa = new TMatrixD(mapSize,mapSize);
			TMatrixD *pa  = new TMatrixD(mapSize,mapSize);
			fa->Zero();
			pa->Zero();
			curFireArray->Add(fa);
			curPosArray->Add(pa);
		}
		fire.Add(curFireArray);
		posCnt.Add(curPosArray);

		ULong64_t startIdx = s * stepWidth;
		ULong64_t endIdx   = startIdx + sampleDur;

		// find start in act
		while (actIdx < actSize) {
			act->GetEntry(actIdx++);
			if (aTime < startIdx)
				continue;
			break;			
		}

		// find start in feed
		while (feedIdx < feedSize) {
			feed->GetEntry(feedIdx++);
			if (fTime < startIdx)
				continue;
			break;
		}		

		if ((aTime != startIdx) || (fTime != startIdx))
			continue;

		// read sample duration
		for (ULong64_t t = startIdx; t < endIdx; ++t) {

			printf("step %d: sample %lld               \r",s+1,t);
			fflush(stdout);

			// convert rawInput to vector - if possible
			fPos = dynamic_cast<TVector2*>(fRawInput);

			if (fPos != 0) {
				// get position in matrices
				Int_t xpos = (Int_t)(TMath::Floor((fPos->X() + shift) * scale * (Double_t)mapSize));
				Int_t ypos = (Int_t)(TMath::Floor((fPos->Y() + shift) * scale * (Double_t)mapSize));

				if ((TMath::IsNaN(xpos)) || (xpos < 0) || (xpos >= mapSize) ||
					(TMath::IsNaN(ypos)) || (ypos < 0) || (ypos >= mapSize)    ) {
					// find next t in act
					while (actIdx < actSize) {
						act->GetEntry(actIdx++);
						if (aTime <= t)
							continue;
						break;			
					}
					// find next t in feed
					while (feedIdx < feedSize) {
						feed->GetEntry(feedIdx++);
						if (fTime <= t)
							continue;
						break;
					}		
					continue;
				}

				while ((aTime == t) && (actIdx < actSize)) {

					// fill in fire and pos data for current unit

					// get matrices
					TMatrixD *curFire = static_cast<TMatrixD*>(curFireArray->At(aUnit));
					TMatrixD *curPos = static_cast<TMatrixD*>(curPosArray->At(aUnit));

					(*curPos)(ypos,xpos) += 1;

					//(*curFire)(ypos,xpos) += aRatio;
					if (TMath::IsNaN(aRatio))
						aRatio = 0;

					(*curFire)(ypos,xpos) += TMath::Exp(-0.5*TMath::Power((1.0 - aRatio)/0.2,2.0));

					act->GetEntry(actIdx++);
				}
			} else {
				// find next t in act
				while (actIdx < actSize) {
					act->GetEntry(actIdx++);
					if (aTime <= t)
						continue;
					break;			
				}
			}

			// find next t in feed
			while (feedIdx < feedSize) {
				feed->GetEntry(feedIdx++);
				if (fTime <= t)
					continue;
				break;
			}		


		}


	}

}

void AnalyseFile(const char *input, const char *output, Int_t steps, 
				 Double_t shift = 0, Double_t scale = 1.0,
	             Int_t mapRes = 50, Int_t samplDur = 30000, 
	             Double_t thres = 0.2, Int_t mxnb = 2)
{
	TFile *inpFile = TFile::Open(input);

	TTree *feedTree = (TTree*)inpFile->Get("feedTree_0");
	TTree *actTree  = (TTree*)inpFile->Get("activityTree");

	TObjArray fire;
	fire.SetOwner(kTRUE);

	TObjArray posCnt;
	posCnt.SetOwner(kTRUE);

	GetCellActivities(actTree, feedTree, steps, samplDur, mapRes, fire, posCnt,shift,scale);

	ULong64_t time;

	actTree->SetBranchStatus("*",0);

	actTree->SetBranchStatus("inputCnt",1);
	actTree->SetBranchAddress("inputCnt",&time);

	Long64_t actSize = actTree->GetEntries();

	ULong64_t  maxTime = 0;

	for (Long64_t i = 0; i < actSize; ++i) {
		actTree->GetEntry(i);
		if (time > maxTime)
			maxTime = time;
	}

	Int_t stepWidth = maxTime / steps;

	if (stepWidth < samplDur) {
		stepWidth = samplDur;
		steps = maxTime / stepWidth;
	}

	Int_t mapRes2 = mapRes*2;	

	TFile *outFile = TFile::Open(output,"new");

	TH2D *gridHist = new TH2D("griddist","distribution of gridness",steps,0,(Double_t)maxTime,40,-2.0,2.0);
	TH2D *spaceHist = new TH2D("spacedist","distribution of grid spacing",steps,0,(Double_t)maxTime,mapRes2,0,mapRes2);
	TH2D *mxDistHist = new TH2D("mxdistdist","distribution of max spacing",steps,0,(Double_t)maxTime,mapRes2,0,mapRes2);
	TH2D *orientHist = new TH2D("orientdist","distribution of grid orientation",steps,0,(Double_t)maxTime,180,-TMath::Pi()/2.0,TMath::Pi()/2.0);
	TH2D *fieldHist = new TH2D("fielddist","distribution of grid field size",steps,0,(Double_t)maxTime,(mapRes2*mapRes2)/100,0,mapRes2*mapRes2/10);	

	Int_t nodeID = 0;
	TTree *mapTree = new TTree("ratemaps","all rate and autocorrelation maps");
	mapTree->Branch("inputCnt",&time,"idx/l");
	mapTree->Branch("nodeID",&nodeID,"nid/I");

	TMatrixD rmap;
	mapTree->Branch("ratemap","TMatrixD",&rmap,160000,0);

	TMatrixD amap;
	mapTree->Branch("automap","TMatrixD",&amap,160000,0);

	Double_t spacing;
	mapTree->Branch("spacing",&spacing,"spc/D");

	Double_t mxDist;
	mapTree->Branch("maxdist",&mxDist,"mxd/D");

	Double_t orientation;
	mapTree->Branch("orientation",&orientation,"or/D");

	Double_t fieldSize;
	mapTree->Branch("fieldsize",&fieldSize,"fsiz/D");

	Double_t gridness;
	mapTree->Branch("gridness",&gridness,"grd/D");

	TMatrixD smo(mapRes, mapRes);

	Int_t nrOfSteps = fire.GetEntriesFast();
	for (Int_t s = 0; s < nrOfSteps; ++s) {
		TObjArray *curFireMaps = static_cast<TObjArray*>(fire.At(s));
		TObjArray *curPosMaps  = static_cast<TObjArray*>(posCnt.At(s));
		Int_t nrOfUnits = curFireMaps->GetEntriesFast();
		for (Int_t u = 0; u < nrOfUnits; ++u) {
			printf("step %d: gng %d of %d           \r",s+1,u+1,nrOfUnits);
			fflush(stdout);

			TMatrixD *fmap = static_cast<TMatrixD*>(curFireMaps->At(u));
			TMatrixD *pmap = static_cast<TMatrixD*>(curPosMaps->At(u));

			SmoothRateMapBC2(*fmap,*pmap,smo);

			TMatrixD *aco = GetAutoCorrelation(&smo);

			nodeID = u;
			time = s * stepWidth;
			rmap.ResizeTo(smo);
			rmap = smo;
			amap.ResizeTo(*aco);
			amap = *aco;
			
			TObjArray mp;
			mp.SetOwner(kTRUE);

			GetMax(*aco, mxnb, &mp);
			ReduceTo6(&mp);

			spacing = GetGridSpacing(&mp);

			mxDist = GetMaxDist(&mp);

			orientation = GetGridOrientation(&mp);

			fieldSize = GetFieldSize(*aco, thres);

			gridness = CalculateGridness(*aco,thres,mxnb);

			mapTree->Fill();

			gridHist->Fill((Double_t)time,gridness);

			spaceHist->Fill((Double_t)time,spacing);

			mxDistHist->Fill((Double_t)time,mxDist);

			orientHist->Fill((Double_t)time,orientation);

			fieldHist->Fill((Double_t)time,fieldSize);

			delete aco;

		}
	}

	printf("\ndone.\n");

	gridHist->Write();
	spaceHist->Write();
	mxDistHist->Write();
	orientHist->Write();
	fieldHist->Write();

	mapTree->Write();

	outFile->Close();
	inpFile->Close();

	delete outFile;
	delete inpFile;
}

void AnalyseDir(const char *dir, const char *inprefix, const char *outprefix, 
	            Int_t steps, Double_t shift = 0, Double_t scale = 1.0, 
	            Int_t mapRes = 50, Int_t samplDur = 30000, 
	            Double_t thres = 0.2, Int_t mxnb = 2)
{
	TSystemDirectory sysdir(dir, dir);
	TList *files = sysdir.GetListOfFiles();

	if (!files) return;

	TSystemFile *file;
	TString fname;
	TIter next(files);
	while ((file=(TSystemFile*)next())) {
		fname = file->GetName();
		char *anafile = gSystem->Which(dir,Form("%s%s",outprefix,fname.Data()));
		if (anafile) {
			printf("File already there\n");
			delete anafile;
			continue;
		}
		if (!file->IsDirectory() && fname.EndsWith(".root") && fname.BeginsWith(inprefix)) {
			printf("Found file: %s\n", fname.Data());
	    	printf("full path:  %s\n", Form("%s/%s",dir,fname.Data()));
	    	printf("output: %s\n", Form("%s/%s%s",dir,outprefix,fname.Data()) );
	    	AnalyseFile(Form("%s/%s",dir,fname.Data()),
	    	  Form("%s/%s%s",dir,outprefix,fname.Data()),steps,shift,scale,mapRes,samplDur,
	    	  thres,mxnb);
		}
	}
}

TMatrixD *GetMatrixFromHist(TH2D *input)
{
	TMatrixD *result = new TMatrixD(input->GetNbinsY(),input->GetNbinsX());
	for (Int_t r = 0; r < input->GetNbinsY(); ++r)
		for (Int_t c = 0; c < input->GetNbinsX(); ++c)
			(*result)(r,c) = input->GetBinContent(c+1,r+1);
	return result;	
}

Double_t GetGridness(TObjArray *ratemaps, Int_t idx,Double_t thres = 0.2, Int_t mxnb = 2)
{
	// get ratemap
	TH2D *rmaphist = static_cast<TH2D*>(ratemaps->At(idx));	
	TMatrixD *rmap = GetMatrixFromHist(rmaphist);
	TMatrixD *aco = GetAutoCorrelation(rmap);
	Double_t result = CalculateGridness(*aco,thres,mxnb);
	delete aco;
	delete rmap;
	return result;
}
