/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of the main TGNG class.

This is an extended implementation of the growing neural gas algorithm 
originally proposed by Bernd Fritzke. 

In addition to the original algorithm this implementation features the following
extensions (all disabled by default):



Epsilon Ramp:
The learning rate Epsilon_b and Epsilon_n can be changed into Epsilon_be and 
Epsilon_ne over a given time period (RampTime).

The interpolation is performed like this:

eps_b = fEpsilon_b * TMath::Power(fEpsilon_be / fEpsilon_b, (Double_t)fInpCnt / fRampTime);
eps_n = fEpsilon_n * TMath::Power(fEpsilon_ne / fEpsilon_n, (Double_t)fInpCnt / fRampTime);

Use SetEpsilonRamp(Bool_t enable) to enable or disable this feature. The
learning rates and the ramp time can be set using:

void SetEpsilon_b(Double_t val);
void SetEpsilon_be(Double_t val);
void SetEpsilon_n(Double_t val);
void SetEpsilon_ne(Double_t val);
void SetRampTime(Double_t val);



Learning Inhibition:
The concept of learning inhibition introduces a kind of refratory period for
every unit of the GNG where the unit will refuse to learn any new input. After
a unit has become BMU its member fLearnInh is set to one. In subsequent steps,
the learning rate is multiplied by (1.0 - fLearnInh) for this unit. In addition
the fLearnInh values of all units are reduced by the parameter fGamma in every
time step, releasing the learning inhibition if the unit becomes not the BMU
for a number of inputs:

curUnit->fLearnInh -= fGamma * curUnit->fLearnInh;
if (curUnit->fLearnInh < 0.0)
	curUnit->fLearnInh = 0.0;

The parameter fGamma can be set using SetGamma(Double_t val). The learning
inhibition can be enabled or disabled using SetLearningInhibition(Bool_t enable).



Lambda delay
The parameter fLambdaDelay allows to set a delay for the GNG to start growing.
Use SetLambdaDelay(UInt_t val) to set this delay in terms of number of inputs
that should pass before the GNG begins to grow.



LISH weighting
If the LISH extension was added to the GNG the learning rate can be augmented
by a factor that is derived from the LISH between BMU and second BMU. The factor
is derived as:

1.0 - 1.0 / TMath::Power(1.0 + 0.5 * TMath::Exp(-fLishWeightSteepness * (avg - fLishWeightCenter)) ,2)

with the parameters fLishWeightSteepness and fLishWeightCenter as well as the
average LISH bin error "avg".

LISH weighting is enabled or disabled using SetLishWeighting(Bool_t enable).
The parameters can be set using:

void SetLishWeightCenter(Double_t center);
void SetLishWeightSteepness(Double_t steepness);



*/
#include <TMath.h>
#include <TH1D.h>
#include <TH2D.h>
#include <TObjString.h>

#include "TGNG.h"



TGNG::TGNG() :
	TNamed("",""),
	fUnits(),
	fAttributes(),
	fEpsilon_b(0.05),
	fEpsilon_be(0.05),
	fEpsilon_n(0.0006), 
	fEpsilon_ne(0.0006), 
	fRampTime(500000),
	fLambda(300),    
	fT_max(100),     
	fAlpha(0.5),     
	fBeta(0.0005),
	fMaxUnitCnt(100),
	fLambdaDelay(0),
	fGamma(0.02),
	fInpCnt(0),
	fDistMeasure(0),
	fInitialized(kFALSE),
	fExtensions(),
	fLishWeight(kFALSE),
	fLishWeightCenter(0.25),
	fLishWeightSteepness(25.0),
	fRawInput(0),
	fUseLearningInhibition(kFALSE),
	fUseEpsilonRamp(kFALSE),
	fGrowth(kTRUE)
{
	// default constructor setting the parameters of the GNG to
	// the values proposed in
	//
	// Fritzke, B. A Growing Neural Gas Network Learns Topologies 
	// Advances in Neural Information Processing Systems 7, 
	// MIT Press, 1995, 625-632
	//
	// the gng is not initialized

	fUnits.SetOwner(kTRUE);
	fExtensions.SetOwner(kTRUE);
	fAttributes.SetOwner(kTRUE);	
}


TGNG::TGNG(const char* name, const char* title) :
	TNamed(name,title),
	fUnits(),
	fAttributes(),
	fEpsilon_b(0.05),
	fEpsilon_be(0.05),
	fEpsilon_n(0.0006), 
	fEpsilon_ne(0.0006), 
	fRampTime(500000),
	fLambda(300),    
	fT_max(100),     
	fAlpha(0.5),     
	fBeta(0.0005),
	fMaxUnitCnt(100),
	fLambdaDelay(0),
	fGamma(0.02),
	fInpCnt(0),
	fDistMeasure(0),
	fInitialized(kFALSE),
	fExtensions(),
	fLishWeight(kFALSE),
	fLishWeightCenter(0.25),
	fLishWeightSteepness(25.0),
	fRawInput(0),
	fUseLearningInhibition(kFALSE),
	fUseEpsilonRamp(kFALSE),
	fGrowth(kTRUE)
{
	// constructor setting the parameters of the GNG to
	// the values proposed in
	//
	// Fritzke, B. A Growing Neural Gas Network Learns Topologies 
	// Advances in Neural Information Processing Systems 7, 
	// MIT Press, 1995, 625-632
	//
	// the gng is not initialized

	fUnits.SetOwner(kTRUE);
	fExtensions.SetOwner(kTRUE);
	fAttributes.SetOwner(kTRUE);
}

TGNG::~TGNG()
{
	// the GNG is reset and the extensions are deleted

	ResetGNG();
	fExtensions.Delete();
	fAttributes.DeleteAll();
}

void TGNG::Init(TGNGDataElement *prototypeA, TGNGDataElement *prototypeB, TGNGDistanceMeasure *distMeasure)
{
	// initializes the gng with two units that have the given prototypes
	// and the supplied distance measure.

	if ((prototypeA == 0) || (prototypeB == 0)) {
		RaiseError(TGNG::ecParamNull,"Prototypes measure can not be 0!");
		return;
	}

	if (distMeasure == 0) {
		RaiseError(TGNG::ecParamNull,"Distance measure can not be 0!");
		return;
	}

	if (distMeasure->IsCompatible(prototypeA,this) == kFALSE) {
		RaiseError(TGNG::ecIncomp,"Distance measure not compatible with data elements!");
		return;		
	}

	if (fInitialized == kTRUE)
		ResetGNG();

	fDistMeasure = distMeasure;

	TGNGUnit *newUnit = new TGNGUnit(prototypeA,0);
	fUnits.Add(newUnit);

	{ // report to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			curExt->OnNewUnit(this,newUnit);
		}
	}

	newUnit = new TGNGUnit(prototypeB,1);
	fUnits.Add(newUnit);

	{ // report to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			curExt->OnNewUnit(this,newUnit);
		}
	}

	fInitialized = kTRUE;

	{ // report to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			curExt->OnGNGInit(this);
		}
	}
}

void TGNG::ManualInit(TGNGDistanceMeasure *distMeasure)
{
	// starts manual initialization of the gng with no units and the supplied 
	// distance measure. New units and edges have to be added manually after 
	// a call to this method. The end of the manual initialization must be
	// signalled by a call to ManualInitDone().

	if (distMeasure == 0) {
		RaiseError(TGNG::ecParamNull,"Distance measure can not be 0!");
		return;
	}

	if (fInitialized == kTRUE)
		ResetGNG();

	fDistMeasure = distMeasure;
}

void TGNG::ManualInitDone()
{
	// finishes the manual initialization 

	if (fInitialized == kTRUE)
		return;

	Int_t nrOfUnits = fUnits.GetEntriesFast();

	if (nrOfUnits == 0) {
		RaiseError(TGNG::ecFatal,"No Units defined during manual init!");
		return;		
	}

	TGNGUnit *firstUnit = static_cast<TGNGUnit*>(fUnits.At(0));

	if (fDistMeasure->IsCompatible(firstUnit->fPrototype,this) == kFALSE) {
		RaiseError(TGNG::ecIncomp,"Distance measure not compatible with data elements!");
		return;		
	}

	// report units and edges to extensions
	{ 
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			for (Int_t j = 0; j < nrOfUnits; ++j) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(j));
				curExt->OnNewUnit(this,curUnit);
			}
		}
	}

	{ 
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			for (Int_t j = 0; j < nrOfUnits; ++j) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(j));
				Int_t nrOfEdges = curUnit->fEdges.GetEntriesFast();
				for (Int_t k = 0; k < nrOfEdges; ++k) {
					TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(k));
					curExt->OnNewEdge(this,curEdge);
				}
			}
		}
	}

	fInitialized = kTRUE;

	{ // report to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			curExt->OnGNGInit(this);
		}
	}
}

void TGNG::ResetGNG()
{
	// resets the GNG

	if (fInitialized == kFALSE)
		return;

	// delete edges manually as they are shared between units
	Int_t unitCnt = fUnits.GetEntriesFast();
	for (Int_t i = 0; i < unitCnt; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));

		Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
		for (Int_t j = 0; j < edgeCnt; ++j) {
			TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));

			// report to extensions
			if ((curEdge->fUnitA != 0) && (curEdge->fUnitB != 0)) { 
				Int_t nrOfExt = fExtensions.GetEntriesFast();
				for (Int_t k = 0; k < nrOfExt; ++k) {
					TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(k));
					curExt->OnEdgeRemoval(this,curEdge);
				}
			}	

			if (curEdge->fUnitA != 0) {
				curEdge->fUnitA = 0;
			} else if (curEdge->fUnitB != 0) {
				curEdge->fUnitB = 0;
			}

			if ((curEdge->fUnitA == 0) && (curEdge->fUnitB == 0)) {
				delete curEdge;
				curUnit->fEdges.RemoveAt(j);
			}
		}

		curUnit->fEdges.Clear();
	}

	for (Int_t i = 0; i < unitCnt; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));

		{ // report to extensions
			Int_t nrOfExt = fExtensions.GetEntriesFast();
			for (Int_t k = 0; k < nrOfExt; ++k) {
				TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(k));
				curExt->OnUnitRemoval(this,curUnit);
			}
		}	
	}

	fUnits.Delete();

	fInitialized = kFALSE;

	fAttributes.DeleteAll();

	Int_t nrOfExt = fExtensions.GetEntriesFast();
	for (Int_t i = 0; i < nrOfExt; ++i) {
		TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
		curExt->OnGNGReset(this);
	}	
}


TGNGUnit* TGNG::Feed(const TGNGDataElement *input, Double_t *minDist, TObject *rawInput, Double_t *ratio)
{
	// Inputs a vector into the GNG and returns the BMU.
	// The GNG is modified.

	if (fRawInput) {
		delete fRawInput;
		fRawInput = 0;
	}

	if (rawInput) {
		fRawInput = rawInput->Clone();
	}

	//printf("check inputs\n");
	if (fInitialized == kFALSE) {
		RaiseError(TGNG::ecNoInit,"GNG not initialized!");
		return 0;				
	}

	if (input == 0) {
		RaiseError(TGNG::ecParamNull,"Input is null!");
		return 0;				
	}

	fInpCnt++;

	// find first and second BMU
	//printf("find first and second BMU\n");
	TGNGUnit *s1 = 0;
	TGNGUnit *s2 = 0;

	Double_t s1ApproxDist = std::numeric_limits<Double_t>::max();
	Double_t s2ApproxDist = std::numeric_limits<Double_t>::max();

	Int_t unitCnt = fUnits.GetEntriesFast();

	for (Int_t i = 0; i < unitCnt; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));

		//Double_t curDist = fDistMeasure->GetApproxDistance(curUnit->fPrototype,input,this);
		Double_t curDist = fDistMeasure->GetDistance(curUnit->fPrototype,input,this);

		if (curDist < s1ApproxDist) {
			s2           = s1;
			s2ApproxDist = s1ApproxDist;
			s1           = curUnit;
			s1ApproxDist = curDist;
		} else if (curDist < s2ApproxDist) {
			s2           = curUnit;
			s2ApproxDist = curDist;
		}
	}

	if ((s1 == 0) || (s2 == 0)) {
		RaiseError(TGNG::ecNoBMU,"Could not find first and second BMU! Something is very wrong...");
		return 0;				
	}

	// get real distances 
	//printf("get real distances \n");
	//Double_t s1Dist = fDistMeasure->GetDistance(s1->fPrototype,input,this);
	//Double_t s2Dist = fDistMeasure->GetDistance(s2->fPrototype,input,this);

	Double_t s1Dist = s1ApproxDist;
	Double_t s2Dist = s2ApproxDist;

	if (minDist != 0) {
		*minDist = s1Dist;
	}

	Double_t dRatio = 1.0;
	Double_t relDist = fDistMeasure->GetDistance(s1->fPrototype,s2->fPrototype,this);

	if ((TMath::IsNaN(relDist) == kFALSE) && (relDist != 0)) {
		dRatio = ((s1Dist - s2Dist) / relDist) + 1.0;
	}

	if (dRatio > 1.0) dRatio = 1.0;
	if (dRatio < 0.0) dRatio = 0.0;		

	if (ratio) {
		*ratio = dRatio;
	}
	

	if (TMath::IsNaN(s1Dist)) {
		RaiseError(TGNG::ecNaN,"Distance to BMU is NaN! Aborting...");
		return 0;				
	}

	if (TMath::IsNaN(s2Dist)) {
		RaiseError(TGNG::ecNaN,"Distance to second BMU is NaN! Aborting...");
		return 0;				
	}

	// increase edge age of all s1 neighbours
	// try to find edge to s2
	//printf("increase edge age of all s1 neighbours / try to find edge to s2\n");
	TGNGEdge *s2Edge = 0;	
	Int_t edgeCnt = s1->fEdges.GetEntriesFast();
	for (Int_t i = 0; i < edgeCnt; ++i) {
		TGNGEdge *curEdge = static_cast<TGNGEdge*>(s1->fEdges.At(i));	
		if (fUseLearningInhibition) {	
			curEdge->fAge += (1.0 - s1->fLearnInh);
		} else {
			curEdge->fAge += 1;			
		}
		if ((curEdge->fUnitA == s2) || (curEdge->fUnitB == s2))
			s2Edge = curEdge;
	}

	// reset edge age to s2, create edge if necessary
	//printf("reset edge age to s2, create edge if necessary\n");
	if (s2Edge != 0) {
		s2Edge->fAge = 0;
	} else {
		s2Edge = new TGNGEdge(s1,s2);
		s1->fEdges.Add(s2Edge);
		s2->fEdges.Add(s2Edge);

		edgeCnt = s1->fEdges.GetEntriesFast();

		{ // report to extensions
			Int_t nrOfExt = fExtensions.GetEntriesFast();
			for (Int_t i = 0; i < nrOfExt; ++i) {
				TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
				curExt->OnNewEdge(this,s2Edge);
			}
		}	

	}


	// delete old edges
	//printf("delete old edges\n");
	for (Int_t i = 0; i < edgeCnt; ++i) {
		TGNGEdge *curEdge = static_cast<TGNGEdge*>(s1->fEdges.At(i));
		if (curEdge->fAge > (Double_t)fT_max) {

			TGNGUnit *otherUnit = (curEdge->fUnitA != s1) ? curEdge->fUnitA : curEdge->fUnitB;

			{ // report to extensions
				Int_t nrOfExt = fExtensions.GetEntriesFast();
				for (Int_t k = 0; k < nrOfExt; ++k) {
					TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(k));
					curExt->OnEdgeRemoval(this,curEdge);
				}
			}	

			s1->fEdges.RemoveAt(i);
			otherUnit->fEdges.Remove(curEdge);
			otherUnit->fEdges.Compress();
			delete curEdge;
		}		
	}
	s1->fEdges.Compress();
	edgeCnt = s1->fEdges.GetEntriesFast();


	// delete isolated units	
	//printf("delete isolated units\n");
	if (fGrowth) {
		for (Int_t i = 0; i < unitCnt; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));
			if (curUnit->fEdges.GetEntriesFast() == 0) {

				{ // report to extensions
					Int_t nrOfExt = fExtensions.GetEntriesFast();
					for (Int_t k = 0; k < nrOfExt; ++k) {
						TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(k));
						curExt->OnUnitRemoval(this,curUnit);
					}
				}	
				fUnits.RemoveAt(i);
				delete curUnit;
			}
		}
		fUnits.Compress();
		unitCnt = fUnits.GetEntriesFast();
		// update unit indices
		for (Int_t i = 0; i < unitCnt; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));
			curUnit->fCurIdx = i;
		}	
	}
	
	// update accumulated error of s1
	//printf("update accumulated error of s1\n");
	s1->fAccErr += TMath::Power(s1Dist,2.0);

	Double_t eps_b = fEpsilon_b;
	Double_t eps_n = fEpsilon_n;

	if (fUseEpsilonRamp) {

		eps_b = fEpsilon_be;
		eps_n = fEpsilon_ne;

		if ((Double_t)fInpCnt < fRampTime) {
			eps_b = fEpsilon_b * TMath::Power(fEpsilon_be / fEpsilon_b, (Double_t)fInpCnt / fRampTime);
			eps_n = fEpsilon_n * TMath::Power(fEpsilon_ne / fEpsilon_n, (Double_t)fInpCnt / fRampTime);
		}

	}

	// adapt s1 prototype
	//printf("adapt s1 prototype\n");
	if (fUseLearningInhibition) {
		s1->fPrototype->Adapt(input,eps_b * (1.0 - s1->fLearnInh));
	} else {
		s1->fPrototype->Adapt(input,eps_b);
	}

	// adapt s1 neighbor prototypes
	//printf("adapt s1 neighbor prototypes\n");

	TObjString lishkey("LISH");

	for (Int_t i = 0; i < edgeCnt; ++i) {
		TGNGEdge *curEdge = static_cast<TGNGEdge*>(s1->fEdges.At(i));
		
		Double_t lw = 1.0;
		if (fLishWeight) {
			TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&lishkey));
			if (lish != 0) {
				lw = GetEdgeLishStrength(lish);
			}			
		}
		
		if (curEdge->fUnitA != s1) {
			if (fUseLearningInhibition) {
				curEdge->fUnitA->fPrototype->Adapt(input,eps_n*lw* (1.0 - s1->fLearnInh));
			} else {
				curEdge->fUnitA->fPrototype->Adapt(input,eps_n*lw);
			}
		} else {
			if (fUseLearningInhibition) {
				curEdge->fUnitB->fPrototype->Adapt(input,eps_n*lw* (1.0 - s1->fLearnInh));
			} else {
				curEdge->fUnitB->fPrototype->Adapt(input,eps_n*lw);
			}
		}
	}

	// set learning inhibition of s1
	s1->fLearnInh = 1.0;


	{ // report to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t k = 0; k < nrOfExt; ++k) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(k));
			curExt->OnUnitUpdate(this,s1,s1Dist,s2,s2Dist,s2Edge,input);
		}
	}	

	// check if we should add a new unit
	//printf("check if we should add a new unit\n");
	if ((fInpCnt > fLambdaDelay) && ((fInpCnt % fLambda) == 0) && (fGrowth == kTRUE)) {

		{ // report to extensions
			Int_t nrOfExt = fExtensions.GetEntriesFast();
			for (Int_t i = 0; i < nrOfExt; ++i) {
				TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
				curExt->OnLambda(this,fInpCnt);
			}
		}	

		if ((UInt_t)unitCnt < fMaxUnitCnt) {

			// find unit with greatest error
			//printf("find unit with greatest error\n");
			TGNGUnit *mxErr    =  0;
			Double_t  mxErrVal = -1;
			for (Int_t i = 0; i < unitCnt; ++i) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));
				if (curUnit->fAccErr > mxErrVal) {
					mxErr    = curUnit;
					mxErrVal = curUnit->fAccErr;
				}
			}

			if (mxErr == 0) {
				RaiseError(TGNG::ecNoMaxErr,"Could not find unit with greatest error! Something is very wrong...");
				return 0;				
			}

			// find neighbor with greatest error
			//printf("find neighbor with greatest error\n");
			TGNGUnit *mxErrNb    =  0;
			Double_t  mxErrNbVal = -1;
			TGNGEdge *mxErrEdge  =  0;
			edgeCnt = mxErr->fEdges.GetEntriesFast();
			for (Int_t i = 0; i < edgeCnt; ++i) {
				TGNGEdge *curEdge   = static_cast<TGNGEdge*>(mxErr->fEdges.At(i));
				TGNGUnit *otherUnit = (curEdge->fUnitA != mxErr) ? curEdge->fUnitA : curEdge->fUnitB;
				if (otherUnit->fAccErr > mxErrNbVal) {
					mxErrNb    = otherUnit;
					mxErrNbVal = otherUnit->fAccErr;
					mxErrEdge  = curEdge;
				}
			}

			if (mxErrNb == 0) {
				RaiseError(TGNG::ecNoMaxErr,"Could not find neighbor unit with greatest error! Something is very wrong...");
				return 0;				
			}

			// create new unit between mxErr and mxErrNb
			//printf("create new unit between mxErr and mxErrNb\n");
			TGNGUnit *newUnit = new TGNGUnit(mxErr->fPrototype->Interpolate(mxErrNb->fPrototype,0.5),unitCnt);
			// delete old edge
			//printf("delete old edge\n");

			{ // report to extensions
				Int_t nrOfExt = fExtensions.GetEntriesFast();
				for (Int_t i = 0; i < nrOfExt; ++i) {
					TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
					curExt->OnEdgeRemoval(this,mxErrEdge);
				}
			}	
			mxErr->fEdges.Remove(mxErrEdge);
			mxErr->fEdges.Compress();
			mxErrNb->fEdges.Remove(mxErrEdge);
			mxErrNb->fEdges.Compress();
			delete mxErrEdge;

			// create two new edges
			//printf("create two new edges\n");
			TGNGEdge *newEdge1 = new TGNGEdge(mxErr,  newUnit);
			TGNGEdge *newEdge2 = new TGNGEdge(mxErrNb,newUnit);

			mxErr->fEdges.Add(newEdge1);
			mxErrNb->fEdges.Add(newEdge2);
			newUnit->fEdges.Add(newEdge1);
			newUnit->fEdges.Add(newEdge2);

			// reduce error of mxErr and mxErrNb
			//printf("reduce error of mxErr and mxErrNb\n");
			mxErr->fAccErr -= fAlpha * mxErr->fAccErr;
			if (mxErr->fAccErr < 0.0)
				mxErr->fAccErr = 0.0;
			mxErrNb->fAccErr -= fAlpha * mxErrNb->fAccErr;
			if (mxErrNb->fAccErr < 0.0)
				mxErrNb->fAccErr = 0.0;

			// set the error of the new unit to the reduced error of mxErr
			//printf("set the error of the new unit to the reduced error of mxErr\n");
			newUnit->fAccErr = mxErr->fAccErr;

			fUnits.Add(newUnit);
			unitCnt = fUnits.GetEntriesFast();

			{ // report to extensions
				Int_t nrOfExt = fExtensions.GetEntriesFast();
				for (Int_t i = 0; i < nrOfExt; ++i) {
					TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
					curExt->OnNewUnit(this,newUnit);
				}
			}

			{ // report to extensions
				Int_t nrOfExt = fExtensions.GetEntriesFast();
				for (Int_t i = 0; i < nrOfExt; ++i) {
					TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
					curExt->OnNewEdge(this,newEdge1);
					curExt->OnNewEdge(this,newEdge2);
				}
			}			
		}
	}

	// reduce the error of all units by beta and
	// reduce the learning inhibition of all units by gamma
	//printf("reduce the error of all units by beta\n");
	for (Int_t i = 0; i < unitCnt; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));

		curUnit->fAccErr -= fBeta * curUnit->fAccErr;
		if (curUnit->fAccErr < 0.0)
			curUnit->fAccErr = 0.0;

		curUnit->fLearnInh -= fGamma * curUnit->fLearnInh;
		if (curUnit->fLearnInh < 0.0)
			curUnit->fLearnInh = 0.0;
	}

	{ // report result to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			curExt->OnFeedResult(this,s1,s1Dist,s2,s2Dist,input);
		}
	}	



	//printf("return\n");
	return s1;
}

TGNGUnit* TGNG::Peek(const TGNGDataElement *input, Double_t *minDist, TObject *rawInput, Double_t *ratio)
{
	// Inputs a vector into the GNG and returns the BMU.
	// The GNG is not modified.

	if (fRawInput) {
		delete fRawInput;
		fRawInput = 0;
	}

	if (rawInput) {
		fRawInput = rawInput->Clone();
	}

	// finds BMU
	if (fInitialized == kFALSE) {
		RaiseError(TGNG::ecNoInit,"GNG not initialized!");
		return 0;				
	}

	if (input == 0) {
		RaiseError(TGNG::ecParamNull,"Input is null!");
		return 0;				
	}

	fInpCnt++;

	TGNGUnit *s1 = 0;
	TGNGUnit *s2 = 0;

	Double_t s1ApproxDist = std::numeric_limits<Double_t>::max();
	Double_t s2ApproxDist = std::numeric_limits<Double_t>::max();

	Int_t unitCnt = fUnits.GetEntriesFast();

	for (Int_t i = 0; i < unitCnt; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));

		Double_t curDist = fDistMeasure->GetApproxDistance(curUnit->fPrototype,input,this);

		if (curDist < s1ApproxDist) {
			s2           = s1;
			s2ApproxDist = s1ApproxDist;
			s1           = curUnit;
			s1ApproxDist = curDist;
		} else if (curDist < s2ApproxDist) {
			s2           = curUnit;
			s2ApproxDist = curDist;
		}
	}

	if ((s1 == 0) || (s2 == 0)) {
		RaiseError(TGNG::ecNoBMU,"Could not find first and second BMU! Something is very wrong...");
		return 0;				
	}

	// get real distances 
	//printf("get real distances \n");
	Double_t s1Dist = fDistMeasure->GetDistance(s1->fPrototype,input,this);
	Double_t s2Dist = fDistMeasure->GetDistance(s2->fPrototype,input,this);

	Int_t nrOfExt = fExtensions.GetEntriesFast();
	for (Int_t i = 0; i < nrOfExt; ++i) {
		TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
		curExt->OnPeekResult(this,s1,s1Dist,s2,s2Dist,input);
	}	

	if (minDist) *minDist = s1Dist;

	if (ratio) {
		Double_t relDist = fDistMeasure->GetDistance(s1->fPrototype,s2->fPrototype,this);

		Double_t rtmp = 1.0;
		if ((TMath::IsNaN(relDist) == kFALSE) && (relDist != 0)) {
			rtmp = ((s1Dist - s2Dist) / relDist) + 1.0;
		}

		if (rtmp > 1.0) rtmp = 1.0;
		if (rtmp < 0.0) rtmp = 0.0;		

		*ratio = rtmp;
	}

	return s1;
}

void TGNG::SetEpsilon_b(Double_t val)
{
	fEpsilon_b = val;
}

void TGNG::SetEpsilon_be(Double_t val)
{
	fEpsilon_be = val;
}

void TGNG::SetEpsilon_n(Double_t val)
{
	fEpsilon_n = val;
}

void TGNG::SetEpsilon_ne(Double_t val)
{
	fEpsilon_ne = val;
}

void TGNG::SetRampTime(Double_t val)
{
	fRampTime = val;
}

void TGNG::SetLambda(UInt_t val)
{
	fLambda = val;
}

void TGNG::SetT_max(UInt_t val)
{
	fT_max = val;
}

void TGNG::SetAlpha(Double_t val)
{
	fAlpha = val;
}

void TGNG::SetBeta(Double_t val)
{
	fBeta = val;
}

void TGNG::SetMaxUnitCnt(UInt_t val)
{
	fMaxUnitCnt = val;
}

void TGNG::SetLambdaDelay(UInt_t val)
{
	fLambdaDelay = val;
}

void TGNG::SetParams(Double_t eps_b, Double_t eps_n, UInt_t lambda, UInt_t tmax, Double_t alpha, Double_t beta, UInt_t maxCnt)
{
	fEpsilon_b  = eps_b;
	fEpsilon_n  = eps_n;
	fLambda     = lambda;
	fT_max      = tmax;
	fAlpha      = alpha;
	fBeta       = beta;
	fMaxUnitCnt = maxCnt;
}

Double_t TGNG::GetEpsilon_b()
{
	return fEpsilon_b;
}

Double_t TGNG::GetEpsilon_be()
{
	return fEpsilon_be;
}

Double_t TGNG::GetEpsilon_n()
{
	return fEpsilon_n;
}

Double_t TGNG::GetEpsilon_ne()
{
	return fEpsilon_ne;
}

Double_t TGNG::GetRampTime()
{
	return fRampTime;
}

UInt_t TGNG::GetLambda()
{
	return fLambda;
}

UInt_t TGNG::GetT_max()
{
	return fT_max;
}

Double_t TGNG::GetAlpha()
{
	return fAlpha;
}

Double_t TGNG::GetBeta()
{
	return fBeta;
}

TGNGDistanceMeasure* TGNG::GetDistanceMeasure() 
{
	// Returns a reference to the currently used distance measure

	return fDistMeasure; 
} 

UInt_t TGNG::GetMaxUnitCnt()
{
	return fMaxUnitCnt;
}

UInt_t TGNG::GetLambdaDelay()
{
	return fLambdaDelay;
}

ULong64_t TGNG::GetInputCount()
{
	// Returns the number of inputs seen by the GNG so far.
	// Inputs provided to Peek() are not counted.

	return fInpCnt;
}

void TGNG::AddExtension(TGNGExtension *ext)
{
	// Add the extension "ext" to the GNG.
	// The GNG will "own" the extension, i.e., it will manage the memory

	fExtensions.Add(ext);
}

void TGNG::RemoveExtension(TGNGExtension *ext)
{
	// Remove the extension from the GNG.
	// The corresponding memory will no longer be managed by the GNG.

	fExtensions.Remove(ext);
	fExtensions.Compress();
}

void TGNG::RaiseError(UInt_t errorCode, const char* errMsg)
{
	// Method that should be used by extensions to inform about errors.

	Int_t nrOfExt = fExtensions.GetEntriesFast();
	for (Int_t i = 0; i < nrOfExt; ++i) {
		TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
		curExt->OnError(this,errorCode,errMsg);
	}		
}

TMatrixD* TGNG::GetAdjacencyMatrix()
{
	// Returns an adjacency matrix of the GNG units.
	// The returned matrix is owned by the caller.

	Int_t nrOfUnits = fUnits.GetEntriesFast();
	if (nrOfUnits <= 0)
		return 0;

	TMatrixD *result = new TMatrixD(nrOfUnits,nrOfUnits);
	result->Zero();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));
		Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
		for (Int_t j = 0; j < edgeCnt; ++j) {
			TGNGEdge *curEdge   = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
			TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;
			(*result)(i,fUnits.IndexOf(otherUnit)) = 1.0;
		}
	}

	return result;
}

TMatrixD* TGNG::GetDistanceMatrix()
{
	// Returns a distance matrix describing the relative distances between the 
	// units prototypes.
	// The returned matrix is owned by the caller.

	Int_t nrOfUnits = fUnits.GetEntriesFast();
	if (nrOfUnits <= 0)
		return 0;

	TMatrixD *result = new TMatrixD(nrOfUnits,nrOfUnits);

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(fUnits.At(i));
		for (Int_t j = 0; j < nrOfUnits; ++j) {
			TGNGUnit *otherUnit = static_cast<TGNGUnit*>(fUnits.At(j));
			
			(*result)(i,j) = fDistMeasure->GetDistance(curUnit->fPrototype,otherUnit->fPrototype,this);
		}
	}

	return result;

}

TH1D* TGNG::GetDistanceDistribution()
{
	// Returns a histogram of all relative distances between units in the GNG
	// The returned histogram is owned by the caller.

	TMatrixD *dm = GetDistanceMatrix();

	if (dm == 0) return 0;

	Double_t maxVal = dm->Max();

	Int_t nrOfUnits = fUnits.GetEntriesFast();

	TH1D *result = new TH1D("GNGDistHist",Form("GNG distance distribution for %s",fDistMeasure->ClassName()),nrOfUnits,0,maxVal+(maxVal / (Double_t)nrOfUnits));
	result->SetDirectory(0);

	for (Int_t i = 0; i < nrOfUnits; ++i)
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if (i == j) continue;

			result->Fill((*dm)(i,j));

		}

	delete dm;

	return result;
}

TH1D* TGNG::GetNbDistanceDistribution()
{
	// Returns a histogram of all relative distances between units that are
	// connected by an edge in the GNG.
	// The returned histogram is owned by the caller.

	TMatrixD *dm = GetDistanceMatrix();
	TMatrixD *am = GetAdjacencyMatrix();

	if ((dm == 0) || (am == 0)) {
		if (dm) delete dm;
		if (am) delete am;
		return 0;
	}

	Double_t maxVal = dm->Max();

	Int_t nrOfUnits = fUnits.GetEntriesFast();

	TH1D *result = new TH1D("GNGNBDistHist",Form("GNG neighbor distance distribution for %s",fDistMeasure->ClassName()),nrOfUnits,0,maxVal+(maxVal / (Double_t)nrOfUnits));
	result->SetDirectory(0);

	for (Int_t i = 0; i < nrOfUnits; ++i)
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if (i == j) continue;

			if ((*am)(i,j) < 0.5) continue;

			result->Fill((*dm)(i,j));

		}

	delete dm;
	delete am;

	return result;
}

void TGNG::FillDistanceDistribution(TH2D *hist, Double_t xVal)
{
	// Fills in the distance distribution of all relative distances between 
	// units in the GNG at position xVal in the provided 2D histogram.

	TMatrixD *dm = GetDistanceMatrix();

	if (dm == 0) return;

	Double_t maxVal = dm->Max();

	Int_t nrOfUnits = fUnits.GetEntriesFast();

	for (Int_t i = 0; i < nrOfUnits; ++i)
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if (i == j) continue;

			hist->Fill(xVal,(*dm)(i,j));

		}

	delete dm;
}

void TGNG::FillNbDistanceDistribution(TH2D *hist, Double_t xVal)
{
	// Fills in the distance distribution of all relative distances between 
	// units that are connected by an edge in the GNG at position xVal in the 
	// provided 2D histogram.

	TMatrixD *dm = GetDistanceMatrix();
	TMatrixD *am = GetAdjacencyMatrix();

	if ((dm == 0) || (am == 0)) {
		if (dm) delete dm;
		if (am) delete am;
		return;
	}

	Double_t maxVal = dm->Max();

	Int_t nrOfUnits = fUnits.GetEntriesFast();

	for (Int_t i = 0; i < nrOfUnits; ++i)
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if (i == j) continue;

			if ((*am)(i,j) < 0.5) continue;

			hist->Fill(xVal,(*dm)(i,j));

		}

	delete dm;
	delete am;
}


TH1D* TGNG::GetGradDistribution()
{
	// Returns a histogram of all unit grades (number of edges incident to a 
	// unit) in the GNG.
	// The returned histogram is owned by the caller.

	TMatrixD *am = GetAdjacencyMatrix();

	if (am == 0) return 0;

	Int_t nrOfUnits = fUnits.GetEntriesFast();

	TH1D *result = new TH1D("GNGGradHist",Form("GNG grad distribution for %s",fDistMeasure->ClassName()),nrOfUnits+1,-0.5,(Double_t)nrOfUnits+0.5);
	result->SetDirectory(0);

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		Double_t grad = 0;
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if ((*am)(i,j) > 0.5)
				grad += 1.0;

		}
		result->Fill(grad);
	}

	delete am;

	return result;	
}

void TGNG::FillGradDistribution(TH2D *hist, Double_t xVal)
{
	// Fills the distribution of units grades (number of edges incident to a 
	// unit) in the GNG at position xVal in the provided 2D histogram.

	TMatrixD *am = GetAdjacencyMatrix();

	if (am == 0) return;

	Int_t nrOfUnits = fUnits.GetEntriesFast();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		Double_t grad = 0;
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if ((*am)(i,j) > 0.5)
				grad += 1.0;

		}
		hist->Fill(xVal,grad);
	}

	delete am;	
}


Int_t TGNG::GetNrOfExtensions()
{
	// returns the number of registered extensions

	return fExtensions.GetEntriesFast();
}

TGNGExtension* TGNG::GetExtension(Int_t idx)
{
	// returns the extension with index idx

	return static_cast<TGNGExtension*>(fExtensions.At(idx));
}

TGNGExtension* TGNG::GetExtension(const char *classname)
{
	// returns the first extension with a matching "classname".

	TString sTerm(classname);
	Int_t nrOfExt = fExtensions.GetEntriesFast();
	for (Int_t i = 0; i < nrOfExt; ++i) {
		TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
		TString curName(curExt->ClassName());
		if (curName == sTerm)
			return curExt;
	}	
	return 0;
}

void TGNG::Print(Option_t* option) const
{
	// print information about the gng
	// option "g" prints gng information (default)
	// option "e" prints extension information
	// option "p" prints all gng parameters

	TString opt = option;
	opt.ToLower();

	printf("\n");
	
	if ((opt.Length() == 0) || (opt.Contains("g"))) {
		// print general gng information
		printf("Growing Neural Gas\n\n");
		printf("Status:   %s\n", fInitialized ? "initialized" : "not initialized");
		printf("# Units:  %d\n",fUnits.GetEntriesFast());
		printf("# Inputs: %lld\n\n",fInpCnt);
		printf("Distance Measure: %s\n", (fDistMeasure != 0) ? fDistMeasure->ClassName() : "null");
		TGNGDataElement *elem = 0;
		if (fInitialized) {
			TGNGUnit *firstUnit = static_cast<TGNGUnit*>(fUnits.At(0));
			if (firstUnit == 0) {
				const_cast<TGNG*>(this)->RaiseError(TGNG::ecFatal,"no first unit although initialized!");
				return;
			}
			elem = firstUnit->fPrototype;
		}
		printf("Data Elements:    %s\n\n", (elem != 0) ? elem->ClassName() : "null");
	}

	if (opt.Contains("e")) {		
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		printf("Number of extensions: %d\n", nrOfExt);
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			printf("Extension %d: %s\n", i, curExt->ClassName());
		}	
		printf("\n");			
	}

	if (opt.Contains("p")) {
		printf("Epsilon b:  %f (adaptation of BMU)\n", fEpsilon_b);
		printf("Epsilon n:  %f (adaptation of neighbors)\n", fEpsilon_n);
		printf("Lambda:     %d (new unit interval)\n", fLambda);
		printf("Alpha:      %f (decrease of accumulated error by insertion)\n", fAlpha);
		printf("Beta:       %f (general decrease of accumulated error)\n", fBeta);
		printf("MaxUnitCnt: %d (max. number of units)\n\n", fMaxUnitCnt);
	}


}

void TGNG::SetLishWeighting(Bool_t enable)
{
	fLishWeight = enable;
}

void TGNG::SetLishWeightCenter(Double_t center)
{
	fLishWeightCenter = center;
}

void TGNG::SetLishWeightSteepness(Double_t steepness)
{
	fLishWeightSteepness = steepness;
}

Bool_t TGNG::GetLishWeighting()
{
	return fLishWeight;
}

Double_t TGNG::GetLishWeightCenter()
{
	return fLishWeightCenter;
}

Double_t TGNG::GetLishWeightSteepness()
{
	return fLishWeightSteepness;
}

Double_t TGNG::GetEdgeLishStrength(TH1D *hist)
{
	Int_t lishSize = hist->GetNbinsX();
	Double_t cnt = 0;
	Double_t avg = 0;
	for (Int_t l = 1; l <= lishSize; ++l) {
		Double_t bc = 0;
		if ((bc = hist->GetBinContent(l)) > 0) {
			avg += hist->GetBinError(l) / bc;
			cnt += 1;
		} else {
			avg += 1;
			cnt += 1;
		}
	}
	if (cnt > 0) {
		avg /= cnt;
	} else {
		avg = 1;
	}

	return 1.0 - 1.0 / TMath::Power(1.0 + 0.5 * TMath::Exp(-fLishWeightSteepness * (avg - fLishWeightCenter)) ,2) ;

}

TObject* TGNG::GetLastRawInput()
{
	return fRawInput;
}

void TGNG::SetLearningInhibition(Bool_t enable)
{
	fUseLearningInhibition = enable;
}

Bool_t TGNG::GetLearningInhibition()
{
	return fUseLearningInhibition;
}

void TGNG::SetGamma(Double_t val)
{
	fGamma = val;
}

Double_t TGNG::GetGamma()
{
	return fGamma;
}

void TGNG::SetEpsilonRamp(Bool_t enable)
{
	fUseEpsilonRamp = enable;
}

Bool_t TGNG::GetEpsilonRamp()
{
	return fUseEpsilonRamp;
}

void TGNG::SetGrowth(Bool_t enable)
{
	fGrowth = enable;
}

Bool_t TGNG::GetGrowth()
{
	return fGrowth;
}
