/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of the TGNGRecursiveData class. This class 
represents a data type that itself is a TGNG. 

It is used to build an RGNG which can, e.g., be used for the simulation of
grid cells.

*/
#include <limits>

#include <TObjArray.h>
#include <TMath.h>

#include "../core/TGNGUnit.h"
#include "../core/TGNGEdge.h"
#include "../core/TGNGDistanceMeasure.h"

#include "TGNGRecursiveData.h"

TGNGRecursiveData::TGNGRecursiveData() :
	TGNGDataElement(),
	fGNG(),
	fNbAttenuation(0.1)
{
}

TGNGRecursiveData::TGNGRecursiveData(TGNGDataElement *prototypeA, 
	                                 TGNGDataElement *prototypeB, 
	                                 TGNGDistanceMeasure *distMeasure) :
	TGNGDataElement(),
	fGNG()
{
	fGNG.Init(prototypeA,prototypeB,distMeasure);
}

TGNGRecursiveData::~TGNGRecursiveData()
{
}

TGNGDataElement* TGNGRecursiveData::Interpolate(const TGNGDataElement *other, Double_t ratio)
{
	// Needs to perform an interpolation between two GNGs

	const TGNGRecursiveData *otherRD = dynamic_cast<const TGNGRecursiveData*>(other);

	if (otherRD == 0) {
		printf("Severe Error! Data used for Interpolate() is not of type TGNGRecursiveData.\n");
		return 0;
	}

	// there is not "the" way to interpolate two GNGs, so we have to improvise a bit

	// create new recursive data
	TGNGRecursiveData *result = new TGNGRecursiveData();

	// manually initialise the fGNG member
	TGNGDistanceMeasure *dm = fGNG.GetDistanceMeasure();
	result->fGNG.ManualInit(dm);

	result->fGNG.SetEpsilon_b(fGNG.GetEpsilon_b());
	result->fGNG.SetEpsilon_be(fGNG.GetEpsilon_be());
	result->fGNG.SetEpsilon_n(fGNG.GetEpsilon_n());
	result->fGNG.SetEpsilon_ne(fGNG.GetEpsilon_ne());
	result->fGNG.SetRampTime(fGNG.GetRampTime());
	result->fGNG.SetLambda(fGNG.GetLambda());
	result->fGNG.SetT_max(fGNG.GetT_max());
	result->fGNG.SetAlpha(fGNG.GetAlpha());
	result->fGNG.SetBeta(fGNG.GetBeta());
	result->fGNG.SetMaxUnitCnt(fGNG.GetMaxUnitCnt());
	result->fGNG.SetLambdaDelay(fGNG.GetLambdaDelay());
	result->fGNG.SetGamma(fGNG.GetGamma());

	result->fGNG.SetLearningInhibition(fGNG.GetLearningInhibition());
	result->fGNG.SetEpsilonRamp(fGNG.GetEpsilonRamp());

	result->SetNbAttenuation(GetNbAttenuation());


	result->fGNG.SetLishWeighting(fGNG.GetLishWeighting());
	result->fGNG.SetLishWeightCenter(fGNG.GetLishWeightCenter());
	result->fGNG.SetLishWeightSteepness(fGNG.GetLishWeightSteepness());

	// copy extensions
	
	Int_t extCnt = fGNG.GetNrOfExtensions();
	for (Int_t i = 0; i < extCnt; ++i) {
		TGNGExtension *curExt  = fGNG.GetExtension(i);
		TGNGExtension *extCopy = dynamic_cast<TGNGExtension*>(curExt->Clone());
		result->fGNG.AddExtension(extCopy);
	}


	// generate interpolated units
	Int_t unitCntA = fGNG.fUnits.GetEntriesFast();
	Int_t unitCntB = otherRD->fGNG.fUnits.GetEntriesFast();
	TObjArray *unitsA = &(fGNG.fUnits);
	TObjArray *unitsB = const_cast<TObjArray*>(&(otherRD->fGNG.fUnits));

	Int_t unitCntC    = result->fGNG.fUnits.GetEntriesFast();
	TObjArray *unitsC = &(result->fGNG.fUnits);

	// place higher number of units in "unitsA"
	if (unitCntA < unitCntB) {
		Int_t      tmpCnt = unitCntB; unitCntB = unitCntA; unitCntA = tmpCnt;
		TObjArray *tmpArr = unitsB;     unitsB = unitsA;     unitsA = tmpArr;
	}

	for (Int_t i = 0; i < unitCntA; ++i) {
		TGNGUnit *curUnitA = static_cast<TGNGUnit*>(unitsA->At(i));
		TGNGUnit *minOther = 0;
		Double_t minDist = numeric_limits<Double_t>::max();
		for (Int_t j = 0; j < unitCntB; ++j) {
			TGNGUnit *curUnitB = static_cast<TGNGUnit*>(unitsB->At(j));
			Double_t curDist = dm->GetDistance(curUnitA->fPrototype,curUnitB->fPrototype,&fGNG);
			if (curDist < minDist) {
				minDist = curDist;
				minOther = curUnitB;
			}
			/*
			if (TMath::IsNaN(curDist)) {
				printf("curDist is NaN\n");
			}
			*/
		}
		if (minOther == 0) {
			printf("Severe Error! Could not find matching unit for Interpolate().\n");
			return 0;			
		}

		TGNGUnit *newUnit = new TGNGUnit(curUnitA->fPrototype->Interpolate(minOther->fPrototype,ratio),unitCntC++);
		unitsC->Add(newUnit);
	}

	// sanity check
	if (unitCntC != unitCntA) {
		printf("Severe Error! Number of interpolated units mismatch.\n");
		return 0;					
	}

	// establish edges
	for (Int_t i = 0; i < unitCntA; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(unitsA->At(i));
		Int_t nrOfEdges = curUnit->fEdges.GetEntriesFast();
		for (Int_t j = 0; j < nrOfEdges; ++j) {
			TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
			TGNGUnit *otherUnit = curEdge->fUnitA;
			if (otherUnit == curUnit) otherUnit = curEdge->fUnitB;

			// get corresponding units of new gng
			TGNGUnit *unitA = static_cast<TGNGUnit*>(unitsC->At(curUnit->fCurIdx));
			TGNGUnit *unitB = static_cast<TGNGUnit*>(unitsC->At(otherUnit->fCurIdx));

			// check if there is already an edge between the units
			Bool_t foundEdge = kFALSE;
			Int_t uaNrOfEdges = unitA->fEdges.GetEntriesFast();
			for (Int_t k = 0; k < uaNrOfEdges; ++k) {
				curEdge = static_cast<TGNGEdge*>(unitA->fEdges.At(k));
				otherUnit = curEdge->fUnitA;
				if (otherUnit == unitA) otherUnit = curEdge->fUnitB;
				if (otherUnit == unitB) {
					foundEdge = kTRUE;
					break;
				}
			}

			if (foundEdge == kTRUE)
				continue;

			// create edge
			TGNGEdge *newEdge = new TGNGEdge(unitA, unitB);
			unitA->fEdges.Add(newEdge);
			unitB->fEdges.Add(newEdge);
		}
	}	

	// finish manual initialization
	result->fGNG.ManualInitDone();	

	return result;
}

void TGNGRecursiveData::Adapt(const TGNGDataElement *other, Double_t factor)
{
	// feeds "other" into fGNG setting epsilon_b = factor and 
	// epsilon_n = factor fNbAttenuation


	Double_t eb = fGNG.GetEpsilon_b();
	Double_t en = fGNG.GetEpsilon_n();

	Bool_t li = fGNG.GetLearningInhibition();
	Bool_t er = fGNG.GetEpsilonRamp();

	fGNG.SetEpsilon_b(factor);
	fGNG.SetEpsilon_n(factor * fNbAttenuation);

	fGNG.SetLearningInhibition(kFALSE);
	fGNG.SetEpsilonRamp(kFALSE);

	fGNG.Feed(other);

	fGNG.SetEpsilon_b(eb);
	fGNG.SetEpsilon_n(en);
	fGNG.SetLearningInhibition(li);
	fGNG.SetEpsilonRamp(er);

	
}

void TGNGRecursiveData::SetNbAttenuation(Double_t factor)
{
	fNbAttenuation = factor;
}

Double_t TGNGRecursiveData::GetNbAttenuation()
{
	return fNbAttenuation;
}
