/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of the TGNGVectorData class. This class 
is the general purpose 1D vector used as input element to the GNG. The vector
data is stored in the member fVector which is of the ROOT type TVectorD.
*/
#include <TRandom.h>
#include <TMath.h>
#include <TStyle.h>

#include "TGNGVectorData.h"

TGNGVectorData::TGNGVectorData() :
	TGNGDataElement(),
	TAttFill(),
	fVector(),
	fPayload(0)
{
}

TGNGVectorData::TGNGVectorData(Int_t dim, TObject *payload) :
	TGNGDataElement(),
	TAttFill(),
	fVector(dim),
	fPayload(payload)
{
	// the vector size is set to "dim" and its fields are initialized with
	// gRandom->Uniform();
	//
	// the Payload is set to "payload". Note that the payload is deleted upon
	// destruction of the TGNGVectorData object

	for (Int_t i = 0; i < dim; ++i)
		fVector[i] = gRandom->Uniform();
}

TGNGVectorData::~TGNGVectorData()
{
	// deletes fPayload if it was set

	if (fPayload)
		delete fPayload;
}

TGNGDataElement* TGNGVectorData::Interpolate(const TGNGDataElement *other, Double_t ratio)
{
	// Performs a linear interpolation between this vector and "other" by 
	// "ratio", i.e.: (ratio * fVector) + ((1.0 - ratio) * otherVector->fVector)

	const TGNGVectorData *otherVector = dynamic_cast<const TGNGVectorData*>(other);

	if (otherVector == 0) {
		printf("Severe Error! Vector used for Interpolate() is not a TGNGVectorData.\n");
		return 0;
	}

	if (otherVector->fVector.GetNrows() != fVector.GetNrows()) {
		printf("Severe Error! Vector used for Interpolate() has different dimension.\n");
		return 0;		
	}

	TGNGVectorData *result = new TGNGVectorData(fVector.GetNrows());

	result->fVector = (ratio * fVector) + ((1.0 - ratio) * otherVector->fVector);

	return result;
}

void TGNGVectorData::Adapt(const TGNGDataElement *other, Double_t factor)
{
	// adapts this vector towards "other" by "factor", i.e.:
	// fVector += factor * (otherVector->fVector - fVector);

	const TGNGVectorData *otherVector = dynamic_cast<const TGNGVectorData*>(other);

	if (otherVector == 0) {
		printf("Severe Error! Vector used for Adapt() is not a TGNGVectorData.\n");
		return;
	}	

	if (otherVector->fVector.GetNrows() != fVector.GetNrows()) {
		printf("Severe Error! Vector used for Interpolate() has different dimension.\n");
		return;		
	}

	fVector += factor * (otherVector->fVector - fVector);

}

void TGNGVectorData::PaintElement(TVirtualPad *pad, Double_t xlow, Double_t ylow, Double_t xup, Double_t yup, std::vector<Color_t> *colors)
{
	// paints this vector into the pad as a column of colored boxes with each 
	// box representing one entry of the vector colored according to the value
	// of that entry. Colors are scaled from fVector.Min() to fVector.Max(); 
	// and selected from the palette provided in "colors".
	
	SetFillStyle(1001);

	Double_t maxDist  = fVector.Max();
	Double_t minDist  = fVector.Min();

	Double_t range = maxDist - minDist;	

	//Int_t palSize = gStyle->GetNumberOfColors();
	Int_t palSize = colors->size();
	Double_t colScale = range / (Double_t)(palSize);
	if (colScale <= 0)
		colScale = 1;

	Double_t ph = yup - ylow;

	Double_t yStep = ph / (Double_t)(fVector.GetNrows());

	Int_t vecSize = fVector.GetNrows();
	for (Int_t i = 0; i < vecSize; ++i) {
		Double_t curVal = fVector[i];

		Int_t colorIdx = TMath::Floor(((curVal-minDist) / colScale) + 0.5);
		if (colorIdx >= palSize)
			colorIdx = palSize-1;

		if (colorIdx < 0)
			colorIdx = 0;

		//SetFillColor(gStyle->GetColorPalette(colorIdx));
		SetFillColor(colors->at(colorIdx));
		TAttFill::Modify();

		pad->PaintBox(xlow, yup - ( (Double_t)(i+1) * yStep ),
			          xup,  yup - ( (Double_t)(i)   * yStep ) );
	}
}

void TGNGVectorData::Randomize()
{
	Int_t dim = fVector.GetNrows();
	for (Int_t i = 0; i < dim; ++i)
		fVector[i] = gRandom->Uniform();
}
