/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a simple clustering approach that uses
the average bin error of the local input space histograms (LISH) to determine, 
if an edge in the gng should be traversed, i.e., if the gng units connected by 
this edge should belong to the same cluster.

The bin error of a LISH bin is calculated as the square root of the bin content
divided by the bin content. If the bin content is zero, the bin error is set to
1. The average bin error of a LISH is the arithmetic mean of all bin errors.

The average bin error is converted into an edge strength by a logistic function:

strength = 1.0 - 1.0 / TMath::Power(1.0 + 0.5 * TMath::Exp(-fLishWeightSteepness * (avg - fLishWeightCenter)) ,2) ;

with fLishWeightSteepness (default: 25.0) the steepness of the logistic function
and fLishWeightCenter (default: 0.25) the logistic function's center.

Finally, this strength value is compared to a threshold value "fClusterThres" 
(default:0.5). If the strength is greater or equal, the edge is traversed and
the two corresponding gng units are assigned to the same cluster.

*/
#include <vector>

#include <TH1D.h>
#include <TObjString.h>
#include <TObjArray.h>
#include <TMath.h>

#include "../../core/TGNGEdge.h"
#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../core/TGNGDataElement.h"

#include "TGNGSimpleClusterExt.h"

TGNGSimpleClusterExt::TGNGSimpleClusterExt() : 
	TGNGExtension(),
	fAttrKey("SCINFO"),
	fColors(),
	fTextColors(),
	fOutlierColor(kGray+1),
	fUnprocessedColor(kGray+2),
	fNextFreeID(1),
	fInterval(10000),
	fInpCnt(0),
	fLishWeightCenter(0.25),
	fLishWeightSteepness(25.0),
	fClusterThres(0.5)
{
	InitColors();
}

void TGNGSimpleClusterExt::OnGNGReset(TGNG*)
{
	fInpCnt = 0;
}

void TGNGSimpleClusterExt::OnFeedResult(TGNG *gng, TGNGUnit*, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	if (++fInpCnt % fInterval != 0) 
		return;


	// reset all units
	fNextFreeID = 1;
	Int_t nrOfUnits = gng->fUnits.GetEntriesFast();
	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(gng->fUnits.At(i));
		TGNGClusterInfo *cInfo = static_cast<TGNGClusterInfo*>(curUnit->fAttributes.GetValue(&fAttrKey));
		if (cInfo == 0) {
			gng->RaiseError(TGNG::ecExtension,"Unit has no attribute \"SCINFO\"!");
			continue;
		}
		cInfo->fClusterID        = -1;
		cInfo->fClusterColor     = fUnprocessedColor;
		cInfo->fClusterTextColor = kBlack;
	}

	TObjString lishkey("LISH");

	// start clustering
	std::vector<TGNGUnit*> candidateUnits;
	std::vector<TGNGClusterInfo*> candidateCIs;

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(gng->fUnits.At(i));
		TGNGClusterInfo *cInfo = static_cast<TGNGClusterInfo*>(curUnit->fAttributes.GetValue(&fAttrKey));
		if (cInfo == 0) {
			gng->RaiseError(TGNG::ecExtension,"Unit has no attribute \"SCINFO\"!");
			continue;
		}
		if (cInfo->fClusterID > -1)
			continue;

		Int_t currentClusterID = fNextFreeID++;
		Int_t clusterSize = 0;

		candidateUnits.push_back(curUnit);
		candidateCIs.push_back(cInfo);
		while (candidateUnits.empty() == false) {
			curUnit = candidateUnits.back(); candidateUnits.pop_back();
			cInfo   = candidateCIs.back(); candidateCIs.pop_back();

			cInfo->fClusterID        = currentClusterID;
			cInfo->fClusterColor     = fColors[currentClusterID % fColors.size()];
			cInfo->fClusterTextColor = fTextColors[currentClusterID % fColors.size()];

			clusterSize++;

			Int_t nrOfEdges = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < nrOfEdges; ++j) {
				TGNGEdge *curEdge   = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
				TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;
				// get lish
				TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&lishkey));
				if (lish == 0) {
					gng->RaiseError(TGNG::ecExtension,"Edge has no attribute \"LISH\"!");
					continue;
				}
				// get edge strength based on avg bin error
				Int_t lishSize = lish->GetNbinsX();
				Double_t cnt = 0;
				Double_t avg = 0;
				for (Int_t l = 1; l <= lishSize; ++l) {
					Double_t bc = 0;
					if ((bc = lish->GetBinContent(l)) > 0) {
						avg += lish->GetBinError(l) / bc;
						cnt += 1;
					} else {
						// test
						avg += 1;
						cnt += 1;
					}
				}
				if (cnt > 0) {
					avg /= cnt;
				} else {
					avg = 1;
				}

				// generalised logistic function as mapping
				Double_t strength = 1.0 - 1.0 / TMath::Power(1.0 + 0.5 * TMath::Exp(-fLishWeightSteepness * (avg - fLishWeightCenter)) ,2) ;

				if (strength < fClusterThres)
					continue;

				// traverse edge
				TGNGClusterInfo *cInfo2 = static_cast<TGNGClusterInfo*>(otherUnit->fAttributes.GetValue(&fAttrKey));	
				if (cInfo2 == 0) {
					gng->RaiseError(TGNG::ecExtension,"Unit has no attribute \"SCINFO\"!");
					continue;
				}

				if (cInfo2->fClusterID > -1) 
					continue;

				candidateUnits.push_back(otherUnit);
				candidateCIs.push_back(cInfo2);		
			}

		}

		if (clusterSize == 1) {
			cInfo->fClusterColor     = fOutlierColor;
			cInfo->fClusterTextColor = kBlack;
			cInfo->fClusterID        = 0;
			fNextFreeID--;
		}

	}
}

void TGNGSimpleClusterExt::OnNewUnit(TGNG *gng, TGNGUnit *newUnit)
{
	if (newUnit->fAttributes.FindObject(&fAttrKey) != 0) {
		gng->RaiseError(TGNG::ecExtension,"Unit has already an attribute \"SCINFO\"!");
		return;
	}
	newUnit->fAttributes.Add(new TObjString(fAttrKey),new TGNGClusterInfo());
}

void TGNGSimpleClusterExt::InitColors()
{
	fColors.push_back(kGray);
	fTextColors.push_back(kBlack);

	fColors.push_back(kGreen);
	fTextColors.push_back(kBlack);

	fColors.push_back(kRed);
	fTextColors.push_back(kBlack);

	fColors.push_back(kBlue);
	fTextColors.push_back(kWhite);

	fColors.push_back(kOrange);
	fTextColors.push_back(kBlack);

	fColors.push_back(kMagenta);
	fTextColors.push_back(kBlack);

	fColors.push_back(kOrange+1);
	fTextColors.push_back(kBlack);

	fColors.push_back(kMagenta-5);
	fTextColors.push_back(kBlack);

	fColors.push_back(kBlue-6);
	fTextColors.push_back(kBlack);

	fColors.push_back(kCyan-6);
	fTextColors.push_back(kBlack);

	fColors.push_back(kAzure+6);
	fTextColors.push_back(kBlack);

	fColors.push_back(kOrange-9);
	fTextColors.push_back(kBlack);

	fColors.push_back(kYellow);
	fTextColors.push_back(kBlack);

	fColors.push_back(kYellow-6);
	fTextColors.push_back(kBlack);

	fColors.push_back(kSpring+6);
	fTextColors.push_back(kBlack);
}

void TGNGSimpleClusterExt::SetLishWeightCenter(Double_t center)
{
	fLishWeightCenter = center;
}

void TGNGSimpleClusterExt::SetLishWeightSteepness(Double_t steepness)
{
	fLishWeightSteepness = steepness;
}

void TGNGSimpleClusterExt::SetClusterThres(Double_t thres)
{
	fClusterThres = thres;
}

void TGNGSimpleClusterExt::SetInterval(Int_t interval)
{
	fInterval = interval;
}
