/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of the local input space histogram extension.
Register this extension with a gng to enable local input space histograms for
the particular gng.

You can use the static functions GetAvgBinErrorDist() and FillAvgBinErrorDist()
to analyse the LISH histograms present in a gng.
*/
#include <TH1D.h>
#include <TMath.h>

#include "../../core/TGNGDataElement.h"
#include "../../core/TGNGDistanceMeasure.h"
#include "../../core/TGNGEdge.h"
#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"


#include "TGNGLishExt.h"

TGNGLishExt::TGNGLishExt() :
	TGNGExtension(),
	fAttrKey("LISH"),
	fLishSize(16)
{
	// default constructor setting the LISH size to 16.
}

TGNGLishExt::TGNGLishExt(UInt_t ls) :
	TGNGExtension(),
	fAttrKey("LISH"),
	fLishSize(ls)
{
	// constructor setting the LISH size to "ls".
}

void TGNGLishExt::OnNewEdge(TGNG *gng, TGNGEdge *newEdge)
{
	if (newEdge->fAttributes.FindObject(&fAttrKey) != 0) {
		gng->RaiseError(TGNG::ecExtension,"Edge has already an attribute \"LISH\"!");
		return;
	}
	TH1D *newLish = new TH1D("lish","local input space histogram",fLishSize,0.0,1.0);
	newLish->SetDirectory(0);

	newEdge->fAttributes.Add(new TObjString(fAttrKey),newLish);
}

void TGNGLishExt::OnUnitUpdate(TGNG                  *gng, 
	                           TGNGUnit              *bmu,
	                           Double_t               bmuDist,
	                           TGNGUnit              *bmu2,
	                           Double_t               bmu2Dist,
	                           TGNGEdge              *bmuEdge,
	                           const TGNGDataElement* )
{
	// get lish
	TH1D *lish = static_cast<TH1D*>(bmuEdge->fAttributes.GetValue(&fAttrKey));
	if (lish == 0) {
		gng->RaiseError(TGNG::ecExtension,"Edge has no attribute \"LISH\"!");
		return;		
	}

	TGNGDistanceMeasure *dm = gng->GetDistanceMeasure();

	Double_t relDist = dm->GetDistance(bmu->fPrototype,bmu2->fPrototype,gng);

	if (TMath::IsNaN(relDist) == kTRUE) {
		gng->RaiseError(TGNG::ecExtension,"Distance between prototypes is NaN!");
		return;				
	}

	if (relDist == 0.0) {
		lish->Fill(0.0,0.5);
		lish->Fill(1.0,0.5);
		return;
	}

	Double_t ratio = (((bmuDist - bmu2Dist) / relDist) + 1.0) / 2.0;

	// update lish
	if (bmuEdge->fUnitA == bmu) {
		lish->Fill(ratio);
	} else {
		lish->Fill(1.0 - ratio);
	}
}



TH1D* TGNGLishExt::GetAvgBinErrorDist(TGNG *gng)
{
	// this static function returns the distribution of the average bin errors
	// exhibited by the LISHs in the GNG "gng".

	TH1D *result = new TH1D("avgBinErrDist","average bin error distribution",100,0,1);
	result->SetDirectory(0);

	TObjString lishkey("LISH");
	Int_t nrOfUnits = gng->fUnits.GetEntriesFast();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(gng->fUnits.At(i));

		Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
		for (Int_t j = 0; j < edgeCnt; ++j) {
			TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));

			// get lish
			TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&lishkey));
			if (lish == 0)
				continue;

			// calculate avg bin error
			Int_t lishSize = lish->GetNbinsX();
			Double_t cnt = 0;
			Double_t avg = 0;
			for (Int_t l = 1; l <= lishSize; ++l) {
				Double_t bc = 0;
				if ((bc = lish->GetBinContent(l)) > 0) {
					avg += lish->GetBinError(l) / bc;
					cnt += 1;
				} else {
					// test
					avg += 1;
					cnt += 1;
				}
			}
			if (cnt > 0) {
				avg /= cnt;
			} else {
				avg = 1;
			}

			result->Fill(avg);
		}			
	}

	return result;
}

void TGNGLishExt::FillAvgBinErrorDist(TGNG *gng, TH2D* hist, Double_t xVal)
{
	// this static function fills the distribution of the average bin errors
	// exhibited by the LISHs in the GNG "gng" into the 2D histogram at position
	// "xval".

	TObjString lishkey("LISH");
	Int_t nrOfUnits = gng->fUnits.GetEntriesFast();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(gng->fUnits.At(i));

		Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
		for (Int_t j = 0; j < edgeCnt; ++j) {
			TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));

			// get lish
			TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&lishkey));
			if (lish == 0)
				continue;

			// calculate avg bin error
			Int_t lishSize = lish->GetNbinsX();
			Double_t cnt = 0;
			Double_t avg = 0;
			for (Int_t l = 1; l <= lishSize; ++l) {
				Double_t bc = 0;
				if ((bc = lish->GetBinContent(l)) > 0) {
					avg += lish->GetBinError(l) / bc;
					cnt += 1;
				} else {
					// test
					avg += 1;
					cnt += 1;
				}
			}
			if (cnt > 0) {
				avg /= cnt;
			} else {
				avg = 1;
			}

			hist->Fill(xVal,avg);
		}			
	}
}






