/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a logging extension that stores not only
the gng at every time step but also every event that takes place, e.g., unit
addition / removal, edge addition / removal, etc.
*/
#include <TFile.h>
#include <TTree.h>
#include <TObjArray.h>

#include "../../core/TGNGEdge.h"
#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../core/TGNGDataElement.h"

#include "TGNGFullLoggerExt.h"

TGNGFullLoggerExt::TGNGFullLoggerExt() : 
	TGNGExtension(),
	fLogFile(0),
	fGngTree(0),
	fErrorTree(0),
	fPeekTree(0),
	fFeedTree(0),
	fEdgeTree(0),
	fUnitTree(0),
	fUpdateTree(0),
	fNextTreeIdx(0),
	fInpCnt(0),
	fGNG(0),
	fBMUIdx(0),
	fBMU2Idx(0),
	fBMUDist(0),
	fBMU2Dist(0),
	fInput(0),
	fEdgeIdx(0),
	fEventType(0),
	fErrorMsg(0),
	fErrorCode(0),
	fErrorCnt(0),
	fGNGLogInterval(1),
	fGNGLogEnabled(kTRUE),
	fErrorLogEnabled(kTRUE),
	fPeekLogEnabled(kTRUE),
	fFeedLogEnabled(kTRUE),
	fEdgeLogEnabled(kTRUE),
	fUnitLogEnabled(kTRUE),
	fUpdateLogEnabled(kTRUE),
	fRawInputDummy(),
	fRawInput(0)
{}

TGNGFullLoggerExt::TGNGFullLoggerExt(const char *file) : 
	TGNGExtension(),
	fLogFile(0),
	fGngTree(0),
	fErrorTree(0),
	fPeekTree(0),
	fFeedTree(0),
	fEdgeTree(0),
	fUnitTree(0),
	fUpdateTree(0),
	fNextTreeIdx(0),
	fInpCnt(0),
	fGNG(0),
	fBMUIdx(0),
	fBMU2Idx(0),
	fBMUDist(0),
	fBMU2Dist(0),
	fInput(0),
	fEdgeIdx(0),
	fEventType(0),
	fErrorMsg(0),
	fErrorCode(0),
	fErrorCnt(0),
	fGNGLogInterval(1),
	fGNGLogEnabled(kTRUE),
	fErrorLogEnabled(kTRUE),
	fPeekLogEnabled(kTRUE),
	fFeedLogEnabled(kTRUE),
	fEdgeLogEnabled(kTRUE),
	fUnitLogEnabled(kTRUE),
	fUpdateLogEnabled(kTRUE),
	fRawInputDummy(),
	fRawInput(0)
{
	SetLogFile(file);
}

TGNGFullLoggerExt::~TGNGFullLoggerExt()
{
	CloseLogFile();	
}

void TGNGFullLoggerExt::SetLogFile(const char *file)
{
	CloseLogFile();
	fLogFile = TFile::Open(file,"recreate");
	if (fLogFile == 0) {
		printf("Could not open logfile: %s\n", file);
		return;
	}
	// create error tree here
	fErrorTree = new TTree("errorTree","error messages of gng");
	fErrorTree->Branch("inputCnt",&fInpCnt,"idx/l");
	fErrorTree->Branch("errorCnt",&fErrorCnt,"errCnt/i");
	fErrorTree->Branch("errorCode",&fErrorCode,"errCode/i");
	fErrorTree->Branch("errorMsg","TString",&fErrorMsg);
	fErrorTree->SetMaxTreeSize(5000000000000LL);	                           
}

void TGNGFullLoggerExt::CloseLogFile()
{
	if (fLogFile) {
		fLogFile->Close();
		delete fLogFile;
		fLogFile = 0;
		fGngTree = 0;
		fErrorTree = 0;
		fPeekTree = 0;
		fFeedTree = 0;
		fEdgeTree = 0;
		fUnitTree = 0;
		fUpdateTree = 0;
	}
}

TFile* TGNGFullLoggerExt::GetCurrentLogFile()
{
	return fLogFile;
}

void TGNGFullLoggerExt::OnError(TGNG *gng, UInt_t errorCode, const char *errorMsg)
{
	if ((fErrorTree == 0) || (fErrorLogEnabled == kFALSE)) return;

	fInpCnt = gng->GetInputCount();
	++fErrorCnt;
	fErrorCode = errorCode;
	fErrorMsg = new TString(errorMsg);

	fErrorTree->Fill();

	delete fErrorMsg;
	fErrorMsg = 0;
}


void TGNGFullLoggerExt::OnGNGInit(TGNG *gng)
{
	if (fLogFile == 0) {
		gng->RaiseError(TGNG::ecExtension,"fLogFile is null on init! Fatal!");
		return;
	}
	if (fGngTree) {
		gng->RaiseError(TGNG::ecExtension,"fGngTree is non null on init!");
		delete fGngTree;
		fGngTree = 0;		
	}
	if (fPeekTree) {
		gng->RaiseError(TGNG::ecExtension,"fPeekTree is non null on init!");
		delete fPeekTree;
		fPeekTree = 0;		
	}
	if (fFeedTree) {
		gng->RaiseError(TGNG::ecExtension,"fFeedTree is non null on init!");
		delete fFeedTree;
		fFeedTree = 0;		
	}
	if (fEdgeTree) {
		gng->RaiseError(TGNG::ecExtension,"fEdgeTree is non null on init!");
		delete fEdgeTree;
		fEdgeTree = 0;		
	}
	if (fUnitTree) {
		gng->RaiseError(TGNG::ecExtension,"fUnitTree is non null on init!");
		delete fUnitTree;
		fUnitTree = 0;		
	}
	if (fUpdateTree) {
		gng->RaiseError(TGNG::ecExtension,"fUpdateTree is non null on init!");
		delete fUpdateTree;
		fUpdateTree = 0;		
	}

	fLogFile->cd();

	fGngTree = new TTree(Form("gngTree_%d",fNextTreeIdx),"the complete gng");
	fGngTree->Branch("inputCnt",&fInpCnt,"idx/l");
	fGngTree->Branch("gng","TGNG",&fGNG,32000,0);
	fGngTree->SetMaxTreeSize(5000000000000LL);

	fPeekTree = new TTree(Form("peekTree_%d",fNextTreeIdx),"peek events");
	fPeekTree->Branch("inputCnt",&fInpCnt,"idx/l");
	fPeekTree->Branch("bmu",&fBMUIdx,"bmu/I");
	fPeekTree->Branch("bmuDist",&fBMUDist,"bmuDist/D");
	fPeekTree->Branch("bmu2",&fBMU2Idx,"bmu2/I");
	fPeekTree->Branch("bmu2Dist",&fBMU2Dist,"bmu2Dist/D");
	fPeekTree->Branch("input","TGNGDataElement",&fInput,32000,0);
	fPeekTree->Branch("rawinput","TObject",&fRawInput,32000,0);
	fPeekTree->SetMaxTreeSize(5000000000000LL);

	fFeedTree = new TTree(Form("feedTree_%d",fNextTreeIdx),"feed events");
	fFeedTree->Branch("inputCnt",&fInpCnt,"idx/l");
	fFeedTree->Branch("bmu",&fBMUIdx,"bmu/I");
	fFeedTree->Branch("bmuDist",&fBMUDist,"bmuDist/D");
	fFeedTree->Branch("bmu2",&fBMU2Idx,"bmu2/I");
	fFeedTree->Branch("bmu2Dist",&fBMU2Dist,"bmu2Dist/D");
	fFeedTree->Branch("input","TGNGDataElement",&fInput,32000,0);
	fFeedTree->Branch("rawinput","TObject",&fRawInput,32000,0);
	fFeedTree->SetMaxTreeSize(5000000000000LL);

	fEdgeTree = new TTree(Form("edgeTree_%d",fNextTreeIdx),"edge events");
	fEdgeTree->Branch("inputCnt",&fInpCnt,"idx/l");
	fEdgeTree->Branch("unitA",&fBMUIdx,"unitA/I");
	fEdgeTree->Branch("unitB",&fBMU2Idx,"unitB/I");
	fEdgeTree->Branch("edge",&fEdgeIdx,"edge/I");
	fEdgeTree->Branch("type",&fEventType,"type/I");
	fEdgeTree->SetMaxTreeSize(5000000000000LL);

	fUnitTree = new TTree(Form("unitTree_%d",fNextTreeIdx),"units events");
	fUnitTree->Branch("inputCnt",&fInpCnt,"idx/l");
	fUnitTree->Branch("unit",&fBMUIdx,"unit/I");
	fUnitTree->Branch("type",&fEventType,"type/I");
	fUnitTree->SetMaxTreeSize(5000000000000LL);

	fUpdateTree = new TTree(Form("updateTree_%d",fNextTreeIdx),"update events");
	fUpdateTree->Branch("inputCnt",&fInpCnt,"idx/l");
	fUpdateTree->Branch("bmu",&fBMUIdx,"bmu/I");
	fUpdateTree->Branch("bmuDist",&fBMUDist,"bmuDist/D");
	fUpdateTree->Branch("bmu2",&fBMU2Idx,"bmu2/I");
	fUpdateTree->Branch("bmu2Dist",&fBMU2Dist,"bmu2Dist/D");
	fUpdateTree->Branch("input","TGNGDataElement",&fInput,32000,0);
	fUpdateTree->Branch("edge",&fEdgeIdx,"edge/I");
	fUpdateTree->SetMaxTreeSize(5000000000000LL);

	fInpCnt = gng->GetInputCount();
	fGNG    = gng;

	fGngTree->Fill();

	fNextTreeIdx++;
}

void TGNGFullLoggerExt::OnGNGReset(TGNG*)
{
	//printf("OnGNGReset() begin\n");
	if (fLogFile == 0)
		return;

	fLogFile->cd();
	if (fGngTree) {
		fGngTree->Write();
		delete fGngTree;
		fGngTree = 0;		
	}
	if (fPeekTree) {
		fPeekTree->Write();
		delete fPeekTree;
		fPeekTree = 0;		
	}
	if (fFeedTree) {
		fFeedTree->Write();
		delete fFeedTree;
		fFeedTree = 0;		
	}
	if (fEdgeTree) {
		fEdgeTree->Write();
		delete fEdgeTree;
		fEdgeTree = 0;		
	}
	if (fUnitTree) {
		fUnitTree->Write();
		delete fUnitTree;
		fUnitTree = 0;		
	}
	if (fUpdateTree) {
		fUpdateTree->Write();
		delete fUpdateTree;
		fUpdateTree = 0;		
	}
	//printf("OnGNGReset() end\n");
}

void TGNGFullLoggerExt::OnPeekResult(TGNG *gng, TGNGUnit *bmu, Double_t bmuDist,
	                                 TGNGUnit *bmu2, Double_t bmu2Dist, const TGNGDataElement *input)
{
	if ((fPeekTree == 0) || (fPeekLogEnabled == kFALSE)) return;

	fInpCnt  = gng->GetInputCount();
	fBMUIdx  = gng->fUnits.IndexOf(bmu);
	fBMUDist = bmuDist;
	fBMU2Idx  = gng->fUnits.IndexOf(bmu2);
	fBMU2Dist = bmu2Dist;
	fInput   = const_cast<TGNGDataElement*>(input);

	fRawInput = gng->GetLastRawInput();
	if (fRawInput == 0)
		fRawInput = &fRawInputDummy;

	fPeekTree->Fill();

	if ((fGngTree == 0) || ((fInpCnt % fGNGLogInterval) != 0) || (fGNGLogEnabled == kFALSE)) return;

	// store gng
	fGNG = gng;
	fGngTree->Fill();
	fGNG = 0;
}

void TGNGFullLoggerExt::OnFeedResult(TGNG *gng, TGNGUnit *bmu, Double_t bmuDist,
                                     TGNGUnit *bmu2, Double_t bmu2Dist, const TGNGDataElement *input)
{
	if ((fFeedTree == 0) || (fFeedLogEnabled == kFALSE)) return;

	fInpCnt = gng->GetInputCount();

	// store feed result
	fBMUIdx  = gng->fUnits.IndexOf(bmu);
	fBMUDist = bmuDist;
	fBMU2Idx  = gng->fUnits.IndexOf(bmu2);
	fBMU2Dist = bmu2Dist;
	fInput   = const_cast<TGNGDataElement*>(input);

	fRawInput = gng->GetLastRawInput();
	if (fRawInput == 0)
		fRawInput = &fRawInputDummy;

	fFeedTree->Fill();

	if ((fGngTree == 0) || ((fInpCnt % fGNGLogInterval) != 0) || (fGNGLogEnabled == kFALSE)) return;

	// store gng
	fGNG = gng;
	fGngTree->Fill();
	fGNG = 0;
}

void TGNGFullLoggerExt::OnNewUnit(TGNG *gng, TGNGUnit *newUnit)
{
	if ((fUnitTree == 0) || (fUnitLogEnabled == kFALSE)) return;

	fInpCnt    = gng->GetInputCount();
	fBMUIdx    = gng->fUnits.IndexOf(newUnit);
	fEventType = 1;
	fUnitTree->Fill();
}

void TGNGFullLoggerExt::OnUnitRemoval(TGNG *gng, TGNGUnit *oldUnit)
{
	if ((fUnitTree == 0) || (fUnitLogEnabled == kFALSE)) return;

	fInpCnt    = gng->GetInputCount();
	fBMUIdx    = gng->fUnits.IndexOf(oldUnit);
	fEventType = -1;
	fUnitTree->Fill();
}

void TGNGFullLoggerExt::OnNewEdge(TGNG *gng, TGNGEdge *newEdge)
{
	if ((fEdgeTree == 0) || (fEdgeLogEnabled == kFALSE)) return;

	fInpCnt    = gng->GetInputCount();
	fBMUIdx    = gng->fUnits.IndexOf(newEdge->fUnitA);
	fBMU2Idx   = gng->fUnits.IndexOf(newEdge->fUnitB);
	fEdgeIdx   = newEdge->fUnitA->fEdges.IndexOf(newEdge);
	fEventType = 1;
	fEdgeTree->Fill();
}

void TGNGFullLoggerExt::OnEdgeRemoval(TGNG *gng, TGNGEdge *newEdge)
{
	if ((fEdgeTree == 0) || (fEdgeLogEnabled == kFALSE)) return;

	fInpCnt    = gng->GetInputCount();
	fBMUIdx    = gng->fUnits.IndexOf(newEdge->fUnitA);
	fBMU2Idx   = gng->fUnits.IndexOf(newEdge->fUnitB);
	fEdgeIdx   = newEdge->fUnitA->fEdges.IndexOf(newEdge);
	fEventType = -1;
	fEdgeTree->Fill();
}

void TGNGFullLoggerExt::OnUnitUpdate(TGNG                  *gng, 
                   					 TGNGUnit              *bmu,
                  					 Double_t               bmuDist,
                  					 TGNGUnit              *bmu2,
                  					 Double_t               bmu2Dist,
                  					 TGNGEdge              *bmuEdge,
                  					 const TGNGDataElement *input    )
{
	if ((fUpdateTree == 0) || (fUpdateLogEnabled == kFALSE)) return;

	fInpCnt   = gng->GetInputCount();
	fBMUIdx   = gng->fUnits.IndexOf(bmu);
	fBMUDist  = bmuDist;
	fBMU2Idx  = gng->fUnits.IndexOf(bmu2);
	fBMU2Dist = bmu2Dist;
	fInput    = const_cast<TGNGDataElement*>(input);
	fEdgeIdx  = bmu->fEdges.IndexOf(bmuEdge);
	fUpdateTree->Fill();
	fInput = 0;
}

void TGNGFullLoggerExt::SetGNGLogInterval(UInt_t val)
{
	fGNGLogInterval = val;
}

UInt_t TGNGFullLoggerExt::GetGNGLogInterval()
{
	return fGNGLogInterval;
}

void TGNGFullLoggerExt::SetGNGLogEnabled(Bool_t enable)
{
	fGNGLogEnabled = enable;
}

void TGNGFullLoggerExt::SetErrorLogEnabled(Bool_t enable)
{
	fErrorLogEnabled = enable;
}

void TGNGFullLoggerExt::SetPeekLogEnabled(Bool_t enable)
{
	fPeekLogEnabled = enable;
}

void TGNGFullLoggerExt::SetFeedLogEnabled(Bool_t enable)
{
	fFeedLogEnabled = enable;
}

void TGNGFullLoggerExt::SetEdgeLogEnabled(Bool_t enable)
{
	fEdgeLogEnabled = enable;
}

void TGNGFullLoggerExt::SetUnitLogEnabled(Bool_t enable)
{
	fUnitLogEnabled = enable;
}

void TGNGFullLoggerExt::SetUpdateLogEnabled(Bool_t enable)
{
	fUpdateLogEnabled = enable;
}


Bool_t TGNGFullLoggerExt::GetGNGLogEnabled()
{
	return fGNGLogEnabled;
}

Bool_t TGNGFullLoggerExt::GetErrorLogEnabled()
{
	return fErrorLogEnabled;
}

Bool_t TGNGFullLoggerExt::GetPeekLogEnabled()
{
	return fPeekLogEnabled;
}

Bool_t TGNGFullLoggerExt::GetFeedLogEnabled()
{
	return fFeedLogEnabled;
}

Bool_t TGNGFullLoggerExt::GetEdgeLogEnabled()
{
	return fEdgeLogEnabled;
}

Bool_t TGNGFullLoggerExt::GetUnitLogEnabled()
{
	return fUnitLogEnabled;
}

Bool_t TGNGFullLoggerExt::GetUpdateLogEnabled()
{
	return fUpdateLogEnabled;
}

