/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a input logging extension that stores
either the last n, best n, or worst n inputs to a gng unit as an unit attribute.

The logging mode can be set by SetLogMode() to the values lmLastN, lmBestN, lmWorstN.
The log size n per unit can be set by SetLogSize()
*/

#include "../../core/TGNGEdge.h"
#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../core/TGNGDataElement.h"

#include "TGNGInputLogExt.h"
#include "TGNGInpLogItem.h"
#include "TGNGInpLogAttr.h"

TGNGInputLogExt::TGNGInputLogExt() : 
	TGNGExtension(),
	fAttrKey("INPLOG"),
	fLogMode(TGNGInputLogExt::lmLastN),
	fLogSize(1),
	fFeedEnable(kFALSE),
	fPeekEnable(kTRUE)
{
	// default constructor setting log size to 1 and mode to last N
}

TGNGInputLogExt::TGNGInputLogExt(Int_t mode, Int_t size, const char *postfix) : 
	TGNGExtension(),
	fAttrKey(Form("INPLOG%s",postfix)),
	fLogMode(mode),
	fLogSize(size),
	fFeedEnable(kFALSE),
	fPeekEnable(kTRUE)
{
	// default constructor setting log size to "size" and mode to "mode"
	// In addition, you can provide a postfix to the attribute key used. This
	// way you can store, e.g., the last 5 and best 10 inputs to a unit by using
	// two input log extension having different postfixes.
}

void TGNGInputLogExt::OnGNGReset(TGNG*)
{
	
}

void TGNGInputLogExt::OnNewUnit(TGNG *gng, TGNGUnit *newUnit)
{
	if (newUnit->fAttributes.FindObject(&fAttrKey) != 0) {
		gng->RaiseError(TGNG::ecExtension,Form("Unit has already an attribute \"%s\"!",fAttrKey.String().Data()));
		return;
	}
	newUnit->fAttributes.Add(new TObjString(fAttrKey),new TGNGInpLogAttr());
}

void TGNGInputLogExt::OnPeekResult(TGNG *gng, TGNGUnit *bmu, Double_t bmuDist,
	                               TGNGUnit*, Double_t, const TGNGDataElement *input)
{
	if (fPeekEnable == kFALSE)
		return;

	LogInput(gng,bmu,bmuDist,input);
}

void TGNGInputLogExt::OnFeedResult(TGNG *gng, TGNGUnit *bmu, Double_t bmuDist,
								   TGNGUnit*, Double_t, const TGNGDataElement *input)
{
	if (fFeedEnable == kFALSE)
		return;

	LogInput(gng,bmu,bmuDist,input);
}

void TGNGInputLogExt::LogInput(TGNG *gng, TGNGUnit *bmu, Double_t bmuDist, const TGNGDataElement *input)
{
	TGNGInpLogAttr *unitLog = static_cast<TGNGInpLogAttr*>(bmu->fAttributes.GetValue(&fAttrKey));
	if (unitLog == 0) {
		gng->RaiseError(TGNG::ecExtension,Form("Unit has no attribute \"%s\"!",fAttrKey.String().Data()));
		return;		
	}

	Int_t nrOfInputs = unitLog->fItems.GetEntriesFast();

	if (nrOfInputs < fLogSize) {
		unitLog->fItems.Add(new TGNGInpLogItem(static_cast<TGNGDataElement*>(input->Clone()),bmuDist));
		unitLog->fNextIdx = (unitLog->fNextIdx + 1) % fLogSize;
		if ((fLogMode == TGNGInputLogExt::lmBestN) ||
			(fLogMode == TGNGInputLogExt::lmWorstN)) 
		{
			unitLog->fItems.Sort();
		}
		return;
	}

	switch (fLogMode) {
		case TGNGInputLogExt::lmLastN : {
			TObject *obj = unitLog->fItems.RemoveAt(unitLog->fNextIdx);
			delete obj;
			unitLog->fItems.AddAt(new TGNGInpLogItem(static_cast<TGNGDataElement*>(input->Clone()),bmuDist),unitLog->fNextIdx);
			unitLog->fNextIdx = (unitLog->fNextIdx + 1) % fLogSize;
		} break;

		case TGNGInputLogExt::lmBestN : {
			TGNGInpLogItem *leastBest = static_cast<TGNGInpLogItem*>(unitLog->fItems.At(nrOfInputs-1));
			if (bmuDist < leastBest->fDist) {
				TObject *obj = unitLog->fItems.RemoveAt(nrOfInputs-1);
				delete obj;
				unitLog->fItems.AddAt(new TGNGInpLogItem(static_cast<TGNGDataElement*>(input->Clone()),bmuDist),nrOfInputs-1);
				unitLog->fItems.Sort();
			}
		} break;

		case TGNGInputLogExt::lmWorstN : {
			TGNGInpLogItem *leastWorst = static_cast<TGNGInpLogItem*>(unitLog->fItems.At(0));
			if (bmuDist > leastWorst->fDist) {
				TObject *obj = unitLog->fItems.RemoveAt(0);
				delete obj;
				unitLog->fItems.AddAt(new TGNGInpLogItem(static_cast<TGNGDataElement*>(input->Clone()),bmuDist),0);
				unitLog->fItems.Sort();
			}			
		} break;

		default: {
			gng->RaiseError(TGNG::ecExtension,"Unknown input log mode!");
		}
	}
}

void TGNGInputLogExt::SetLogMode(Int_t mode)
{
	fLogMode = mode;
}

Int_t TGNGInputLogExt::GetLogMode()
{
	return fLogMode;
}

void TGNGInputLogExt::SetLogSize(Int_t size)
{
	fLogSize = size;
}

Int_t TGNGInputLogExt::GetLogSize()
{
	return fLogSize;
}

void TGNGInputLogExt::SetFeedLog(Bool_t enable)
{
	fFeedEnable = enable;
}

void TGNGInputLogExt::SetPeekLog(Bool_t enable)
{
	fPeekEnable = enable;
}
