/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a gng graph visualization by a 
force-directed graph drawing approach based on Eades.

See the Draw() method for options.
*/
#include <limits>
#include <cstdio>
#include <TSystem.h>
#include <TObjString.h>
#include <TH1D.h>
#include <TRandom.h>
#include <TVector2.h>
#include <TMath.h>

#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../datatypes/TGNGVectorData.h"
#include "../functional/TGNGIdExt.h"

#include "TGNGEadesVis.h"

TGNGEadesVis::TGNGEadesVis() :
	TGNGExtension(),
	TAttLine(),
	TAttMarker(),
	TAttText(),
	TAttFill(),
	fPosAttr("EadesPos"),
	fIterations(10),
	fCurrentPad(0),
	fUpdateInterval(0.05),
	fLastUpdate(),
	fCurrentGNG(0),
	fCurrentBMU(0),
	fEdgeColor(kBlack),
	fEdgeStyle(1),
	fEdgeWidth(1),
	fUnitColor(kAzure+1),
	fUnitSize(1),
	fUnitStyle(21),
	fUnitMColor(kOrange+1),
	fUnitMSize(2),
	fUnitMStyle(21),
	fUnitTextOffsetX(0.015),
	fUnitTextOffsetY(-0.015),
	fUnitTextColor(kBlue+2),
	fUnitTextFont(100),
	fUnitTextSize(0.02)	
{}


void TGNGEadesVis::Draw(Option_t* option)
{
	// draws the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e - do not draw edges
	// !u - do not draw units
	// m - mark current BMU
	// t - draw node IDs (if attribute "ID" is present)
	// l - draw LISHs (if attribute "LISH" is present)

	// clear current Pad
	if (gPad)
		gPad->Clear(); 

	TObject::Draw(option);
	fCurrentPad = gPad;
}

void TGNGEadesVis::Paint(Option_t *option)
{
	// paints the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e - do not draw edges
	// !u - do not draw units
	// m - mark current BMU
	// t - draw node IDs (if attribute "ID" is present)
	// l - draw LISHs (if attribute "LISH" is present)

	if ((fCurrentGNG == 0) || (fCurrentPad == 0))
		return;

	// calculate positions
	Double_t c1 = 2.0;
	Double_t c2 = 1.0;
	Double_t c3 = 1.0;
	Double_t c4 = 0.1;

	Int_t nrOfUnits = fCurrentGNG->fUnits.GetEntriesFast();

	for (Int_t i = 0; i < fIterations; ++i) {
		TVector2 centerPos(0.0,0.0);
		for (Int_t j = 0; j < nrOfUnits; ++j) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(j));
			// get unit pos
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
				continue;
			}
			TVector2 unitDelta(0.0,0.0);
			// calculate repellant forces
			for (Int_t k = 0; k < nrOfUnits; ++k) {
				if (j == k) continue;
				TGNGUnit *otherUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(k));
				// get unit pos
				TVector2 *otherPos = static_cast<TVector2*>(otherUnit->fAttributes.GetValue(&fPosAttr));
				if (otherPos == 0) {
					fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
					continue;
				}
				TVector2 curDelta = *unitPos - *otherPos;
				unitDelta += curDelta.Unit() * (c3 / curDelta.Mod2());
			}

			// calculate attractive forces
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t k = 0; k < edgeCnt; ++k) {
				TGNGEdge *curEdge   = static_cast<TGNGEdge*>(curUnit->fEdges.At(k));
				TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;
				// get other units pos
				TVector2 *otherPos = static_cast<TVector2*>(otherUnit->fAttributes.GetValue(&fPosAttr));
				if (otherPos == 0) {
					fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
					continue;
				}
				TVector2 curDelta = *unitPos - *otherPos;
				// subtract previously added repellant force
				unitDelta -= curDelta.Unit() * (c3 / curDelta.Mod2());
				// add attractive force
				unitDelta += curDelta.Unit() * (-c1 * TMath::Log(curDelta.Mod() / c2));
			}

			// update vertex pos
			*unitPos += c4 * unitDelta;
			centerPos += *unitPos;
		}
		centerPos /= (Double_t)nrOfUnits;
		for (Int_t j = 0; j < nrOfUnits; ++j) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(j));
			// get unit pos
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
				continue;
			}
			*unitPos -= centerPos;
		}		
	}

	// get space dimension
	Double_t minX = numeric_limits<Double_t>::max();
	Double_t maxX = numeric_limits<Double_t>::min();
	Double_t minY = numeric_limits<Double_t>::max();
	Double_t maxY = numeric_limits<Double_t>::min();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit       *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));

		TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
		if (unitPos == 0) {
			fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
			continue;
		}

		Double_t curX = unitPos->X();
		Double_t curY = unitPos->Y();
		if (curX < minX) minX = curX;
		if (curX > maxX) maxX = curX;
		if (curY < minY) minY = curY;
		if (curY > maxY) maxY = curY;
	}

	Double_t dimX = maxX - minX;
	Double_t dimY = maxY - minY;

	// get pad dimensions
	Double_t to_w = fCurrentPad->UtoPixel(1.);
	Double_t to_h = fCurrentPad->VtoPixel(0.);

	Double_t left_m   = fCurrentPad->GetLeftMargin();
	Double_t right_m  = fCurrentPad->GetLeftMargin();
	Double_t top_m    = fCurrentPad->GetTopMargin();
	Double_t bottom_m = fCurrentPad->GetBottomMargin();

	Double_t mw = 1.0 - left_m - right_m;
	Double_t mh = 1.0 - top_m - bottom_m;

	TString opt = option;
	opt.ToLower();

	// draw edges
	if (!opt.Contains("!e")) {

		SetLineColor(fEdgeColor);
		SetLineStyle(fEdgeStyle);
		SetLineWidth(fEdgeWidth);
		TAttLine::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
				continue;
			}
			Double_t x1 = unitPos->X();
			Double_t y1 = unitPos->Y();
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
				if (curEdge->fUnitA != curUnit)
					continue;
				// get other units positions
				TVector2 *otherPos = static_cast<TVector2*>(curEdge->fUnitB->fAttributes.GetValue(&fPosAttr));
				if (otherPos == 0) {
					fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
					continue;
				}
				Double_t x2 = otherPos->X();
				Double_t y2 = otherPos->Y();
				// draw line
				fCurrentPad->PaintLine((x1-minX) / dimX * mw + left_m,
								1.0 - ((y1-minY) / dimY * mh) - top_m,
					            (x2-minX) / dimX * mw + left_m,
					            1.0 - ((y2-minY) / dimY * mh) - top_m);
			}			
		}

	}

	// draw LISHs
	
	if (opt.Contains("l")) {
		TObjString key("LISH");

		SetFillColor(kBlack);
		SetFillStyle(1001);
		TAttFill::Modify();

		SetLineColor(kOrange-9);
		SetLineStyle(1);
		SetLineWidth(1);
		TAttLine::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit pos
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
				continue;
			}
			Double_t x1 = unitPos->X();
			Double_t y1 = unitPos->Y();
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
				if (curEdge->fUnitA != curUnit)
					continue;
				// get other units positions
				TVector2 *otherPos = static_cast<TVector2*>(curEdge->fUnitB->fAttributes.GetValue(&fPosAttr));
				if (otherPos == 0) {
					fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
					continue;
				}
				Double_t x2 = otherPos->X();
				Double_t y2 = otherPos->Y();

				// get lish
				TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&key));
				if (lish == 0)
					continue;

				// center pos on line
				Double_t cpx = ((x1+x2)/2.0 - minX) / dimX * mw + left_m;
				Double_t cpy = 1.0 - (((y1+y2)/2.0 - minY) / dimY * mh) - top_m;

				Double_t bw2 = 0.015;
				Double_t bh2 = 0.01;

				Double_t xmargin = 0.002;
				Double_t ymargin = 0.002;

				// draw black box centered on the line
				fCurrentPad->PaintBox(cpx-bw2,cpy+bh2,cpx+bw2,cpy-bh2);

				cpx -= bw2 - xmargin;
				cpy -= bh2 - ymargin;

				// draw mini histogram
				Int_t nrOfBins  = lish->GetNbinsX();
				Double_t maxVal = lish->GetBinContent(lish->GetMaximumBin());
				if (maxVal == 0.0) maxVal = 1.0;
				Double_t xFrac  = ((bw2 - xmargin) * 2.0) / (Double_t)nrOfBins;
				Double_t yFact  = ((bh2 - ymargin) * 2.0) / maxVal;

				for (Int_t k = 0; k < nrOfBins; ++k) {
					Double_t curVal = (x1 < x2) ? lish->GetBinContent(k+1) : 
												  lish->GetBinContent(nrOfBins-k);
					// draw line
					Double_t tmpX = cpx + (Double_t)k * xFrac;
					fCurrentPad->PaintLine(tmpX,cpy,tmpX,cpy + (curVal * yFact));	
				}


			}			
		}
	}
	

	// draw units
	if (!opt.Contains("!u")) {


		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit pos
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
				continue;
			}
			Double_t x1 = (unitPos->X() - minX) / dimX * mw + left_m;
			Double_t y1 = 1.0 - ((unitPos->Y() - minY) / dimY * mh) - top_m;
			// later: set style dependend on attributes
			if ((curUnit == fCurrentBMU) && (opt.Contains("m"))) {
				SetMarkerColor(fUnitMColor);
				SetMarkerSize(fUnitMSize);
				SetMarkerStyle(fUnitMStyle);
			} else {
				SetMarkerColor(fUnitColor);
				SetMarkerSize(fUnitSize);
				SetMarkerStyle(fUnitStyle);
			}
			TAttMarker::Modify();				
			fCurrentPad->PaintPolyMarker(1,&x1,&y1);

		}

	}

	// draw uint IDs 
	if (opt.Contains("t")) {
		TObjString key("ID");

		char buffer[10]; // buffer for node IDs
		SetTextAlign(22);
		SetTextColor(fUnitTextColor);
		SetTextFont(fUnitTextFont);
		SetTextSize(fUnitTextSize);
		TAttText::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit id
			TGNGId *uid = static_cast<TGNGId*>(curUnit->fAttributes.GetValue(&key));
			if (uid == 0)
				continue;

			// get unit pos
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"EadesPos\"!");
				continue;
			}
			Double_t x1 = (unitPos->X() - minX) / dimX * mw + left_m;
			Double_t y1 = 1.0 - ((unitPos->Y() - minY) / dimY * mh) - top_m;

			std::sprintf(buffer,"%d",uid->fID);
			fCurrentPad->PaintText(x1+fUnitTextOffsetX,
				            y1+fUnitTextOffsetY,buffer);
		}
	}


	//fCurrentGNG = 0;
	//fCurrentBMU = 0;
}

void TGNGEadesVis::OnPeekResult(TGNG *gng, TGNGUnit *bmu, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGEadesVis::OnFeedResult(TGNG *gng, TGNGUnit *bmu, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGEadesVis::OnNewUnit(TGNG *gng, TGNGUnit *newUnit)
{
	// add needed attributes
	if (newUnit->fAttributes.FindObject(&fPosAttr) != 0) {
		gng->RaiseError(TGNG::ecExtension,"Unit has already an attribute \"EadesPos\"!");
		return;
	}
	TVector2 *nPos = new TVector2(0.0,0.0);

	Int_t edgeCnt = newUnit->fEdges.GetEntriesFast();
	if (edgeCnt > 1) {
		Double_t minX = numeric_limits<Double_t>::max();
		Double_t maxX = numeric_limits<Double_t>::min();
		Double_t minY = numeric_limits<Double_t>::max();
		Double_t maxY = numeric_limits<Double_t>::min();

		for (Int_t i = 0; i < edgeCnt; ++i) {
			TGNGEdge *curEdge = static_cast<TGNGEdge*>(newUnit->fEdges.At(i));
			TGNGUnit *otherUnit = (curEdge->fUnitA != newUnit) ? curEdge->fUnitA : curEdge->fUnitB;
			TVector2 *otherPos = static_cast<TVector2*>(otherUnit->fAttributes.GetValue(&fPosAttr));
			if (otherPos == 0) {
				continue;
			}
			Double_t curX = otherPos->X();
			Double_t curY = otherPos->Y();
			if (curX < minX) minX = curX;
			if (curX > maxX) maxX = curX;
			if (curY < minY) minY = curY;
			if (curY > maxY) maxY = curY;
		}

		Double_t dimX = maxX - minX;
		Double_t dimY = maxY - minY;

		nPos->Set(minX + gRandom->Uniform(dimX),minY + gRandom->Uniform(dimY));

	} else {
		nPos->Set(gRandom->Uniform(10),gRandom->Uniform(10));
	}

	newUnit->fAttributes.Add(new TObjString(fPosAttr),nPos);

}

void TGNGEadesVis::SetUpdateInterval(Double_t sec)
{
	fUpdateInterval = sec;
}

void TGNGEadesVis::SetEdgeColor(Color_t color)
{
	fEdgeColor = color;
}

void TGNGEadesVis::SetEdgeStyle(Style_t style)
{
	fEdgeStyle = style;
}

void TGNGEadesVis::SetEdgeWidth(Width_t width)
{
	fEdgeWidth = width;
}


void TGNGEadesVis::SetUnitColor(Color_t color)
{
	fUnitColor = color;
}

void TGNGEadesVis::SetUnitSize(Size_t size)
{
	fUnitSize = size;
}

void TGNGEadesVis::SetUnitStyle(Style_t style)
{
	fUnitStyle = style;
}

void TGNGEadesVis::SetUnitTextOffsetX(Double_t ox)
{
	fUnitTextOffsetX = ox;
}

void TGNGEadesVis::SetUnitTextOffsetY(Double_t oy)
{
	fUnitTextOffsetY = oy;
}

void TGNGEadesVis::SetUnitTextColor(Color_t color)
{
	fUnitTextColor = color;
}

void TGNGEadesVis::SetUnitTextFont(Font_t font)
{
	fUnitTextFont = font;
}

void TGNGEadesVis::SetUnitTextSize(Float_t size)
{
	fUnitTextSize = size;
}



void TGNGEadesVis::UpdateVis()
{
	TTimeStamp current;
	if (current - fLastUpdate < fUpdateInterval)
		return;

	fLastUpdate = current;

	if (fCurrentPad) {
		fCurrentPad->Modified();
		fCurrentPad->Update();
		gSystem->ProcessEvents();
	}	
}

