/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a gng graph visualization that uses a
force-directed graph drawing approach based the work of Fruchterman and 
Reingold.

See Draw() method for options.
*/
#include <limits>
#include <cstdio>
#include <TSystem.h>
#include <TObjString.h>
#include <TH1D.h>
#include <TRandom.h>
#include <TVector2.h>
#include <TMath.h>
#include <TColor.h>
#include <TStyle.h>

#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../datatypes/TGNGVectorData.h"
#include "../functional/TGNGIdExt.h"
#include "../cluster/TGNGSimpleClusterExt.h"

#include "TGNGFruReVis.h"
#include "TGNGVisEdge.h"


TGNGFruReVis::TGNGFruReVis() :
	TGNGExtension(),
	TAttLine(),
	TAttMarker(),
	TAttText(),
	TAttFill(),
	fPosAttr("FruRePos"),
	fPadAttr("FruRePad"),
	fIterations(1),
	fTemp(1.0),
	fCurrentPad(0),
	fUpdateInterval(0.05),
	fLastUpdate(),
	fCurrentGNG(0),
	fCurrentBMU(0),
	fEdgeColor(kBlack),
	fEdgeStyle(1),
	fEdgeWidth(1),
	fUnitColor(kAzure-8),
	fUnitSize(1),
	fUnitStyle(21),
	fUnitMColor(kOrange+1),
	fUnitMSize(2),
	fUnitMStyle(21),
	fUnitTextColor(kBlue+4),
	fUnitTextFont(100),
	fUnitTextSize(0.02),
	fLishWeight(kTRUE),
	fLishWeightCenter(0.25),
	fLishWeightSteepness(25.0),
	fVisEdges(),
	fPaletteA(52),
	fLocalSpace(40)
{
	fVisEdges.SetOwner(kTRUE);
	SetPalette(fPaletteA,kTRUE);
}


void TGNGFruReVis::Draw(Option_t* option)
{
	// draws the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e - do not draw edges
	// !u - do not draw units
	// m - mark current BMU
	// t - draw node IDs (if attribute "ID" is present)
	// l - draw LISHs (if attribute "LISH" is present)
	// c - draw Clusters (if attribute "SCINFO" is present)

	// clear current Pad
	if (gPad)
		gPad->Clear(); 

	TObject::Draw(option);
	fCurrentPad = gPad;
}

Double_t TGNGFruReVis::GetEdgeLishStrength(TH1D *hist)
{
	Int_t lishSize = hist->GetNbinsX();
	Double_t cnt = 0;
	Double_t avg = 0;
	for (Int_t l = 1; l <= lishSize; ++l) {
		Double_t bc = 0;
		if ((bc = hist->GetBinContent(l)) > 0) {
			avg += hist->GetBinError(l) / bc;
			cnt += 1;
		} else {
			// test
			avg += 1;
			cnt += 1;
		}
	}
	if (cnt > 0) {
		avg /= cnt;
	} else {
		avg = 1;
	}

	//lw = TMath::Power(1.0 - avg,2.0);
	// generalised logistic function 
	return 1.0 - 1.0 / TMath::Power(1.0 + 0.5 * TMath::Exp(-fLishWeightSteepness * (avg - fLishWeightCenter)) ,2) ;

}

void TGNGFruReVis::Paint(Option_t *option)
{
	// paints the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e - do not draw edges
	// !u - do not draw units
	// m - mark current BMU
	// t - draw node IDs (if attribute "ID" is present)
	// l - draw LISHs (if attribute "LISH" is present)
	// c - draw Clusters (if attribute "SCINFO" is present)

	if ((fCurrentGNG == 0) || (fCurrentPad == 0))
		return;

	TObjString lishkey("LISH");

	// calculate positions
	Int_t nrOfUnits = fCurrentGNG->fUnits.GetEntriesFast();

	Double_t w     = 1000;
	Double_t area  = w*w;
	Double_t c1    = 0.1 * TMath::Sqrt(area / (Double_t)nrOfUnits);
	Double_t c2    = 0.01 * area / (Double_t)nrOfUnits;
	//Double_t local = fLocalSpace;

	fTemp = 1.0;
	for (Int_t i = 0; i < fIterations; ++i) {
		for (Int_t j = 0; j < nrOfUnits; ++j) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(j));
			// get unit pos
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				addPos(fCurrentGNG,curUnit,kTRUE);
				unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
				/*
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
				continue;
				*/
			}
			TVector2 unitDelta(0.0,0.0);
			// calculate repellant forces
			for (Int_t k = 0; k < nrOfUnits; ++k) {
				if (j == k) continue;
				TGNGUnit *otherUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(k));
				// get unit pos
				TVector2 *otherPos = static_cast<TVector2*>(otherUnit->fAttributes.GetValue(&fPosAttr));
				if (otherPos == 0) {
					fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
					continue;
				}
				TVector2 curDelta = *unitPos - *otherPos;
				Double_t curDist = curDelta.Mod();
				if (curDist < 0.0000001)
					curDist = TMath::Sign(0.0000001,curDist);

				Double_t localPush = 1.0;
				if (curDist < fLocalSpace)
					localPush = fLocalSpace * (1.0 - (curDist / fLocalSpace));

				unitDelta += curDelta.Unit() * (c2 / curDist) * localPush;
			}

			// calculate attractive forces
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t k = 0; k < edgeCnt; ++k) {
				TGNGEdge *curEdge   = static_cast<TGNGEdge*>(curUnit->fEdges.At(k));
				TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;
				// get other units pos
				TVector2 *otherPos = static_cast<TVector2*>(otherUnit->fAttributes.GetValue(&fPosAttr));
				if (otherPos == 0) {
					fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
					continue;
				}
				Double_t lw = 1.0;
				if (fLishWeight == kTRUE) {
					// get lish
					TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&lishkey));
					if (lish != 0) {
						lw = GetEdgeLishStrength(lish);
					}
				}

				TVector2 curDelta = *unitPos - *otherPos;
				// add attractive force
				unitDelta -= curDelta.Unit() * (curDelta.Mod2() / c1) * lw;
			}

			// update vertex pos
			Double_t deltaLength = unitDelta.Mod() * fTemp * 0.005;
			if (TMath::Abs(deltaLength) > c1)
				deltaLength = TMath::Sign(c1,deltaLength);
			*unitPos += unitDelta.Unit() * deltaLength;

			Double_t newX = unitPos->X();
			Double_t newY = unitPos->Y();

			if (newX < 0) newX = 0;
			if (newX > w) newX = w;
			if (newY < 0) newY = 0;
			if (newY > w) newY = w;

			unitPos->Set(newX,newY);
		}
		fTemp *= 0.95;
		if (fTemp < 0.05) fTemp = 0.05;
	}

	// get space dimension
	Double_t minX = numeric_limits<Double_t>::max();
	Double_t maxX = numeric_limits<Double_t>::min();
	Double_t minY = numeric_limits<Double_t>::max();
	Double_t maxY = numeric_limits<Double_t>::min();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit       *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));

		TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
		if (unitPos == 0) {
			fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
			continue;
		}

		Double_t curX = unitPos->X();
		Double_t curY = unitPos->Y();
		if (curX < minX) minX = curX;
		if (curX > maxX) maxX = curX;
		if (curY < minY) minY = curY;
		if (curY > maxY) maxY = curY;
	}

	Double_t dimX = maxX - minX;
	Double_t dimY = maxY - minY;

	// get pad dimensions
	Double_t to_w = fCurrentPad->UtoPixel(1.);
	Double_t to_h = fCurrentPad->VtoPixel(0.);

	Double_t left_m   = fCurrentPad->GetLeftMargin();
	Double_t right_m  = fCurrentPad->GetLeftMargin();
	Double_t top_m    = fCurrentPad->GetTopMargin();
	Double_t bottom_m = fCurrentPad->GetBottomMargin();

	Double_t mw = 1.0 - left_m - right_m;
	Double_t mh = 1.0 - top_m - bottom_m;

	TString opt = option;
	opt.ToLower();

	// draw edges
	if (!opt.Contains("!e")) {

		// prepare edge buffer
		Int_t vEdgeCnt = fVisEdges.GetEntriesFast();
		for (Int_t i = 0; i < vEdgeCnt; ++i) {
			TGNGVisEdge *curVisEdge = static_cast<TGNGVisEdge*>(fVisEdges.At(i));
			curVisEdge->fFlag = 0;
		}
		Int_t usedVisEdges = 0;

		SetLineColor(fEdgeColor);
		SetLineStyle(fEdgeStyle);
		SetLineWidth(fEdgeWidth);
		TAttLine::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
				continue;
			}
			Double_t x1 = unitPos->X();
			Double_t y1 = unitPos->Y();
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
				if (curEdge->fUnitA != curUnit)
					continue;
				// get other units positions
				TVector2 *otherPos = static_cast<TVector2*>(curEdge->fUnitB->fAttributes.GetValue(&fPosAttr));
				if (otherPos == 0) {
					fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
					continue;
				}

				Double_t lw = 1.0;
				if (fLishWeight == kTRUE) {
					// get lish
					TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&lishkey));
					if (lish != 0) {
						// calculate average bin error
						lw = GetEdgeLishStrength(lish);
					}
				}

				TGNGVisEdge *curVisEdge;
				if (usedVisEdges < vEdgeCnt) {
					curVisEdge = static_cast<TGNGVisEdge*>(fVisEdges.At(usedVisEdges++));
				} else {
					curVisEdge = new TGNGVisEdge();
					fVisEdges.Add(curVisEdge);
					++usedVisEdges;
					++vEdgeCnt;
				}

				Double_t x2 = otherPos->X();
				Double_t y2 = otherPos->Y();

				curVisEdge->fStrength = lw;
				curVisEdge->fX1       = (x1-minX) / dimX * mw + left_m;
				curVisEdge->fY1       = 1.0 - ((y1-minY) / dimY * mh) - top_m;
				curVisEdge->fX2       = (x2-minX) / dimX * mw + left_m;
				curVisEdge->fY2       = 1.0 - ((y2-minY) / dimY * mh) - top_m;
				curVisEdge->fFlag     = 1;

			}			
		}

		fVisEdges.UnSort();
		fVisEdges.Sort();

		for (Int_t i = 0; i < vEdgeCnt; ++i) {
			TGNGVisEdge *curVisEdge = static_cast<TGNGVisEdge*>(fVisEdges.At(i));

			if (curVisEdge->fFlag == 0)
				continue;

			if (fLishWeight == kTRUE) {
				Double_t lw = curVisEdge->fStrength;
				// we fake transparency with line color
				Int_t palSize = fPalA.size();
				Int_t colIdx = (Double_t)(palSize-1) * lw;
				if (colIdx > palSize-1) colIdx = palSize-1;
				if (colIdx < 0) colIdx = 0;
				//SetLineColor(TColor::GetColor((Float_t)(1.0 - lw), (Float_t)(1.0 - lw),(Float_t)( 1.0 - lw)));
				SetLineColor(fPalA[colIdx]);
				Double_t ew = 3.0 * (Double_t)fEdgeWidth * lw;
				if (ew < (Double_t)fEdgeWidth)
					ew = (Double_t)fEdgeWidth;
				SetLineWidth(ew);
				TAttLine::Modify();
			}

			// draw line
			fCurrentPad->PaintLine(curVisEdge->fX1,curVisEdge->fY1,curVisEdge->fX2,curVisEdge->fY2);

		}

	}

	// draw LISHs
	
	if (opt.Contains("l")) {

		SetFillColor(kBlack);
		SetFillStyle(1001);
		TAttFill::Modify();

		SetLineColor(kOrange-9);
		SetLineStyle(1);
		SetLineWidth(1);
		TAttLine::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit pos
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
				continue;
			}
			Double_t x1 = unitPos->X();
			Double_t y1 = unitPos->Y();
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
				if (curEdge->fUnitA != curUnit)
					continue;
				// get other units positions
				TVector2 *otherPos = static_cast<TVector2*>(curEdge->fUnitB->fAttributes.GetValue(&fPosAttr));
				if (otherPos == 0) {
					fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
					continue;
				}
				Double_t x2 = otherPos->X();
				Double_t y2 = otherPos->Y();

				// get lish
				TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&lishkey));
				if (lish == 0)
					continue;

				// calculate average bin error - sort of
				Double_t lw = GetEdgeLishStrength(lish);

				if (lw < 0.66)
					continue;

				// center pos on line
				Double_t cpx = ((x1+x2)/2.0 - minX) / dimX * mw + left_m;
				Double_t cpy = 1.0 - (((y1+y2)/2.0 - minY) / dimY * mh) - top_m;

				Double_t bw2 = 0.015;
				Double_t bh2 = 0.01;

				Double_t bx1 = cpx-bw2;
				Double_t by1 = cpy-bh2;
				Double_t bx2 = cpx+bw2;
				Double_t by2 = cpy+bh2;

				// paint black box
				fCurrentPad->PaintBox(bx1,by1,bx2,by2);

				Double_t bw = bx2-bx1;
				Double_t bh = by2-by1;

				Double_t xmargin = 0.002;
				Double_t ymargin = 0.002;

				// draw mini histogram
				Int_t nrOfBins  = lish->GetNbinsX();
				Int_t maxBin    = lish->GetMaximumBin();
				Double_t maxVal = lish->GetBinContent(maxBin);
				if (maxVal == 0.0) maxVal = 1.0;
				Double_t xFrac  = (bw - 2*xmargin)  / (Double_t)nrOfBins;
				Double_t yFact  = (bh - 2*ymargin)  / maxVal; // bh is negative

				Bool_t reverse = (curEdge->fUnitA != curUnit);

				//SetLineColor(fLishLineColor);
				SetLineColor(kOrange-9);
				TAttLine::Modify();
				for (Int_t k = 0; k < nrOfBins; ++k) {
					Double_t curVal = !reverse ? lish->GetBinContent(k+1) : 
											     lish->GetBinContent(nrOfBins-k);
					// draw line
					Double_t tmpX = bx1 + (Double_t)k * xFrac + xFrac / 2.0 + xmargin;
					fCurrentPad->PaintLine(tmpX,by1 + ymargin,tmpX,by1 + (curVal * yFact)+ ymargin);	
				}

				//SetLineColor(fLishErrorLineColor);
				SetLineColor(kRed);
				TAttLine::Modify();
				for (Int_t k = 0; k < nrOfBins; ++k) {
					Double_t curVal = !reverse ? lish->GetBinError(k+1) : 
											     lish->GetBinError(nrOfBins-k);
					// draw line
					Double_t tmpX = bx1 + (Double_t)k * xFrac + xFrac / 2.0 + xmargin;
					fCurrentPad->PaintLine(tmpX,by1 + ymargin,tmpX,by1 + (curVal * yFact)+ ymargin);	
				}



			}			
		}
	}
	

	// draw units
	if ((!opt.Contains("!u")) && (!opt.Contains("t"))) {

		TObjString sckey("SCINFO");

		SetMarkerColor(fUnitColor);
		SetMarkerSize(fUnitSize);
		SetMarkerStyle(fUnitStyle);
		TAttMarker::Modify();				

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit pos
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
				continue;
			}
			Double_t x1 = (unitPos->X() - minX) / dimX * mw + left_m;
			Double_t y1 = 1.0 - ((unitPos->Y() - minY) / dimY * mh) - top_m;
			// later: set style dependend on attributes

			if (opt.Contains("c")) {
				TGNGClusterInfo *cInfo = static_cast<TGNGClusterInfo*>(curUnit->fAttributes.GetValue(&sckey));
				if (cInfo != 0) {
					SetMarkerColor(cInfo->fClusterColor);
					TAttMarker::Modify();				

				} 
			}

			fCurrentPad->PaintPolyMarker(1,&x1,&y1);

		}

	}

	// draw uint IDs 
	if (opt.Contains("t")) {
		TObjString key("ID");
		TObjString sckey("SCINFO");

		SetFillColor(fUnitColor);
		SetFillStyle(1001);
		TAttFill::Modify();

		char buffer[10]; // buffer for node IDs
		SetTextAlign(22);
		SetTextColor(fUnitTextColor);
		SetTextFont(fUnitTextFont);
		SetTextSize(fUnitTextSize);
		TAttText::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit id
			TGNGId *uid = static_cast<TGNGId*>(curUnit->fAttributes.GetValue(&key));
			if (uid == 0)
				continue;

			// see if there is cluster info
			if (opt.Contains("c")) {
				TGNGClusterInfo *cInfo = static_cast<TGNGClusterInfo*>(curUnit->fAttributes.GetValue(&sckey));
				if (cInfo != 0) {
					SetFillColor(cInfo->fClusterColor);
					TAttFill::Modify();

					SetTextColor(cInfo->fClusterTextColor);
					TAttText::Modify();
				} 
			}

			// get unit pos
			TVector2 *unitPos = static_cast<TVector2*>(curUnit->fAttributes.GetValue(&fPosAttr));
			if (unitPos == 0) {
				fCurrentGNG->RaiseError(TGNG::ecExtension,"Unit has no attribute \"FruRePos\"!");
				continue;
			}
			Double_t x1 = (unitPos->X() - minX) / dimX * mw + left_m;
			Double_t y1 = 1.0 - ((unitPos->Y() - minY) / dimY * mh) - top_m;

			Double_t d = fUnitTextSize * 0.6;
			fCurrentPad->PaintBox(x1-d,y1+d,x1+d,y1-d);

			std::sprintf(buffer,"%d",uid->fID);
			fCurrentPad->PaintText(x1,y1+0.001,buffer);
		}
	}


	//fCurrentGNG = 0;
	//fCurrentBMU = 0;

	fLastUpdate = TTimeStamp();

}

void TGNGFruReVis::OnPeekResult(TGNG *gng, TGNGUnit *bmu, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGFruReVis::OnFeedResult(TGNG *gng, TGNGUnit *bmu, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGFruReVis::addPos(TGNG *gng, TGNGUnit *newUnit, Bool_t forceRnd )
{
	// add needed attributes
	if (newUnit->fAttributes.FindObject(&fPosAttr) != 0) {
		gng->RaiseError(TGNG::ecExtension,"Unit has already an attribute \"FruRePos\"!");
		return;
	}
	TVector2 *nPos = new TVector2(0.0,0.0);

	Int_t edgeCnt = newUnit->fEdges.GetEntriesFast();
	if ((edgeCnt > 1) && (forceRnd == kFALSE)) {
		Double_t minX = numeric_limits<Double_t>::max();
		Double_t maxX = numeric_limits<Double_t>::min();
		Double_t minY = numeric_limits<Double_t>::max();
		Double_t maxY = numeric_limits<Double_t>::min();

		for (Int_t i = 0; i < edgeCnt; ++i) {
			TGNGEdge *curEdge = static_cast<TGNGEdge*>(newUnit->fEdges.At(i));
			TGNGUnit *otherUnit = (curEdge->fUnitA != newUnit) ? curEdge->fUnitA : curEdge->fUnitB;
			TVector2 *otherPos = static_cast<TVector2*>(otherUnit->fAttributes.GetValue(&fPosAttr));
			if (otherPos == 0) {
				continue;
			}
			Double_t curX = otherPos->X();
			Double_t curY = otherPos->Y();
			if (curX < minX) minX = curX;
			if (curX > maxX) maxX = curX;
			if (curY < minY) minY = curY;
			if (curY > maxY) maxY = curY;
		}

		Double_t dimX = maxX - minX;
		Double_t dimY = maxY - minY;

		//nPos->Set(minX + gRandom->Uniform(dimX),minY + gRandom->Uniform(dimY));
		nPos->Set(minX + dimX / 2.0 + gRandom->Uniform(dimX*0.1),minY + dimY / 2.0 + gRandom->Uniform(dimY*0.1));

	} else {
		nPos->Set(gRandom->Uniform(1000),gRandom->Uniform(1000));
	}

	newUnit->fAttributes.Add(new TObjString(fPosAttr),nPos);

	fTemp = 1.0;
}

void TGNGFruReVis::OnNewUnit(TGNG *gng, TGNGUnit *newUnit)
{
	addPos(gng,newUnit);
}

void TGNGFruReVis::SetUpdateInterval(Double_t sec)
{
	fUpdateInterval = sec;
}

void TGNGFruReVis::SetIterations(Int_t iter)
{
	fIterations = iter;
}

void TGNGFruReVis::SetEdgeColor(Color_t color)
{
	fEdgeColor = color;
}

void TGNGFruReVis::SetEdgeStyle(Style_t style)
{
	fEdgeStyle = style;
}

void TGNGFruReVis::SetEdgeWidth(Width_t width)
{
	fEdgeWidth = width;
}


void TGNGFruReVis::SetUnitColor(Color_t color)
{
	fUnitColor = color;
}

void TGNGFruReVis::SetUnitSize(Size_t size)
{
	fUnitSize = size;
}

void TGNGFruReVis::SetUnitStyle(Style_t style)
{
	fUnitStyle = style;
}

void TGNGFruReVis::SetUnitTextColor(Color_t color)
{
	fUnitTextColor = color;
}

void TGNGFruReVis::SetUnitTextFont(Font_t font)
{
	fUnitTextFont = font;
}

void TGNGFruReVis::SetUnitTextSize(Float_t size)
{
	fUnitTextSize = size;
}



void TGNGFruReVis::UpdateVis()
{
	TTimeStamp current;
	Double_t delta = current - fLastUpdate;
	if (delta < fUpdateInterval)
		return;

	if (fCurrentPad) {
		fCurrentPad->Modified();
		fCurrentPad->Update();
		gSystem->ProcessEvents();
	}	

	//fLastUpdate = current;
}

void TGNGFruReVis::SetLishWeighting(Bool_t enable)
{
	fLishWeight = enable;
}

void TGNGFruReVis::SetLishWeightCenter(Double_t center)
{
	fLishWeightCenter = center;
}

void TGNGFruReVis::SetLishWeightSteepness(Double_t steepness)
{
	fLishWeightSteepness = steepness;
}

void TGNGFruReVis::SetPalette(Int_t a, Bool_t invert)
{
	fPaletteA = a;

	fPalA.clear();

	gStyle->SetPalette(fPaletteA);
	Int_t palSize = gStyle->GetNumberOfColors();
	for (Int_t i = 0; i < palSize; ++i)
		fPalA.push_back(gStyle->GetColorPalette(invert ? palSize-1-i : i));

}

void TGNGFruReVis::SetLocalSpace(Double_t localSpace)
{
	fLocalSpace = localSpace;
}
