/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a gng unit visualization. In case of 
1D vector data it draws the prototype vector of each unit as a column of colored
boxes.
*/
#include <limits>
#include <cstdio>
#include <TSystem.h>
#include <TObjString.h>
#include <TH1D.h>
#include <TRandom.h>
#include <TVector2.h>
#include <TMath.h>
#include <TStyle.h>
#include <TPad.h>

#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../core/TGNGDataElement.h"
#include "../functional/TGNGIdExt.h"

#include "TGNGUnitVis.h"


TGNGUnitVis::TGNGUnitVis() :
	TGNGExtension(),
	TAttLine(),
	TAttMarker(),
	TAttText(),
	TAttFill(),
	fCurrentPad(0),
	fUpdateInterval(0.05),
	fLastUpdate(),
	fCurrentGNG(0),
	fCurrentBMU(0),
	fEdgeColor(kGray+1),
	fEdgeStyle(1),
	fEdgeWidth(1),
	fTableTextTopOffsetX(0.0),
	fTableTextTopOffsetY(-0.1),
	fTableTextColor(kBlue+2),
	fTableTextFont(100),
	fTableTextSize(0.8),
	fMinUnit(-1),
	fMaxUnit(-1),
	fRows(1),
	fPaletteA(51),
	fPaletteB(51)
{
	SetPalettes(fPaletteA,fPaletteB);
}


void TGNGUnitVis::Draw(Option_t* option)
{
	// draws the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e  - do not draw matrix lines
	// t    - print node IDs 
	// tr   - print raw node idx (position in TObjArray) instead of IDs 

	// clear current Pad
	if (gPad)
		gPad->Clear(); 

	TObject::Draw(option);
	fCurrentPad = gPad;
}

void TGNGUnitVis::Paint(Option_t *option)
{
	// paints the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e   - do not draw matrix lines
	// t    - print node IDs 
	// tr   - print raw node idx (position in TObjArray) instead of IDs 

	if ((fCurrentGNG == 0) || (fCurrentPad == 0)) {
		fLastUpdate = TTimeStamp();
		return;
	}

	Int_t nrOfUnits = fCurrentGNG->fUnits.GetEntriesFast();	

	Int_t maxUnit = nrOfUnits - 1;
	if ((fMaxUnit > 0) && (fMaxUnit < nrOfUnits))
		maxUnit = fMaxUnit;

	Int_t minUnit = 0;
	if ((fMinUnit > 0) && (fMinUnit <= maxUnit))
		minUnit = fMinUnit;

	Int_t range = maxUnit - minUnit + 1;

	Double_t dimX = (Double_t)(range);
	Double_t dimY = 50;


	// get pad dimensions
	Double_t to_w = fCurrentPad->UtoPixel(1.);
	Double_t to_h = fCurrentPad->VtoPixel(0.);

	Double_t left_m   = fCurrentPad->GetLeftMargin();
	Double_t right_m  = fCurrentPad->GetLeftMargin();
	Double_t top_m    = fCurrentPad->GetTopMargin();
	Double_t bottom_m = fCurrentPad->GetBottomMargin();

	Double_t mw = 1.0 - left_m - right_m;
	Double_t mh = 1.0 - top_m - bottom_m;

	TString opt = option;
	opt.ToLower();

	// draw uint IDs 
	if (opt.Contains("t")) {
		TObjString key("ID");

		char buffer[10]; // buffer for node IDs
		SetTextAlign(21);
		SetTextColor(fTableTextColor);
		SetTextFont(fTableTextFont);
		SetTextSize(TMath::Min((1.0 / dimX) * fTableTextSize,(1.0 / dimY) * fTableTextSize));
		TAttText::Modify();

		for (Int_t i = minUnit; i <= maxUnit; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));

			Int_t idx = -1;
			// get unit id
			if (opt.Contains("tr")) {
				idx = i;
			} else {
				TGNGId *uid = static_cast<TGNGId*>(curUnit->fAttributes.GetValue(&key));
				if (uid == 0)
					continue;
				idx = uid->fID;
			}

			Double_t x1 = ((Double_t)(i-minUnit+fTableTextTopOffsetX) + 0.5) / dimX * mw + left_m;
			Double_t y1 = 1.0 - ((1.0 + fTableTextTopOffsetY) / dimY * mh) - top_m;

			std::sprintf(buffer,"%d",idx);
			fCurrentPad->PaintText(x1,y1,buffer);

		}
	}

	// draw prototypes
	for (Int_t i = minUnit; i <= maxUnit; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));

		Double_t x1 = ( (Double_t)(i-minUnit) ) / dimX * mw + left_m;
		Double_t y1 = (1.0 / dimY * mh) + bottom_m;
		
		Double_t x2 = ( (Double_t)(i-minUnit+1) ) / dimX * mw + left_m;
		Double_t y2 = 1.0 - (1.0 / dimY * mh) - top_m;		

		curUnit->fPrototype->PaintElement(fCurrentPad,x1,y1,x2,y2,((i & 1) == 0) ? &fPalA : &fPalB);
	}



	// draw matrix lines
	if (!opt.Contains("!e")) {

		SetLineColor(fEdgeColor);
		SetLineStyle(fEdgeStyle);
		SetLineWidth(fEdgeWidth);
		TAttLine::Modify();

		for (Int_t i = minUnit; i < maxUnit; ++i) {

			Double_t x1 = i-minUnit+1;
			Double_t y1 = 0;

			Double_t x2 = i-minUnit+1;
			Double_t y2 = dimY;

			fCurrentPad->PaintLine(x1 / dimX * mw + left_m,
							1.0 - (y1 / dimY * mh) - top_m,
				            x2 / dimX * mw + left_m,
				            1.0 - (y2 / dimY * mh) - top_m);
		}

	}
	
	fLastUpdate = TTimeStamp();

}

void TGNGUnitVis::OnPeekResult(TGNG *gng, TGNGUnit *bmu, Double_t,
							   TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGUnitVis::OnFeedResult(TGNG *gng, TGNGUnit *bmu, Double_t,
							   TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGUnitVis::SetUpdateInterval(Double_t sec)
{
	fUpdateInterval = sec;
}

void TGNGUnitVis::SetRange(Int_t minUnit, Int_t maxUnit)
{
	fMinUnit = minUnit;
	fMaxUnit = maxUnit;
}

void TGNGUnitVis::SetEdgeColor(Color_t color)
{
	fEdgeColor = color;
}

void TGNGUnitVis::SetEdgeStyle(Style_t style)
{
	fEdgeStyle = style;
}

void TGNGUnitVis::SetEdgeWidth(Width_t width)
{
	fEdgeWidth = width;
}

void TGNGUnitVis::SetTableTextTopOffsetX(Double_t ox)
{
	fTableTextTopOffsetX = ox;
}

void TGNGUnitVis::SetTableTextTopOffsetY(Double_t oy)
{
	fTableTextTopOffsetY = oy;
}

void TGNGUnitVis::SetTableTextColor(Color_t color)
{
	fTableTextColor = color;
}

void TGNGUnitVis::SetTableTextFont(Font_t font)
{
	fTableTextFont = font;
}

void TGNGUnitVis::SetTableTextSize(Float_t size)
{
	fTableTextSize = size;
}

void TGNGUnitVis::UpdateVis()
{
	TTimeStamp current;
	Double_t delta = current - fLastUpdate;
	if (delta < fUpdateInterval)
		return;

	if (fCurrentPad) {
		fCurrentPad->Modified();
		fCurrentPad->Update();
		gSystem->ProcessEvents();
	}	

	//fLastUpdate = current;
}

void TGNGUnitVis::SetRows(Int_t rows)
{
	fRows = rows;
}

void TGNGUnitVis::SetPalettes(Int_t a, Int_t b)
{
	// The gng unit prototypes can be visualized using two alternating palettes
	// "a" and "b". If you do not whish to use alternating palettes, just set
	// both palettes to the same value.

	fPaletteA = a;
	fPaletteB = b;

	fPalA.clear();
	fPalB.clear();

	gStyle->SetPalette(fPaletteA);
	Int_t palSize = gStyle->GetNumberOfColors();
	for (Int_t i = 0; i < palSize; ++i)
		fPalA.push_back(gStyle->GetColorPalette(i));

	gStyle->SetPalette(fPaletteB);
	palSize = gStyle->GetNumberOfColors();
	for (Int_t i = 0; i < palSize; ++i)
		fPalB.push_back(gStyle->GetColorPalette(i));

}
